/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for boxlist utility (boxlist.c)
   created: Mark Wahl DL4YBG 94/10/03
   updated: Mark Wahl DL4YBG 95/01/05
*/

#include "tnt.h"
#include "boxlist.h"

extern int act_mode;
extern int act_channel;
extern char download_dir[];
extern int LINES,COLS;
extern char ok_text[];
extern struct window statlin;
extern int txecho_flag;
extern char att_monitor;
extern char att_normal;
extern struct rx_file rx_file[CHANNELS];
#ifdef USE_IFACE
extern struct rx_file mb_file;
#endif

int blist_add_plus;
struct boxlist_file boxlist_file[CHANNELS];
static struct window boxlistwin[CHANNELS];
struct real_layout boxlist_layout[CHANNELS][5];

#ifdef USE_IFACE
struct boxlist_file mbboxlist_file;
static struct window mbboxlistwin;
struct real_layout mbboxlist_layout[5];
int bl_mode;
extern struct rx_file mb_file;
#endif

static struct tmpname_entry *tmpname_root;


/* display line on boxlist-screen */
static void blist_textout_len(buf,len,mode,channel)
char *buf;
int len;
int mode;
int channel;
{
#ifdef USE_IFACE
  if (mode == M_MAILBOX) {
    win_textout_len(buf,len,&mbboxlistwin,1);
  }
  else {
    win_textout_len(buf,len,&boxlistwin[channel],1);
  }
#else
  win_textout_len(buf,len,&boxlistwin[channel],1);
#endif
}

/* character to boxlist screen */
void blist_charout_cntl(channel,ch)
int channel;
char ch;
{
#ifdef USE_IFACE
  if (bl_mode == M_MAILBOX) {
    win_charout_cntl(ch,&mbboxlistwin);
  }
  else {
    win_charout_cntl(ch,&boxlistwin[channel]);
  }
#else
  win_charout_cntl(ch,&boxlistwin[channel]);
#endif
}

/* mark selected line in boxlist screen */
static void blist_mark_line(channel,str,len)
int channel;
char *str;
int len;
{
#ifdef USE_IFACE
  if (bl_mode == M_MAILBOX) {
    win_attrib(att_monitor,&mbboxlistwin);
    win_textout_len(str,len,&mbboxlistwin,1);
    win_attrib(att_normal,&mbboxlistwin);
    win_charout_cntl(C_STLINE,&mbboxlistwin);
    win_charout_cntl(C_CUUP,&mbboxlistwin);
  }
  else {
    win_attrib(att_monitor,&boxlistwin[channel]);
    win_textout_len(str,len,&boxlistwin[channel],1);
    win_attrib(att_normal,&boxlistwin[channel]);
    win_charout_cntl(C_STLINE,&boxlistwin[channel]);
    win_charout_cntl(C_CUUP,&boxlistwin[channel]);
  }
#else
  win_attrib(att_monitor,&boxlistwin[channel]);
  win_textout_len(str,len,&boxlistwin[channel],1);
  win_attrib(att_normal,&boxlistwin[channel]);
  win_charout_cntl(C_STLINE,&boxlistwin[channel]);
  win_charout_cntl(C_CUUP,&boxlistwin[channel]);
#endif
}

/* send generated line */
static void blist_send_line(channel,str,len)
int channel;
char *str;
int len;
{
  int i;
  
#ifdef USE_IFACE
  if (bl_mode == M_MAILBOX) {
    mb_input(str,len);
    for (i=0;i<len;i++) {
      if (str[i] == '\r') {
        mb_nextline();
      }
      else {
        mb_charout(str[i]);
      }
    }
    if (txecho_flag) {
      rem_mb_display_buf(str,len);
    }
  }
  else {
    dat_input(act_channel,str,len);
    for (i=0;i<len;i++) {
      if (str[i] == '\r') {
        pre_nextline(channel);
      }
      else {
        pre_charout(channel,str[i]);
      }
    }
    if (txecho_flag) {
      insert_cr_rx(channel);
      rem_data_display_buf(channel,str,len);
    }
  }
#else
  dat_input(act_channel,str,len);
  for (i=0;i<len;i++) {
    if (str[i] == '\r') {
      pre_nextline(channel);
    }
    else {
      pre_charout(channel,str[i]);
    }
  }
  if (txecho_flag) {
    insert_cr_rx(channel);
    rem_data_display_buf(channel,str,len);
  }
#endif
}

/* analyse boxlist line */
void blist_analyse_line(channel,func)
int channel;
int func;
{
  char str[MAXCOLS+1];
  int len;
  char call[MAXCOLS+1];
  char file[MAXCOLS+1];
  char date[MAXCOLS+1];
  char time[MAXCOLS+1];
  char mbx[MAXCOLS+1];
  char title[MAXCOLS+1];
  int check_nr;
  int bytes;
  int lt;
  int i;
  int res;
  char rubrik[MAXCOLS+1];
  int number;

#ifdef USE_IFACE
  if (bl_mode == M_MAILBOX) {
    win_charout_cntl(C_EOLINE,&mbboxlistwin);
    len = get_line_noconv(str,&mbboxlistwin);
    win_charout_cntl(C_STLINE,&mbboxlistwin);
  }
  else {
    win_charout_cntl(C_EOLINE,&boxlistwin[channel]);
    len = get_line_noconv(str,&boxlistwin[channel]);
    win_charout_cntl(C_STLINE,&boxlistwin[channel]);
  }
#else  
  win_charout_cntl(C_EOLINE,&boxlistwin[channel]);
  len = get_line_noconv(str,&boxlistwin[channel]);
  win_charout_cntl(C_STLINE,&boxlistwin[channel]);
#endif

/* DIEBOX checklist 

    7 DL4BCU > TERMINE...16 24.09.94 DL      2214   5 2m Mobilfuchsjagd I05 08.

*/  

  res = sscanf(str,"%d %s > %s %s %s %d %d %s",
               &check_nr, call, file, date, mbx, &bytes, &lt, title);
  if (res != 8) {
    /* special scan if mbx is empty */
    res = sscanf(str,"%d %s > %s %s %d %d %s",
               &check_nr, call, file, date, &bytes, &lt, title);
    if (res == 7) {
      mbx[0] = '\0';
      res++;
    }
  }
  if (res == 8) {
    for (i=0;i<strlen(file);i++) {
      if (file[i] == '.') file[i] = ' ';
    }
    res = sscanf(file,"%s %d",rubrik,&number);
    if (res != 2) {
      strncpy(rubrik,file,8);
      rubrik[8] = '\0';
      sscanf(file+8,"%d",&number);
    }
    blist_mark_line(channel,str,len);
    switch (func) {
    case FUNC_READ:
      if (blist_add_plus)
        sprintf(str,"R %s %d +\r",rubrik,number);
      else
        sprintf(str,"R %s %d\r",rubrik,number);
      break;
    case FUNC_LIST:
      sprintf(str,"L %s %d\r",rubrik,number);
      break;
    case FUNC_ERASE:
      sprintf(str,"E %s %d\r",rubrik,number);
      break;
    case FUNC_KILL:
      sprintf(str,"K %s %d\r",rubrik,number);
      break;
    }
    len = strlen(str);
    blist_send_line(channel,str,len);
    return;
  }

/* DIEBOX list 

 263 DL1ZAX 02.11.94 18:03   6763  DL-RUNDSPRUCH NR. 39/94

*/  

  res = sscanf(str,"%d %s %s %s %d %s",
               &number, call, date, time, &bytes, title);
  if (res == 6) {
    blist_mark_line(channel,str,len);
    switch (func) {
    case FUNC_READ:
      if (blist_add_plus)
        sprintf(str,"R %d +\r",number);
      else
        sprintf(str,"R %d\r",number);
      break;
    case FUNC_LIST:
      sprintf(str,"L %d\r",number);
      break;
    case FUNC_ERASE:
      sprintf(str,"E %d\r",number);
      break;
    case FUNC_KILL:
      sprintf(str,"K %d\r",number);
      break;
    }
    len = strlen(str);
    blist_send_line(channel,str,len);
    return;
  }
  
/* all other not yet implemented */

  beep();
  return;
}

/* select boxlist screen */
void sel_boxlist()
{
  if (act_mode == M_BOXLIST) {
#ifdef USE_IFACE
    if (bl_mode == M_MAILBOX) {
      sel_mailbox();
    }
    else {
      sel_connect(act_channel);
    }
#else
    sel_connect(act_channel);
#endif
  }
  else {
#ifdef USE_IFACE  
    if (act_mode == M_MAILBOX) {
      if (mbboxlist_file.type == -1) return;
      real_screen(&mbboxlist_layout[0]);
      bl_mode = M_MAILBOX;
    }
    else {
      if (boxlist_file[act_channel].type == -1) return;
      real_screen(&boxlist_layout[act_channel][0]);
      bl_mode = M_CONNECT;
    }
#else
    if (boxlist_file[act_channel].type == -1) return;
    real_screen(&boxlist_layout[act_channel][0]);
#endif
    act_mode = M_BOXLIST;
    statlin_update();
  }
}

/* open a temporary file for boxlist */
void cmd_logblist(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  char *tmpname;
  struct tmpname_entry *newentry;
  struct tmpname_entry *listentry;
  
  tmpname = strdup("/tmp/tntbliXXXXXX");
  mktemp(tmpname);
  /* do entry in tempfile list */
  newentry = (struct tmpname_entry *)malloc(sizeof(struct tmpname_entry));
  newentry->name = tmpname;
  newentry->next = NULL;
  if (tmpname_root == NULL) {
    tmpname_root = newentry;
  }
  else {
    listentry = tmpname_root;
    while (listentry->next != NULL) {
      listentry = listentry->next;
    }
    listentry->next = newentry;    
  }
  
  open_logfile(RX_NORM,RX_RCV,channel,strlen(tmpname),mode,tmpname);
}

#define BUFSIZE 512

/* start a boxlist screen with specified file */
void cmd_blist(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int file_len;
  int end;
  int len2;
  char buf[BUFSIZE];
  int i;
  int lines;
  int lines2;
  struct boxlist_file *blfile;
  char *linebuf;
  int linelen;

#ifdef USE_IFACE  
  if (mode == M_MAILBOX) {
    blfile = &mbboxlist_file;
  }
  else {
    blfile = &boxlist_file[channel];
  }
#else
  blfile = &boxlist_file[channel];
#endif
  
  if (blfile->type != -1) {
    cmd_display(mode,channel,"Boxlist already active",1);
    return;
  }
  if (len == 0) {
    close_file(1,0,channel,0,mode,NULL);
#ifdef USE_IFACE
    if ((mode == M_MAILBOX) && (mb_file.name[0] != '\0')) {
      strcpy(blfile->name,mb_file.name);
    }
    else if ((mode != M_MAILBOX) && (rx_file[channel].name[0] != '\0')) {
      strcpy(blfile->name,rx_file[channel].name);
    }
#else
    if (rx_file[channel].name[0] != '\0') {
      strcpy(blfile->name,rx_file[channel].name);
    }
#endif
    else {
      cmd_display(mode,channel,"No boxlist file stored",1);
      return;
    }
  }
  else {
    if(strchr(str,'/') != NULL) {
      strcpy(blfile->name,str);
    }
    else {
      strcpy(blfile->name,download_dir);
      strcat(blfile->name,str);
    }
  }
  blfile->fd = open(blfile->name,O_RDONLY); 
  if (blfile->fd == -1) {
    /* file does not exist */
    cmd_display(mode,channel,"File not existing",1);
    return;
  }
  
  blfile->type = BL_NORMAL;
  /* counting number of LFs in file to get number of lines */
  file_len = 0;
  lines = 0;
  end = 1;
  while (end) {
    len2 = read(blfile->fd,buf,BUFSIZE);
    file_len += len2;
    for (i=0;i<len2;i++) {
      if (buf[i] == '\n') lines++;
    }
    if (len2 < BUFSIZE) end = 0;
  }
  blfile->len = file_len;
  lines += 2; /* just to be sure... */

#ifdef USE_IFACE
  if (mode == M_MAILBOX) {
    open_window(lines,&mbboxlistwin,1);
    mbboxlist_layout[0].win = &mbboxlistwin;
    mbboxlist_layout[0].first_real_line = 0;
    mbboxlist_layout[0].win_num_lines = LINES-1;
    mbboxlist_layout[0].pagesize = LINES-4;
    mbboxlist_layout[1].win = &statlin;
    mbboxlist_layout[1].first_real_line = LINES-1;
    mbboxlist_layout[1].win_num_lines = 1;
    mbboxlist_layout[1].pagesize = 0;
  }
  else {
    open_window(lines,&boxlistwin[channel],1);
    boxlist_layout[channel][0].win = &boxlistwin[channel];
    boxlist_layout[channel][0].first_real_line = 0;
    boxlist_layout[channel][0].win_num_lines = LINES-1;
    boxlist_layout[channel][0].pagesize = LINES-4;
    boxlist_layout[channel][1].win = &statlin;
    boxlist_layout[channel][1].first_real_line = LINES-1;
    boxlist_layout[channel][1].win_num_lines = 1;
    boxlist_layout[channel][1].pagesize = 0;
  }
#else
  open_window(lines,&boxlistwin[channel],1);
  boxlist_layout[channel][0].win = &boxlistwin[channel];
  boxlist_layout[channel][0].first_real_line = 0;
  boxlist_layout[channel][0].win_num_lines = LINES-1;
  boxlist_layout[channel][0].pagesize = LINES-4;
  boxlist_layout[channel][1].win = &statlin;
  boxlist_layout[channel][1].first_real_line = LINES-1;
  boxlist_layout[channel][1].win_num_lines = 1;
  boxlist_layout[channel][1].pagesize = 0;
#endif
  
  /* seek to file begin */
  lseek(blfile->fd,0,SEEK_SET);
  linelen = 0;
  end = 1;
  linebuf = (char *)malloc(COLS+2);
  while (end) {
    len2 = read(blfile->fd,buf,BUFSIZE);
    for (i=0;i<len2;i++) {
      if (buf[i] == '\n') {
        linebuf[linelen] = '\r';
        linelen++;
        blist_textout_len(linebuf,linelen,mode,channel);
        linelen = 0;
      }
      else {
        if (linelen < (COLS-1)) {
          linebuf[linelen] = buf[i];
          linelen++;
        }
      }
    }
    if (len2 < BUFSIZE) end = 0;
  }
  if (linelen) {
    linebuf[linelen] = '\r';
    linelen++;
    blist_textout_len(linebuf,linelen,mode,channel);
    linelen = 0;
  }
  free(linebuf);
  close(blfile->fd);
  cmd_display(mode,channel,ok_text,1);
}

/* finish boxlist */
void cmd_xblist(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int i;
  struct boxlist_file *blfile;

#ifdef USE_IFACE  
  if (mode == M_MAILBOX) {
    blfile = &mbboxlist_file;
  }
  else {
    blfile = &boxlist_file[channel];
  }
#else
  blfile = &boxlist_file[channel];
#endif
  
  if (blfile->type == -1) {
    cmd_display(mode,channel,"No boxlist active",1);
    return;
  }

#ifdef USE_IFACE
  if (mode == M_MAILBOX) {
    close_window(&mbboxlistwin);
  }
  else {
    close_window(&boxlistwin[channel]);
  }
#else
  close_window(&boxlistwin[channel]);
#endif

  blfile->type = -1;  
  cmd_display(mode,channel,ok_text,1);
}

/* init of boxlist */
void init_blist()
{
  int i;
  int j;


  tmpname_root = NULL;
  
  for (i=0;i<CHANNELS;i++) {
    boxlist_file[i].type = -1;
    for (j = 0; j < 5; j++) {
      boxlist_layout[i][j].win = NULL;
      boxlist_layout[i][j].first_real_line = 0;
      boxlist_layout[i][j].win_num_lines = 0;
      boxlist_layout[i][j].pagesize = 0;
    }
  }
#ifdef USE_IFACE
  mbboxlist_file.type = -1;
  for (j = 0; j < 5; j++) {
    mbboxlist_layout[j].win = NULL;
    mbboxlist_layout[j].first_real_line = 0;
    mbboxlist_layout[j].win_num_lines = 0;
    mbboxlist_layout[j].pagesize = 0;
  }
  bl_mode = M_CONNECT;
#endif  
}

/* exit of boxlist */
void exit_blist()
{
  int i;
  int j;
  struct tmpname_entry *listentry;
  struct tmpname_entry *oldentry;
  
  if (tmpname_root != NULL) {
    listentry = tmpname_root;
    do {
      oldentry = listentry;
      listentry = oldentry->next;
      unlink(oldentry->name);
      free(oldentry->name);
      free(oldentry);
    } while (listentry != NULL);
  }
  for (i=0;i<CHANNELS;i++) {
    if (boxlist_file[i].type != -1) {
      close_window(&boxlistwin[i]);
      boxlist_file[i].type = -1;
    }
  }
#ifdef USE_IFACE
  if (mbboxlist_file.type != -1) {
    close_window(&mbboxlistwin);
    mbboxlist_file.type = -1;
  }
#endif  
}
