/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993,1994 by Mark Wahl
   For license details see documentation
   Procedures for connect (connect.c)
   created: Mark Wahl DL4YBG 94/03/12
   updated: Mark Wahl DL4YBG 94/12/21
*/

#include "tnt.h"
#include "connect.h"

extern struct channel_stat ch_stat[CHANNELS];
extern char name_file_name[];
extern char tnt_dir[];
extern int tnc_command;
extern char ok_text[];

char connect_str[] = "onnected to ";

static char xc_active_text[] = "XCONNECT already active";
static char xc_notfound_text[] = "Routing file not found";
static char xc_corrupt_text[] = "Routing file corrupt";
static char xc_maxdigis_text[] = "Digipeater-table full";
static char xc_maxalloc_text[] = "Too many recursions";
static char xc_abort_text[] = "XCONNECT aborted";
static char xc_noxconn_text[] = "XCONNECT not active";
static char xc_inv_arg_text[] = "invalid arguments";
static char xc_inv_call_text[] = "invalid callsign";

static struct xconn_stat xc_stat[CHANNELS];

static int test_connected(channel,buffer)
int channel;
char *buffer;
{
  char str_buf[257];
  int len;
  char *str_ptr;
  int result;
  
  
  /* copy buffer to string, \0 may occure in buffer, but if connected
     is assumed to be received, plain ASCII will be received */
  result = 0;
  len = (int)(*buffer + 1);
  memcpy(str_buf,buffer+1,len);
  str_buf[len] = '\0';
  /* try to find connect-string */
  if ((str_ptr = strstr(str_buf,connect_str)) == NULL) return(T_NOCON);
  /* check if connected */
  if (str_ptr > (str_buf + 2)) {
    if ((tolower(*(str_ptr - 1)) == 'c') && ((*(str_ptr - 2)) == ' '))
      result = T_CON;
  }
  /* else check if reconnected */
  if (!result && (str_ptr > (str_buf + 3))) {
    if (((*(str_ptr - 1)) == 'c') && ((*(str_ptr - 2)) == 'e') &&
        (tolower(*(str_ptr - 3)) == 'r'))
      result = T_RECON;
  }
  if (!result) return(result);
  str_ptr += strlen(connect_str);
  /* now get new callsign */
  if (sscanf(str_ptr,"%30s",ch_stat[channel].disp_call) != 1) return(0);
  stat_display(channel);
  if (result == T_CON) /* if connect copy to call for logbook */
    strcpy(ch_stat[channel].log_call,ch_stat[channel].disp_call);
  return(result);
}

#define LINELENGTH 127

static int route_call(channel,len,mode,str,nextlink,callflag)
int channel;
int len;
int mode;
char *str;
int nextlink;
int callflag;
{  
  int found;
  int corrupt;
  int num;
  char *line_ptr;
  FILE *xc_fp;
  char *call_ptr;
  char *tmp_ptr;
  char tmpstr[160];

  strcpy(tmpstr,tnt_dir);
  strcat(tmpstr,name_file_name);
  if ((xc_fp = fopen(tmpstr,"r")) == NULL) {
    cmd_display(mode,channel,xc_notfound_text,1);
    return(0);
  }
  found = 0;
  line_ptr = malloc(LINELENGTH + 1);
  while (!found) {
    corrupt = 1;
    if ((fgets(line_ptr,LINELENGTH,xc_fp)) != NULL) {
      if (*(line_ptr + strlen(line_ptr) -1) == '\n') {
        if ((call_ptr = strtok(line_ptr," \t")) != NULL ) {
          corrupt = 0;
          tmp_ptr = callflag ? call_ptr : (call_ptr + 2);
          if (strcmp(tmp_ptr,str) == 0) {
            found = 1;
          }
        }
      }
    }
    else {
      return(C_NOTFOUND);
    }
    if (corrupt) {
      cmd_display(mode,channel,xc_corrupt_text,1);
      free(line_ptr);
      fclose(xc_fp);
      return(0);
    }
  }
  tmp_ptr = strtok(NULL,";"); 
  if ((tmp_ptr == NULL) || (*(tmp_ptr + strlen(tmp_ptr) - 1) == '\n')) {
    /* station can be reached directly */
    free(line_ptr);
    fclose(xc_fp);
    return(C_DIRECT);  
  }
  /* station has routing data available */
  num = xc_stat[channel].num_digi;
  while ((xc_stat[channel].digiinfo[num].nextdigi =
         strtok(NULL," \t")) != NULL) {
    tmp_ptr = xc_stat[channel].digiinfo[num].nextdigi;
    if (*(tmp_ptr = tmp_ptr + strlen(tmp_ptr) - 1) == '\n')
      *tmp_ptr = '\0';
    num++;
    if (num >= MAXDIGIS) {
      cmd_display(mode,channel,xc_maxdigis_text,1);
      free(line_ptr);
      fclose(xc_fp);
      return(0);
    }
    xc_stat[channel].digiinfo[num - 1].nextlink = num;
  }
  if (num == xc_stat[channel].num_digi) {
    /* no routing data found after ; */
    cmd_display(mode,channel,xc_corrupt_text,1);
    free(line_ptr);
    fclose(xc_fp);
    return(0);
  }
  xc_stat[channel].digiinfo[num - 1].nextlink = nextlink;
  xc_stat[channel].num_digi = num;
  xc_stat[channel].alloc[xc_stat[channel].num_alloc] = line_ptr;
  xc_stat[channel].num_alloc++;
  fclose(xc_fp);
  return(C_ROUTE);
}

static int gen_con_string(channel,str,ctype)
int channel;
char *str;
int ctype;
{
  char tmp_str[256];
  char *tmp_ptr;
  int nodetype;
  int digi;
  
  strcpy(tmp_str,"");
  digi = xc_stat[channel].cur_digi;
  while (1) {
    tmp_ptr = xc_stat[channel].digiinfo[digi].nextdigi;
    digi = xc_stat[channel].digiinfo[digi].nextlink;
    xc_stat[channel].cur_digi = digi;
    if (digi == D_EOL) {
      nodetype = CT_END;
      break;
    }
    if (*tmp_ptr == 'N') {
      nodetype = CT_NODE;
      break;
    }
    if (*tmp_ptr == 'T') {
      nodetype = CT_USER;
      break;
    }
    strcat(tmp_str," "); 
    strcat(tmp_str,tmp_ptr+2);
  }
  switch (ctype) {
  case CT_CONNECT:
    strcpy(str,"C");
    break;
  case CT_NODE:
    strcpy(str,"C ");
    break;
  case CT_USER:
    strcpy(str,"//C ");
    break;
  }
  strcat(str,tmp_ptr+2);
  strcat(str,tmp_str);
  if ((ctype == CT_NODE) || (ctype == CT_USER)) {
    strcat(str,"\r");
  }
  return(nodetype);
}

static void free_alloc(channel)
int channel;
{
  int i;
  
  if (xc_stat[channel].num_alloc) {
    for (i = 0; i < xc_stat[channel].num_alloc; i++) {
      free(xc_stat[channel].alloc[i]);
    }
  }
}

void close_xconnect(channel,success)
int channel;
int success;
{
#ifdef USE_IFACE
  char socket_str[256];
#endif

  if (xc_stat[channel].active) {
    if (success) {
#ifdef USE_IFACE
      if (xc_stat[channel].inform_box) {
        get_iface_name(xc_stat[channel].iface,socket_str);
        cmd_actiface(1,1,channel,strlen(socket_str),M_CMDSCRIPT,socket_str);
      }
#endif
    }
    else {
#ifdef USE_IFACE
      if (xc_stat[channel].inform_box) {
        abort_routing(xc_stat[channel].callsign);
      }
#endif
    }
    xc_stat[channel].active = 0;
    statlin_update();
    free_alloc(channel);
  }
}

void cmd_xconnect(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int result;
  char tnc_string[256];
  int i;
  int first_digi;
  int second_digi;
  int digi;
  int expand_end;
  char call[256];
  int timeout;
  int numpar;
  
  numpar = sscanf(str,"%s %d",call,&timeout);
  switch (numpar) {
  case 1:
    timeout = 0;
    break;
  case 2:
    break;
  default:
    cmd_display(mode,channel,xc_inv_arg_text,1);
    return;
    break;
  }
  if (strlen(call) > 9) {
    cmd_display(mode,channel,xc_inv_call_text,1);
    return;
  }
  for (i = 0; i < strlen(call); i++) {
    call[i] = toupper(call[i]);
  }
  if (xc_stat[channel].active) {
    if (strcmp(call,"OFF") == 0) {
      close_xconnect(channel,0);
      cmd_display(mode,channel,xc_abort_text,1);
      return;
    }
    else {
      cmd_display(mode,channel,xc_active_text,1);
      return;
    }
  }
  if (strcmp(call,"OFF") == 0) {
    cmd_display(mode,channel,xc_noxconn_text,1);
    return;
  }
  xc_stat[channel].num_digi = 0;
  xc_stat[channel].num_alloc = 0;
  xc_stat[channel].first_digi = 0;
  if ((result = route_call(channel,len,mode,call,D_EOL,0)) == 0) {
    /* somethings wrong with routing file */
#ifdef USE_IFACE  
    abort_routing(call);
#endif
    return;
  }
  switch (result) {
  case C_NOTFOUND:
  case C_DIRECT:
    xc_stat[channel].ctype = CT_END;
    xc_stat[channel].active = 1;
    statlin_update();
    xc_stat[channel].timeout = timeout;
    xc_stat[channel].starttime = time(NULL);
#ifdef USE_IFACE
    if (xc_stat[channel].inform_box = par1) {
      strcpy(xc_stat[channel].callsign,call);
      xc_stat[channel].iface = par2;
    }
#endif
    tnc_command = 1;
    strcpy(tnc_string,"C");
    strcat(tnc_string,call);
    queue_cmd_data(channel,X_COMM,strlen(tnc_string),mode,tnc_string);
    break;
  case C_ROUTE:
    expand_end = 0;
    while (!expand_end) {
      first_digi = xc_stat[channel].first_digi;
      second_digi = xc_stat[channel].digiinfo[first_digi].nextlink;
      xc_stat[channel].first_digi = xc_stat[channel].num_digi;
      if ((result = route_call(channel,len,mode,
                    xc_stat[channel].digiinfo[first_digi].nextdigi,
                    second_digi,1)) == 0) return;
      switch (result) {
      case C_NOTFOUND:
      case C_DIRECT:
        xc_stat[channel].first_digi = first_digi;
        expand_end = 1;
        break;
      case C_ROUTE:
        break;
      }
    }
    xc_stat[channel].cur_digi = xc_stat[channel].first_digi;
    xc_stat[channel].ctype = gen_con_string(channel,tnc_string,CT_CONNECT);
    xc_stat[channel].active = 1;
    statlin_update();
    xc_stat[channel].timeout = timeout;
    xc_stat[channel].starttime = time(NULL);
#ifdef USE_IFACE
    if (xc_stat[channel].inform_box = par1) {
      strcpy(xc_stat[channel].callsign,call);
      xc_stat[channel].iface = par2;
    }
#endif
    tnc_command = 1;
    queue_cmd_data(channel,X_COMM,strlen(tnc_string),mode,tnc_string);
    break;
  }
}

void next_connect(channel)
int channel;
{
  char tnc_string[256];
  int flag;

  if (xc_stat[channel].active) {
    if (xc_stat[channel].ctype == CT_END) {
      close_xconnect(channel,1);
      return;
    }
    xc_stat[channel].ctype = gen_con_string(channel,tnc_string,
                                            xc_stat[channel].ctype);
    flag = 0;
    rem_data_display(channel,tnc_string);
    queue_cmd_data(channel,X_DATA,strlen(tnc_string),flag,tnc_string);
  }
}

void connect_update(channel,buffer)
int channel;
char *buffer;
{
  switch (test_connected(channel,buffer)) {
  case T_CON:
    next_connect(channel);
    set_pwmode(channel);
    break;
  case T_RECON:
    close_xconnect(channel,0);
    break;
  default:
    break;
  }
}

int xconnect_active(channel)
int channel;
{
  return(xc_stat[channel].active > 0);
}

void init_xconnect()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    xc_stat[i].active = 0;
  }
}

void exit_xconnect()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    close_xconnect(i,0);
  }
}

void check_xconnect_timeout()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    if (xc_stat[i].active) {
      if (xc_stat[i].timeout != 0) {
        if ((time(NULL) - xc_stat[i].starttime) > xc_stat[i].timeout) {
          close_xconnect(i,0);
          queue_cmd_data(i,X_COMM,1,M_CMDSCRIPT,"D");
        }
      }
    }
  }
}

#ifdef USE_IFACE
void get_call_xc(channel,call)
int channel;
char *call;
{
  if (xc_stat[channel].active)
    strcpy(call,xc_stat[channel].callsign);
  else
    strcpy(call,"");
}
#endif

int own_connection(call1,call2)
char *call1;
char *call2;
{
  char destcall[10];
  int result;
  int i;
  
  result = 0;
  for (i=1;i<CHANNELS;i++) {
    if (ch_stat[i].state >= 4) {
      strip_call(destcall,i);
      if (((strcmp(destcall,call1) == 0) && 
           (strcmp(ch_stat[i].curcall,call2) == 0)) ||
          ((strcmp(destcall,call2) == 0) &&
           (strcmp(ch_stat[i].curcall,call1) == 0))) {
        result = 1;
        break;
      } 
    }
  }
  return(result);
}

void cmd_concall(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  if (ch_stat[channel].state < 4) {
    cmd_display(mode,channel,"Only while connected",1);
    return;
  }
  if (sscanf(str,"%30s",ch_stat[channel].disp_call) != 1) {
    cmd_display(mode,channel,"Invalid callsign",1);
    return;
  }
  strcpy(ch_stat[channel].log_call,ch_stat[channel].disp_call);
  stat_display(channel);
  set_pwmode(channel);
  cmd_display(mode,channel,ok_text,1);
}
