/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for display (display.c)
   created: Mark Wahl DL4YBG 93/08/01
   updated: Mark Wahl DL4YBG 95/01/06
*/

#include "tnt.h"
#include "xmon.h"

extern int act_channel;
extern int act_mode;
extern int xmon_screen;
extern struct channel_stat ch_stat[CHANNELS];
extern char signon[];
extern char add_signon[];
extern int is_root;
extern int chan_flag;
extern int pause_flag[CHANNELS];
extern int resync;
extern char att_normal;
extern char att_statline;
extern char att_monitor;
extern char att_cstatline;
extern char att_controlchar; 
extern char att_remote;
extern int lines_command;
extern int lines_monitor;
extern int lines_input;
extern int lines_output;
extern char rem_tnt_str[];
extern char rem_newlin_str[];
extern char tnt_dir[];
extern int infobell_flag;
extern int cbell_flag;
extern int layer3_flag;
extern int LINES,COLS;
extern int tnc_busy;
extern struct tx_file tx_file_com;
#ifdef USE_SOCKET
extern struct sock_data sd[CHANNELS];
#endif
#ifdef USE_IFACE
extern int bl_mode;
#endif

int scr_divide;
static int xmon_ch;
static int layer3_frame;
static struct window prewin[CHANNELS];
static struct window statwin[CHANNELS];
static struct window textwin[CHANNELS];
static struct window cmdwin;
static struct window monwin;
static struct window helpwin;
struct window statlin;

static struct real_layout layout[CHANNELS+2][5];
static struct real_layout help_layout[5];
extern struct real_layout xmon_layout[XMON_SCREENS][5];
extern struct real_layout boxlist_layout[CHANNELS][5];
extern struct real_layout heard_layout[5];
#ifdef USE_IFACE
extern struct real_layout mb_layout[5];
extern struct real_layout mbboxlist_layout[5];
#endif

static int old_mode;
static int help_available;
char tnt_help_file[80];

static char *ax_state[] = {
  "************",
  " LINK SETUP ",
  " FRAME REJ *",
  " DISC. LINK ",
  " INFO XFER *",
  " REJECT SND ",
  " WAIT ACK **",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *",
  " LOCAL BUSY ",
  " REM. BUSY *",
  " BOTH BUSY *"
};
  
void statlin_update()
{
  int i;
  char c;
  char tmp;
  char *tmpstr;
  int channel;
  
  tmpstr = (char *)malloc(COLS+2);
  switch (act_mode) {
  case M_CONNECT:
    strcpy(tmpstr,"\015*** CONNECT *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (pause_flag[act_channel])
      tmp = 'P';
    else if (xconnect_active(act_channel))
      tmp = 'X';
    else if (insert_active())
      tmp = 'I';
    else
      tmp = '*';
    break;
      strncat(tmpstr,&tmp,1);
  case M_COMMAND:
    strcpy(tmpstr,"\015*** COMMAND *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (insert_active())
      tmp = 'I';
    else 
      tmp = '*';
    break;
  case M_MONITOR:
    strcpy(tmpstr,"\015*** MONITOR *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (pause_flag[0])
      tmp = 'P';
    else 
      tmp = '*';
    break;  
  case M_EXTMON:
    strcpy(tmpstr,"\015*** EXTMONI *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (insert_active())
      tmp = 'I';
    else 
      tmp = '*';
    break;  
  case M_BOXLIST:
    strcpy(tmpstr,"\015*** BOXLIST *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else 
      tmp = '*';
    break;  
  case M_HEARD:
    strcpy(tmpstr,"\015**** HEARD **");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else 
      tmp = '*';
    break;  
  case M_HELP:
    strcpy(tmpstr,"\015**** HELP ***");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else 
      tmp = '*';
    break;  
#ifdef USE_IFACE
  case M_MAILBOX:
    strcpy(tmpstr,"\015*** MAILBOX *");
    if (resync)
      tmp = 'S';
    else if (tnc_busy)
      tmp = 'B';
    else if (insert_active())
      tmp = 'I';
    else 
      tmp = '*';
    break;  
#endif
  }
  strncat(tmpstr,&tmp,1);
#ifdef USE_IFACE
  if (act_mode == M_MAILBOX) {
    strcat(tmpstr,"*************");
  }
  else if ((act_mode == M_BOXLIST) && (bl_mode == M_MAILBOX)) {
    strcat(tmpstr,"* MAILBOX ***");
  }
  else {
    strcat(tmpstr,"* CHANNEL ");
    if (chan_flag) {
      tmp = '?';
    }
    else {
      if (act_mode == M_EXTMON) {
        channel = xmon_screen;
      }
      else {
        channel = act_channel;
      }
      if (channel > 9) {
        tmp = channel-10+'@';
      }
      else {
        tmp = channel+'0';
      }
    }
    strncat(tmpstr,&tmp,1);
    strcat(tmpstr," *");
  }
#else
  strcat(tmpstr,"* CHANNEL ");
  if (chan_flag) {
    tmp = '?';
  }
  else {
    if (act_mode == M_EXTMON) {
      channel = xmon_screen;
    }
    else {
      channel = act_channel;
    }
    if (channel > 9) {
      tmp = channel-10+'@';
    }
    else {
      tmp = channel+'0';
    }
  }
  strncat(tmpstr,&tmp,1);
  strcat(tmpstr," *");
#endif
  show_rxfile(act_channel,tmpstr);
  tmp = '*';
  strncat(tmpstr,&tmp,1);
  show_txfile(act_channel,tmpstr);
  strcat(tmpstr,"* ");
  for (i = 0; i < CHANNELS; i++) {
    if ((i == act_channel) && (act_mode == M_CONNECT)) {
      ch_stat[i].not_disp = 0;
    }
    if (ch_stat[i].not_disp) {
      c = '0';
      if (i > 9) c = '@' - 10;
      tmp = i+c;
    }
    else if (ch_stat[i].state) {
      tmp = '+';
    }
    else {
      tmp = SPACE;
    }
    strncat(tmpstr,&tmp,1);
  }
  strcat(tmpstr," **");
  for (i=strlen(tmpstr);i<COLS+1;i++) {
    tmpstr[i] = '*';
  }
  tmpstr[COLS+1] = '\0';
#ifdef TNT_CHAMBER
  for (i=0;i<strlen(tmpstr);i++) {
    if (tmpstr[i] == '*')
      tmpstr[i] = '-';
  }
#endif
  win_stringout(tmpstr,&statlin,0);
  free(tmpstr);
}

void newdata(channel)
int channel;
{
  ch_stat[channel].not_disp = 1;
  statlin_update();
}

void sel_monitor()
{
  if (act_mode == M_MONITOR) {
    switch (old_mode) {
    case M_COMMAND:
      real_screen(&layout[COMMAND][0]);
      break;
    case M_CONNECT:
      real_screen(&layout[act_channel][0]);
      break;
    case M_EXTMON:
      real_screen(&xmon_layout[xmon_screen][0]);
      break;
    case M_HEARD:
      real_screen(&heard_layout[0]);
      break;
    case M_HELP:
      real_screen(&help_layout[0]);
      break;
    case M_BOXLIST:
#ifdef USE_IFACE
      if (bl_mode == M_MAILBOX) {
        real_screen(&mbboxlist_layout[0]);
      }
      else {
        real_screen(&boxlist_layout[act_channel][0]);
      }
#else
      real_screen(&boxlist_layout[act_channel][0]);
#endif
      break;
#ifdef USE_IFACE
    case M_MAILBOX:
      real_screen(&mb_layout[0]);
      break;
#endif
    }
    act_mode = old_mode;
  }
  else {
    old_mode = act_mode;
    real_screen(&layout[MONITOR][0]);
    act_mode = M_MONITOR;
  }
  statlin_update();
}

void sel_command()
{
  if (act_mode == M_COMMAND) return;
  real_screen(&layout[COMMAND][0]);
  act_mode = M_COMMAND;
  statlin_update();
}

void sel_connect(channel)
int channel;
{
  if ((channel == act_channel) && (act_mode == M_CONNECT)) return;  
  real_screen(&layout[channel][0]);
  act_mode = M_CONNECT;
  act_channel = channel;
  statlin_update();
}

void stat_display(channel)
{
  int i;
  int stat_len;
  int snd_frms;
  char tmp;
  char *stat_str;
  
  stat_str = (char *)malloc(COLS+2);
  strcpy(stat_str,"");
  strcat(stat_str,"\015**");
  if (channel == 0) {
    strcat(stat_str,"************************************* ");
    strcat(stat_str,"UNPROTO CHANNEL ************************");
  }
  else {
    strcat(stat_str,ax_state[ch_stat[channel].state & 0x0f]);
    strcat(stat_str,"**");
    if (ch_stat[channel].snd_frms) {
      strcat(stat_str," S=");
      snd_frms = ch_stat[channel].snd_frms + ch_stat[channel].snd_queue_frms;
      if (snd_frms > 9) {
        tmp = snd_frms/10+'0';
        strncat(stat_str,&tmp,1);
        tmp = snd_frms%10+'0';
        strncat(stat_str,&tmp,1);
        tmp = SPACE;
        strncat(stat_str,&tmp,1);
      }
      else {
        tmp = snd_frms+'0';
        strncat(stat_str,&tmp,1);
        strcat(stat_str," *");
      }
    }
    else strcat(stat_str,"******");
    strcat(stat_str,"**");
    if (ch_stat[channel].unacked) {
      strcat(stat_str," U=");
      tmp = ch_stat[channel].unacked+'0';
      strncat(stat_str,&tmp,1);
      tmp = SPACE;
      strncat(stat_str,&tmp,1);
    }
    else strcat(stat_str,"*****");
    strcat(stat_str,"**");
    if (ch_stat[channel].tries) {
      strcat(stat_str," R=");
      if (ch_stat[channel].tries > 9) {
        tmp = ch_stat[channel].tries/10+'0';
        strncat(stat_str,&tmp,1);
        tmp = ch_stat[channel].tries%10+'0';
        strncat(stat_str,&tmp,1);
        tmp = SPACE;
        strncat(stat_str,&tmp,1);
      }
      else {
        tmp = ch_stat[channel].tries+'0';
        strncat(stat_str,&tmp,1);
        strcat(stat_str," *");
      }
    }
    else strcat(stat_str,"******");
    strcat(stat_str,"** ");

    if (ch_stat[channel].disp_call[0] == '\0')   
      strncat(stat_str,ch_stat[channel].call,40);
    else
      strncat(stat_str,ch_stat[channel].disp_call,40);
    if ((stat_len = strlen(stat_str)) < (COLS + 1)) {
      strncat(stat_str,
              " ***************************************"
              "****************************************"
              "****************************************",
              COLS + 1 - stat_len);
    }
  }
#ifdef TNT_CHAMBER
  for (i=0;i<strlen(stat_str);i++) {
    if (stat_str[i] == '*')
      stat_str[i] = '-';
  }
#endif
  win_stringout(stat_str,&statwin[channel],0);
  free(stat_str);
}

void data_display(channel,buffer)
int channel;
char *buffer;
{
  char tmpstring[280];
  
  strcpy(tmpstring,"\r*** ");
  strcat(tmpstring,buffer);
  strcat(tmpstring,"\r");
#ifdef USE_SOCKET
  sock_status_out(channel,tmpstring);
  if (output_screen_sock(channel)) {
    if (cbell_flag) beep();
    win_stringout(tmpstring,&textwin[channel],1);
    newdata(channel);
  }
#else
  if (cbell_flag) beep();
  win_stringout(tmpstring,&textwin[channel],1);
  newdata(channel);
#endif
  write_file(channel,strlen(tmpstring),RX_RCV,tmpstring);
}

void rem_data_display(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_stringout(buffer,&textwin[channel],1);
  win_attrib(att_normal,&textwin[channel]);
}

void rem_data_display_buf(channel,buffer,len)
int channel;
char *buffer;
int len;
{
  win_attrib(att_remote,&textwin[channel]);
  win_textout_len(buffer,len,&textwin[channel],1);
  win_attrib(att_normal,&textwin[channel]);
}

void rem_stat_display(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_stringout(buffer,&textwin[channel],1);
  win_charout_cntl(C_STLINE,&textwin[channel]);
  win_attrib(att_normal,&textwin[channel]);
}

void data_display_len(channel,buffer)
int channel;
char *buffer;
{
#ifdef USE_IFACE
  write_iface(channel,(*buffer)+1,buffer+1);
#endif
  write_pty(channel,(*buffer)+1,buffer+1);
  if (output_screen(channel)) {
    if (infobell_flag) beep();
    newdata(channel);
  }
  if (!act_abin_rx(channel)) {
    if (!check_spec_strings(channel,buffer)) {
      if (output_screen(channel)) {
        insert_cr_tx(channel);
#ifdef USE_IFACE
        boxcut(channel,buffer);
#endif        
        win_textout(buffer,&textwin[channel],1);
      }
    }
    else return;
    connect_update(channel,buffer);
  }
  scan_login_received(channel,buffer);
  if (!remote_input(channel,buffer))
    write_file(channel,(*buffer)+1,RX_RCV,buffer+1);
  /* set lastcr if last character of packet is a CR */
  ch_stat[channel].lastcr = (buffer[(*buffer)+1] == '\015') ? 1 : 0;
}

void data_display_len2(channel,buffer)
int channel;
char *buffer;
{
  if (output_screen(channel)) {
#ifdef USE_IFACE
    boxcut(channel,buffer);
#endif
    win_textout(buffer,&textwin[channel],1);
  }
}

void rem_data_display_len(channel,buffer)
int channel;
char *buffer;
{
  win_attrib(att_remote,&textwin[channel]);
  win_textout(buffer,&textwin[channel],1);
  win_attrib(att_normal,&textwin[channel]);
}

void cmd_display(flag,channel,buffer,cr)
int flag;
int channel;
char *buffer;
int cr;
{
  int ans_flag;
  int ans_len;
  char ans_str[256];
  int con_channel;
  
  switch (flag) {
  case M_CMDSCRIPT:
    return;
  case M_CONNECT:
    if (tx_file_com.type == -1) {
      con_channel = channel;
    }
    else {
      con_channel = act_channel;
    } 
    win_stringout(buffer,&prewin[con_channel],0);
    if (cr) win_charout(CR,&prewin[con_channel],0);
    break;
#ifdef USE_SOCKET
  case M_SOCKET:
    if(sd[channel].login_stat>=2) {
      strcpy(ans_str,rem_tnt_str);
      strcat(ans_str,buffer);
      strcat(ans_str,rem_newlin_str);
      if(sd[channel].login_stat==2)
        strcat(ans_str,"\nCmd> ");
      ans_len = strlen(ans_str);
      write_pty(channel,ans_len,ans_str);
    }
    break;
#endif
  case M_REMOTE_TEMP:
  case M_REMOTE:
    strcpy(ans_str,rem_tnt_str);
    strcat(ans_str,buffer);
    strcat(ans_str,rem_newlin_str);
    ans_len = strlen(ans_str);
    ans_flag = 0;
    rem_data_display(channel,ans_str);
    queue_cmd_data(channel,X_DATA,ans_len,ans_flag,ans_str);
    break;
  case M_EXTMON:
    xm_pre_display(channel,buffer,cr);
    break;
#ifdef USE_IFACE
  case M_MAILBOX:
    mb_pre_display(buffer,cr);
    break;
#endif 
  default:
    win_stringout(buffer,&cmdwin,0);
    if (cr) win_charout(CR,&cmdwin,0);
    break;
  }
}
  
void moni_display2(buffer,attrib)
char *buffer;
int attrib;
{
  if (attrib)
    win_attrib(att_monitor,&monwin);
  win_stringout(buffer,&monwin,0);
  if (attrib)
    win_attrib(att_normal,&monwin);
  win_charout(CR,&monwin,0);
  write_monfile(strlen(buffer),buffer,1);
}

void moni_display(channel,buffer)
int channel;
char *buffer;
{
  layer3_frame = 0;
  if (layer3_flag) {
    layer3_frame = (strstr(buffer,"pid CF") != NULL);
  }
  xmon_ch = analyse_mon_header(buffer);
  moni_display2(buffer,1);
}

void moni_display_len(channel,buffer)
int channel;
char *buffer;
{
  int append_cr;

  if (xmon_ch != -1) {
    write_mon_info(buffer,xmon_ch);
  }
  else {
    analyse_mon_body(buffer);
  }
  if (layer3_frame) {
    if (!display_layer3(buffer)) return;
  }
  win_textout(buffer,&monwin,0);
  append_cr = 0;
  if (*(buffer + *buffer + 1) != CR) {
    win_charout(CR,&monwin,0);
    append_cr = 1;
  }
  write_monfile((*buffer + 1),buffer+1,append_cr);
}

void moni_display_len2(buffer,len)
char *buffer;
{
  int append_cr;
  char *buffer2;

  buffer2 = (char *)malloc(len+1);
  *buffer2 = (char)(len - 1);
  memcpy(buffer2 + 1,buffer,len);
  win_textout(buffer2,&monwin,0);
  append_cr = 0;
  if (*(buffer + len - 1) != CR) {
    win_charout(CR,&monwin,0);
    append_cr = 1;
  }
  write_monfile(len,buffer,append_cr);
}

int pre_newline(channel,str)
int channel;
char *str;
{
  int len;
  
  len = get_line(str,&prewin[channel]);
  win_charout(CR,&prewin[channel],0);
  return(len);
}

int pre_nextline(channel)
int channel;
{
  win_charout(CR,&prewin[channel],0);
}

void pre_charout(channel,ch)
int channel;
char ch;
{
  if (prewin[channel].column < COLS-1) {
    win_charout_cntl(ch,&prewin[channel]);
  }
  else {
    beep();
  }
}

void pre_charout_nobnd(channel,ch)
int channel;
char ch;
{
  win_charout_cntl(ch,&prewin[channel]);
}

void pre_charout_cntl(channel,ch)
int channel;
char ch;
{
  if (prewin[channel].column < COLS-1) {
    ch &= 0x1f;
    ch |= 0x40;
    win_attrib(att_controlchar,&prewin[channel]);
    win_charout_cntl(ch,&prewin[channel]);
    win_attrib(att_normal,&prewin[channel]);
  }
  else {
    beep();
  }
}

int cmd_newline(channel,str)
int channel;
char *str;
{
  int len;
  char tmpstr[MAXCOLS+1];
  
  len = get_line(tmpstr,&cmdwin);
  win_charout(CR,&cmdwin,0);
  if (tmpstr[0] == ':') {
    strcpy(str,tmpstr+1);
    len--;
  }
  else
    strcpy(str,tmpstr);
  return(len);
}

void cmd_charout(channel,ch)
int channel;
char ch;
{
  if (cmdwin.column < COLS-1) {
    win_charout_cntl(ch,&cmdwin,0);
  }
  else {
    beep();
  }
}

void cmd_charout_nobnd(channel,ch)
int channel;
char ch;
{
  win_charout_cntl(ch,&cmdwin);
}

int init_screen()
{
  int i;
  int j;

  infobell_flag = 0;
  cbell_flag = 1;
  layer3_flag = 1;  
  for (i = 0; i <= MONITOR; i++) {
    for (j = 0; j < 5; j++) {
      layout[i][j].win = NULL;
      layout[i][j].first_real_line = 0;
      layout[i][j].win_num_lines = 0;
      layout[i][j].pagesize = 0;
    }
  }
  window_init();
  open_window(1,&statlin,1);
  open_window(lines_command,&cmdwin,1);
  layout[COMMAND][0].win = &cmdwin;
  layout[COMMAND][0].first_real_line = 0;
  layout[COMMAND][0].win_num_lines = LINES-1;
  layout[COMMAND][0].pagesize = LINES-4;
  layout[COMMAND][1].win = &statlin;
  layout[COMMAND][1].first_real_line = LINES-1;
  layout[COMMAND][1].win_num_lines = 1;
  layout[COMMAND][1].pagesize = 0;
  open_window(lines_monitor,&monwin,1);
  layout[MONITOR][0].win = &monwin;
  layout[MONITOR][0].first_real_line = 0;
  layout[MONITOR][0].win_num_lines = LINES-1;
  layout[MONITOR][0].pagesize = LINES-4;
  layout[MONITOR][1].win = &statlin;
  layout[MONITOR][1].first_real_line = LINES-1;
  layout[MONITOR][1].win_num_lines = 1;
  layout[MONITOR][1].pagesize = 0;
  for (i = 0; i < CHANNELS; i++) {
    open_window(lines_input,&prewin[i],1);
    open_window(1,&statwin[i],1);
    win_attrib(att_cstatline,&statwin[i]);
    open_window(lines_output,&textwin[i],1);
    layout[i][0].win = &prewin[i];
    layout[i][0].first_real_line = 0;
    layout[i][0].win_num_lines = LINES/scr_divide-1;
    layout[i][0].pagesize = LINES/scr_divide-2;
    layout[i][1].win = &statwin[i];
    layout[i][1].first_real_line = LINES/scr_divide-1;
    layout[i][1].win_num_lines = 1;
    layout[i][1].pagesize = 0;
    layout[i][2].win = &textwin[i];
    layout[i][2].first_real_line = LINES/scr_divide;
    layout[i][2].win_num_lines = LINES-LINES/scr_divide-1;
    layout[i][2].pagesize = LINES-LINES/scr_divide-4;
    layout[i][3].win = &statlin;
    layout[i][3].first_real_line = LINES-1;
    layout[i][3].win_num_lines = 1;
    layout[i][3].pagesize = 0;
  }
  real_screen(&layout[COMMAND][0]);
  act_mode = M_COMMAND;
  old_mode = M_COMMAND;
  act_channel = 1;
  xmon_ch = -1;
  layer3_frame = 0;
  win_attrib(att_statline,&statlin);
  for (i = 0; i < CHANNELS;i++) {
    ch_stat[i].not_disp = 0;
  }
  statlin_update();
  if (!is_root) {
    cmd_display(M_COMMAND,1,signon,0);
    cmd_display(M_COMMAND,1,add_signon,1);
  }
  else {
    cmd_display(M_COMMAND,1,signon,1);
  }
  return(0);
}

void exit_screen()
{
  int i;
  
  close_window(&cmdwin);
  close_window(&monwin);
  close_window(&statlin);
  for (i = 0; i < CHANNELS; i++) {
    close_window(&prewin[i]);
    close_window(&statwin[i]);
    close_window(&textwin[i]);
  }
  window_exit();
}

/* display line on help-screen */
static void help_textout_len(buf,len)
char *buf;
int len;
{
  win_textout_len(buf,len,&helpwin,1);
}

/* character to help-screen */
void help_charout_cntl(ch)
char ch;
{
  win_charout_cntl(ch,&helpwin);
}

/* select help screen */
void sel_help()
{
  if ((!help_available) || (act_mode == M_HELP)) return;
  real_screen(&help_layout[0]);
  act_mode = M_HELP;
  statlin_update();
}

#define BUFSIZE 512

/* init of help-screen */
void init_help()
{
  char helpfile[160];
  int helpfd;
  int file_len;
  int end;
  int len2;
  char buf[BUFSIZE];
  int i;
  int j;
  int lines;
  char *linebuf;
  int linelen;

  
  help_available = 0;
  strcpy(helpfile,tnt_dir);
  strcat(helpfile,tnt_help_file);
  helpfd = open(helpfile,O_RDONLY); 
  if (helpfd == -1) {
    /* file does not exist */
    return;
  }
  
  /* counting number of LFs in file to get number of lines */
  file_len = 0;
  lines = 0;
  end = 1;
  while (end) {
    len2 = read(helpfd,buf,BUFSIZE);
    file_len += len2;
    for (i=0;i<len2;i++) {
      if (buf[i] == '\n') lines++;
    }
    if (len2 < BUFSIZE) end = 0;
  }
  lines += 1; /* just to be sure... */

  for (j = 0; j < 5; j++) {
    help_layout[j].win = NULL;
    help_layout[j].first_real_line = 0;
    help_layout[j].win_num_lines = 0;
    help_layout[j].pagesize = 0;
  }
  open_window(lines,&helpwin,0);
  help_layout[0].win = &helpwin;
  help_layout[0].first_real_line = 0;
  help_layout[0].win_num_lines = LINES-1;
  help_layout[0].pagesize = LINES-4;
  help_layout[1].win = &statlin;
  help_layout[1].first_real_line = LINES-1;
  help_layout[1].win_num_lines = 1;
  help_layout[1].pagesize = 0;
  
  /* seek to file begin */
  lseek(helpfd,0,SEEK_SET);
  linelen = 0;
  end = 1;
  linebuf = (char *)malloc(COLS+2);
  while (end) {
    len2 = read(helpfd,buf,BUFSIZE);
    for (i=0;i<len2;i++) {
      if (buf[i] == '\n') {
        linebuf[linelen] = '\r';
        linelen++;
        help_textout_len(linebuf,linelen);
        linelen = 0;
      }
      else {
        if (linelen < (COLS-1)) {
          linebuf[linelen] = buf[i];
          linelen++;
        }
      }
    }
    if (len2 < BUFSIZE) end = 0;
  }
  if (linelen) {
    linebuf[linelen] = '\r';
    linelen++;
    help_textout_len(linebuf,linelen);
    linelen = 0;
  }
  close(helpfd);
  free(linebuf);
  helpwin.line = 0;
  if ((helpwin.phys_line = helpwin.line + helpwin.line_offset) >=
       helpwin.num_lines)
    helpwin.phys_line -= helpwin.num_lines;
  helpwin.column = 0;
  help_available = 1;
}

/* exit of help-screen */
void exit_help()
{
  if (help_available) {
    close_window(&helpwin);
  }
}
