/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993,1994 by Mark Wahl
   For license details see documentation
   Procedures for interface (iface.c)
   created: Mark Wahl DL4YBG 94/07/03
   updated: Mark Wahl DL4YBG 94/12/20
*/

#undef IFACE_DEBUG

#include "tnt.h"
#ifdef USE_IFACE
#include "ifacedef.h"
#include "iface.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>

extern int act_mode;
extern int att_controlchar;
extern int att_normal;
extern int att_remote;
extern int att_cstatline;
extern int lines_mbinput;
extern int lines_mboutput;
extern int LINES,COLS;
extern int mbscr_divide;
extern int ptyecho_flag;
static char iface_err_txt[80];
static char iface_open_txt[80];
static char sock_err_txt[80];
static char conn_err_txt[80];
static char end_err_txt[80];
static char back_tnt_txt[80];
char iface_active_txt[80];
char box_socket[80];
extern char sh_active_txt[];
static struct iface_list if_list[MAX_IFACE];
static struct iface_stat if_stat[CHANNELS];
extern struct channel_stat ch_stat[CHANNELS];
int box_active_flag;

struct real_layout mb_layout[5];
static struct window mb_prewin;
static struct window mb_statwin;
static struct window mb_textwin;
extern struct window statlin;

static void deactivate_program();
void send_command_packet_if();
void mb_display_buf();
void mb_stat_display();

/* close iface connection on channel */
void close_iface_con(channel,deact)
int channel;
int deact;
{
  if (if_stat[channel].iface != -1) {
    if (deact) deactivate_program(channel);
    if_stat[channel].iface = -1;
    if (channel == 0)
      mb_stat_display();
  }
}

/* close the iface-socket */
void close_iface(iface,deact)
int iface;
int deact;
{
  int i;
  struct queue_entry *oldq_ptr;
  
  if (if_list[iface].active) {
    for (i=0;i<CHANNELS;i++) {
      if (if_stat[i].iface == iface) {
        close_iface_con(i,deact);
      }
    }
    if (strcmp(if_list[iface].name,box_socket) == 0) {
      monbox_close();
      box_active_flag = 0;
    }
    while (if_list[iface].first_q != NULL) {
      oldq_ptr = if_list[iface].first_q;
      free(oldq_ptr->buffer);
      if_list[iface].first_q = oldq_ptr->next;
      free((char *)oldq_ptr);
    }
    close(if_list[iface].sockfd);
    if_list[iface].active = 0;
  }
}

static void blocking_test(iface,len)
int iface;
int len;
{
  IFACE_CMDBUF command;
  
  if_list[iface].sendlength += len;
  if (if_list[iface].sendlength > BLOCKING_SIZE) {
    if_list[iface].blocked = 1;
    command.command = CMD_BLOCK;
    send_command_packet_if(NO_CHANNEL,NO_USERNR,LEN_SIMPLE,
                           (char *)&command,iface);
  }
}

static void unblocking(iface)
int iface;
{
  struct queue_entry *oldq_ptr;
  
  if_list[iface].blocked = 0;
  if_list[iface].sendlength = 0;
  while ((!if_list[iface].blocked) &&
         (if_list[iface].first_q != NULL)) {
    oldq_ptr = if_list[iface].first_q;
    if (write(if_list[iface].sockfd,oldq_ptr->buffer,
              oldq_ptr->len) < oldq_ptr->len) {
      close_iface(iface,0);
      return;
    }
    blocking_test(iface,oldq_ptr->len);
    free(oldq_ptr->buffer);
    if_list[iface].first_q = oldq_ptr->next;
    free((char *)oldq_ptr);
  }
}

/* analysis of received packet via interface */
static void packet_analysis(indicator,channel,usernr,len,buf,iface)
char indicator;
int channel;
int usernr;
int len;
char *buf;
int iface;
{
  int con_channel;
  char p1str[256];
  char p2str[256];
  char hlpstr[256];
  char *bufptr;
  IFACE_CMDBUF *rec_command;
  IFACE_CMDBUF command;
  
  if ((channel != NO_CHANNEL) &&
      (channel < CHANNELS) && 
      (if_stat[channel].iface == -1)) return;
  switch (indicator) {
  case IF_COMMAND:
    rec_command = (IFACE_CMDBUF *)buf;
    switch (rec_command->command) {
    case CMD_DISCONNECT:
      if (channel != 0)
        queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
      close_iface_con(channel,0);
      break;
    case CMD_FINISH:
      if (channel != 0) {
        if (if_stat[channel].direct_started) {
          queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
          close_iface_con(channel,0);
        }
        else {
          close_iface_con(channel,0);
          cmd_display(M_REMOTE,channel,back_tnt_txt,1);
        }
      }
      else {
        close_iface_con(channel,0);
      }
      break;
    case CMD_BLOCK:
      command.command = CMD_UNBLOCK;
      send_command_packet_if(NO_CHANNEL,NO_USERNR,LEN_SIMPLE,(char *)&command,iface);
      break;
    case CMD_UNBLOCK:
      unblocking(iface);
      break;
    case CMD_BOXPBOXSF:
      bufptr = rec_command->data.boxpboxsf.buffer;
      strcpy(p1str,bufptr);
      bufptr += strlen(p1str) + 1;
      strcpy(p2str,bufptr);
      
      con_channel = find_free_channel();
      if (con_channel == -1) {
        abort_routing(p2str);
      }
      else {
        strcpy(hlpstr,"I");
        strcat(hlpstr,p1str);
        queue_cmd_data(con_channel,X_COMM,strlen(hlpstr),M_CMDSCRIPT,hlpstr);
        sprintf(hlpstr,"%s %d",p2str,rec_command->data.boxpboxsf.timeout);
        cmd_xconnect(1,iface,
                     con_channel,strlen(hlpstr),M_CONNECT,hlpstr);
      }
      break;
    case CMD_BOXPRDISC:
      if (channel != 0)
        queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
      close_iface_con(channel,1);
      break;
    case CMD_BOXPABORTSF:
      queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
      close_iface_con(channel,0);
      if (rec_command->data.immediate) {
        queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
      }
      break;
    case CMD_ACT_RESP:
      if ((len-LEN_SIMPLE) > 0) {
        if (channel == 0) {
          mb_display_buf(rec_command->data.buffer,len-LEN_SIMPLE);
        }
        else {
          if (ptyecho_flag)
            rem_data_display_buf(channel,rec_command->data.buffer,len-LEN_SIMPLE);
          dat_input(channel,rec_command->data.buffer,len-LEN_SIMPLE);
        }
      }
      if (usernr == 0) {
        if (if_stat[channel].direct_started) {
          queue_cmd_data(channel,X_COMM,1,M_CMDSCRIPT,"D");
          close_iface_con(channel,0);
        }
        else {
          close_iface_con(channel,0);
          cmd_display(M_REMOTE,channel,back_tnt_txt,1);
        }
      }
      else {
        if_stat[channel].usernr = usernr;
      }
      break;
    default:
      break;
    }
    break;
  case IF_DATA:
    if (channel == 0) {
      mb_display_buf(buf,len);
    }
    else {
      if (ptyecho_flag)
        rem_data_display_buf(channel,buf,len);
      dat_input(channel,buf,len);
    }
    break;
  default:
    break;
  }
}

void send_command_packet_if(channel,usernr,len,buf,iface)
short channel;
short usernr;
unsigned short len;
char *buf;
int iface;
{
  char buffer[MAX_LEN];
  int res;
  IFACE_HEADER header;

  if ((iface > MAX_IFACE) || (len > IFACE_PACLEN)) return;
  header.indicator = IF_COMMAND;
  header.channel = channel;
  header.usernr = usernr;
  header.len = len;
  memcpy(buffer,(char *)&header,HEAD_LEN);
  memcpy(buffer + HEAD_LEN,buf,len);
  if ((res = write(if_list[iface].sockfd,buffer,len+HEAD_LEN)) < len+HEAD_LEN) {
    close_iface(iface,0);
    return;
  }
}

static void send_command_packet(channel,usernr,len,buf)
short channel;
short usernr;
unsigned short len;
char *buf;
{
  char buffer[MAX_LEN];
  int iface;
  int res;
  IFACE_HEADER header;

  if (((iface = if_stat[channel].iface) == -1) || (len > IFACE_PACLEN)) return;
  header.indicator = IF_COMMAND;
  header.channel = channel;
  header.usernr = usernr;
  header.len = len;
  memcpy(buffer,(char *)&header,HEAD_LEN);
  memcpy(buffer + HEAD_LEN,buf,len);
  if ((res = write(if_list[iface].sockfd,buffer,len+HEAD_LEN)) < len+HEAD_LEN) {
    close_iface(iface,0);
    return;
  }
}

/* channel status is sent via interface */
void send_cstatus(channel)
int channel;
{
  IFACE_CMDBUF command;
  int len;

  if ((if_stat[channel].iface == -1) || (channel == 0)) return;
  command.command = CMD_CSTATUS;
  command.data.snd_frames = ch_stat[channel].snd_frms + 
                            ch_stat[channel].snd_queue_frms;
  send_command_packet((short)channel,if_stat[channel].usernr,
                      LEN_CSTATUS,(char *)&command);
}

static void activate_program(channel)
int channel;
{
  int len;
  char callsign[10];
  char *ptr;
  IFACE_CMDBUF command;

  command.command = CMD_ACTIVATE;
  len = LEN_SIMPLE;
  if (channel != 0) {
    /* fill callsign */
    strip_call_log(callsign,channel);
    ptr = strchr(callsign,'-');
    if (ptr != NULL) *ptr = '\0';
    strcpy(command.data.buffer,callsign);
    len += strlen(callsign);
    *(command.data.buffer + len - LEN_SIMPLE) = '\0';
    len++;
  }
  send_command_packet((short)channel,NO_USERNR,len,(char *)&command);

  send_cstatus(channel);
}

static void deactivate_program(channel)
int channel;
{
  IFACE_CMDBUF command;
  
  command.command = CMD_DEACTIVATE;
  send_command_packet((short)channel,if_stat[channel].usernr,
                      LEN_SIMPLE,(char *)&command);
}

/* write data to interface */
int write_iface(channel,len,str)
int channel;
int len;
char *str;
{
  int iface;
  char buffer[MAX_LEN];
  int res;
  char *ptr;
  struct queue_entry *q_ptr;
  IFACE_HEADER header;
  
  if (((iface = if_stat[channel].iface) == -1) || 
      (len > IFACE_PACLEN)) return(0);
  header.indicator = IF_DATA;
  header.channel = (short) channel;
  if (channel == NO_CHANNEL)
    header.usernr = NO_USERNR;
  else
    header.usernr = if_stat[channel].usernr;
  header.len = len;
  memcpy(buffer,(char *)&header,HEAD_LEN);
  memcpy(&buffer[HEAD_LEN],str,len);
  if (if_list[iface].blocked) {
    q_ptr = (struct queue_entry *)malloc(sizeof(struct queue_entry));
    if (q_ptr == NULL) {
      close_iface(iface,0);
      return;
    }
    ptr = (char *)malloc(len+HEAD_LEN);
    if (ptr == NULL) {
      free(q_ptr);
      close_iface(iface,0);
      return;
    }
    memcpy(ptr,buffer,len+HEAD_LEN);
    q_ptr->buffer = ptr;
    q_ptr->len = len+HEAD_LEN;
    q_ptr->next = NULL;
    if (if_list[iface].first_q == NULL) {
      if_list[iface].first_q = q_ptr;
    }
    else {
      if_list[iface].last_q->next = q_ptr;
    }
    if_list[iface].last_q = q_ptr;
  }
  else {
    if ((res = write(if_list[iface].sockfd,
                   buffer,len+HEAD_LEN)) < len+HEAD_LEN) {
      close_iface(iface,0);
      return;
    }
    blocking_test(iface,len+HEAD_LEN);
  }
}

/* read data from interface */
static int read_data_iface(iface)
int iface;
{
  char buffer[MAX_LEN];
  char *bufptr;
  int max;
  int len;
  int remain;
#ifdef IFACE_DEBUG
  unsigned short tmplen;
#endif
  

  if (!if_list[iface].active) return(0);  
  len = read(if_list[iface].sockfd,buffer,MAX_LEN);
  if (len == -1) {
    if (errno != EAGAIN) {
      close_iface(iface);
    }
    return(0);
  }
  if (len == 0) {
    close_iface(iface);
  }
  bufptr = buffer;
  remain = len;
  while (remain) {
    if (if_list[iface].status < HEAD_LEN) {
      *(((char *)&if_list[iface].header) + if_list[iface].status) = *bufptr;
      if_list[iface].status++;
      remain--;
      bufptr++;
    }
    else {
      /* packet data */
      if (if_list[iface].header.len > 0) {
        if_list[iface].buffer[if_list[iface].status - HEAD_LEN] = *bufptr;
        if_list[iface].status++;
      }
      if ((if_list[iface].status - HEAD_LEN) == if_list[iface].header.len) {
        if_list[iface].status = 0;
#ifdef IFACE_DEBUG
        tmplen = 0;
        fprintf(stderr,"IF: %d, ind: %d, ch: %d, us: %d, len: %d\n",
                iface,
                if_list[iface].header.indicator,
                if_list[iface].header.channel,
                if_list[iface].header.usernr,
                if_list[iface].header.len);
        while (tmplen < if_list[iface].header.len) {
          fprintf(stderr,"%x ",if_list[iface].buffer[tmplen]);
          tmplen++;
        }
        fprintf(stderr,"\n");
#endif
        packet_analysis(if_list[iface].header.indicator,
                        if_list[iface].header.channel,
                        if_list[iface].header.usernr,
                        if_list[iface].header.len,
                        if_list[iface].buffer,
                        iface);
      }
      remain--;
      bufptr++;
    }
  }
  return(1);
}

/* install link to other program */
void cmd_iface(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int sockfd;
  int servlen;
  struct sockaddr_un serv_addr;
  int iface;

  iface = 0;
  while (iface < MAX_IFACE) {
    if (!if_list[iface].active) {
      break;
    }
    if (strcmp(if_list[iface].name,str) == 0) {
      cmd_display(mode,channel,iface_open_txt,1);
      return;
    }
    iface++;
    if (iface == MAX_IFACE) {
      cmd_display(mode,channel,iface_err_txt,1);
      return;
    }
  }  
  /* fill structure "serv_addr" */
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sun_family = AF_UNIX;
  strcpy(serv_addr.sun_path,str);
  servlen = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
  /* open socket */
  if ((sockfd = socket(AF_UNIX,SOCK_STREAM,0)) < 0) {
    cmd_display(mode,channel,sock_err_txt,1);
    return;
  }
#ifndef USE_SELECT
  fcntl(sockfd,F_SETFL,O_NONBLOCK);
#endif
  /* connect other program */
  if (connect(sockfd, (struct sockaddr *) &serv_addr, servlen) < 0) {
    cmd_display(mode,channel,conn_err_txt,1);
    return;
  }
  if_list[iface].active = IF_ACTIVE;
  if_list[iface].sockfd = sockfd;
  if_list[iface].status = 0;
  if_list[iface].blocked = 0;
  if_list[iface].sendlength = 0;
  strcpy(if_list[iface].name,str);
  if (strcmp(if_list[iface].name,box_socket) == 0) {
    box_active_flag = 1;
  }
  cmd_display(mode,channel,"Ok",1);
}

/* cut link to other program */
void cmd_endiface(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int iface;
  
  iface = 0;
  while (iface < MAX_IFACE) {
    if (if_list[iface].active) {
      if (strcmp(if_list[iface].name,str) == 0) {
        close_iface(iface,1);
        cmd_display(mode,channel,"Ok",1);
        return;
      }
    }
    iface++;
    if (iface == MAX_IFACE) {
      cmd_display(mode,channel,end_err_txt,1);
      return;
    }
  }
}

/* finish execution of other program */
void cmd_finiface(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int iface;
  char buffer[10];
  int blen;
  IFACE_CMDBUF command;
  
  iface = 0;
  while (iface < MAX_IFACE) {
    if (if_list[iface].active) {
      if (strcmp(if_list[iface].name,str) == 0) {
        command.command = CMD_FINISH_PRG;
        send_command_packet_if(NO_CHANNEL,NO_USERNR,LEN_SIMPLE,(char *)&command,iface);
        close_iface(iface,0);
        cmd_display(mode,channel,"Ok",1);
        return;
      }
    }
    iface++;
    if (iface == MAX_IFACE) {
      cmd_display(mode,channel,end_err_txt,1);
      return;
    }
  }
}

/* activate link to other program */
void cmd_actiface(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int iface;
  
  if (mode == M_MAILBOX) {
    channel = 0;
  }
  else {
    if (shell_active(channel)) {
      cmd_display(mode,channel,sh_active_txt,1);
      return;
    }
  }
  
  if (if_stat[channel].iface != -1) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
  iface = 0;
  while (iface < MAX_IFACE) {
    if (if_list[iface].active) {
      if (strcmp(if_list[iface].name,str) == 0) {
        if_stat[channel].iface = iface;
        if_stat[channel].usernr = 0;
        if_stat[channel].direct_started = par1;
        if (par2 && (channel != 0)) {
          start_sf(channel,iface);
          cmd_display(mode,channel,"Starting S&F",1);
        }
        else {
          activate_program(channel);
          cmd_display(mode,channel,"Activating program",1);
        }
        if (channel == 0)
          mb_stat_display();
        return;
      }
    }
    iface++;
    if (iface == MAX_IFACE) {
      cmd_display(mode,channel,end_err_txt,1);
      return;
    }
  }
}

/* deactivate link to other program */
void cmd_deactiface(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int iface;

  if (mode == M_MAILBOX) channel = 0;  
  iface = 0;
  while (iface < MAX_IFACE) {
    if (if_list[iface].active) {
      if (strcmp(if_list[iface].name,str) == 0) {
        close_iface_con(channel,1);
        cmd_display(mode,channel,"Ok",1);
        return;
      }
    }
    iface++;
    if (iface == MAX_IFACE) {
      cmd_display(mode,channel,end_err_txt,1);
      return;
    }
  }
}

#ifdef USE_SELECT
void iface_fdset(max_fd,fdmask)
int *max_fd;
struct fd_set *fdmask;
{
  int iface;
  int fd;
  
  for (iface=0;iface<MAX_IFACE;iface++) {
    if (if_list[iface].active) {
      fd = if_list[iface].sockfd;
      FD_SET(fd,fdmask);
      if (fd > ((*max_fd) - 1)) {
        *max_fd = fd + 1;
      }
    }
  }
}

int iface_receive(fdmask)
struct fd_set *fdmask;
{
  int iface;
  int result;
  
  result = 1;
  for (iface=0;iface<MAX_IFACE;iface++) {
    if (if_list[iface].active) {
      fd = if_list[iface].sockfd;
      if (FD_ISSET(fd,fdmask)) {
        read_data_iface(iface);
        result = 0;
      }
    }
  }
  return(result);
}

#else
int iface_receive()
{
  int iface;
  int result;
  
  result = 1;
  for (iface=0;iface<MAX_IFACE;iface++) {
    if (if_list[iface].active) {
      if (read_data_iface(iface)) {
        result = 0;
      }
    }
  }
  return(result);
}
#endif

/* returns if iface active on channel */
int iface_active(channel)
int channel;
{
  return(if_stat[channel].iface != -1);
}

/* init of iface-variables */
void init_iface()
{
  int i;
  
  for (i=0;i<MAX_IFACE;i++) {
    if_list[i].active = 0;
  }
  for (i=0;i<CHANNELS;i++) {
    if_stat[i].iface = -1;
  }
  strcpy(iface_err_txt,"All interfaces in use");
  strcpy(iface_open_txt,"Interface already open");
  strcpy(sock_err_txt,"Can't open stream socket");
  strcpy(conn_err_txt,"Can't connect to program");
  strcpy(end_err_txt,"Interface not active");
  strcpy(iface_active_txt,"Interface already active");
  strcpy(back_tnt_txt,"Back at TNT");

  box_active_flag = 0;  
  init_monbox();
  init_boxcut();
}

/* exit of iface, close all sockets */
void exit_iface()
{
  int i;

  exit_boxcut();
  exit_monbox();
    
  for (i=0;i<MAX_IFACE;i++) {
    close_iface(i,1);
  }
}

/* activate box */
void cmd_actbox(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_iface(par1,par2,channel,strlen(box_socket),mode,box_socket);
}

/* deactivate box */
void cmd_deactbox(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_endiface(par1,par2,channel,strlen(box_socket),mode,box_socket);
}

/* finish box program */
void cmd_finbox(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_finiface(par1,par2,channel,strlen(box_socket),mode,box_socket);
}

/* start box */
void cmd_box(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_actiface(par1,par2,channel,strlen(box_socket),mode,box_socket);
}

/* end box */
void cmd_endbox(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_deactiface(par1,par2,channel,strlen(box_socket),mode,box_socket);
}

/* looking for iface of name str, if found, return iface-nr, else -1 */
int conv_name_to_iface(str)
char *str;
{
  int iface;
  iface = 0;
  while (iface < MAX_IFACE) {
    if (if_list[iface].active) {
      if (strcmp(if_list[iface].name,str) == 0) {
        return(iface);
      }
    }
    iface++;
  }
  return(-1);
}

/* file string with socket name of iface */
void get_iface_name(iface,str)
int iface;
char *str;
{
  if (if_list[iface].active) {
    strcpy(str,if_list[iface].name);
  }
  else {
    str[0] = '\0';
  }
}

/* procedures for mailbox-user-frontend */

void sel_mailbox()
{
  if (act_mode == M_MAILBOX) return;
  real_screen(&mb_layout[0]);
  act_mode = M_MAILBOX;
  statlin_update();
}

int mb_newline(str)
char *str;
{
  int len;
  
  len = get_line(str,&mb_prewin);
  win_charout(CR,&mb_prewin,0);
  return(len);
}

int mb_nextline()
{
  win_charout(CR,&mb_prewin,0);
}

void mb_charout(ch)
char ch;
{
  if (mb_prewin.column < COLS-1) {
    win_charout_cntl(ch,&mb_prewin);
  }
  else {
    beep();
  }
}

void mb_charout_nobnd(ch)
char ch;
{
  win_charout_cntl(ch,&mb_prewin);
}

void mb_charout_cntl(ch)
char ch;
{
  if (mb_prewin.column < COLS-1) {
    ch &= 0x1f;
    ch |= 0x40;
    win_attrib(att_controlchar,&mb_prewin);
    win_charout_cntl(ch,&mb_prewin);
    win_attrib(att_normal,&mb_prewin);
  }
  else {
    beep();
  }
}

void rem_mb_display_buf(buffer,len)
char *buffer;
int len;
{
  win_attrib(att_remote,&mb_textwin);
  win_textout_len(buffer,len,&mb_textwin,1);
  win_attrib(att_normal,&mb_textwin);
}

void mb_display_buf(buffer,len)
char *buffer;
int len;
{
  win_textout_len(buffer,len,&mb_textwin,1);
  write_mbfile(len,RX_RCV,buffer);
}

void mb_pre_display(buffer,cr)
char *buffer;
int cr;
{
  win_stringout(buffer,&mb_prewin,0);
  if (cr) win_charout(CR,&mb_prewin,0);
}

void mb_input(str,len)
char *str;
int len;
{
  write_iface(0,len,str);
}

void mb_stat_display()
{
  char *stat_str;
  int i;

  stat_str = (char *)malloc(COLS+2);
  strcpy(stat_str,"");
  strcat(stat_str,"\015**");
  strcat(stat_str,"************************************* ");
  if (if_stat[0].iface != -1)
    strcat(stat_str,"MAILBOX **************");
  else
    strcat(stat_str,"MAILBOX NOT CONNECTED ");
  strcat(stat_str,"******************");
  for (i=strlen(stat_str);i<COLS+1;i++) {
    stat_str[i] = '*';
  }
  stat_str[COLS+1] = '\0';
  win_stringout(stat_str,&mb_statwin,0);
  free(stat_str);
}

void init_mb()
{
  open_window(lines_mbinput,&mb_prewin,1);
  open_window(1,&mb_statwin,1);
  win_attrib(att_cstatline,&mb_statwin);
  open_window(lines_mboutput,&mb_textwin,1);
  mb_layout[0].win = &mb_prewin;
  mb_layout[0].first_real_line = 0;
  mb_layout[0].win_num_lines = LINES/mbscr_divide-1;
  mb_layout[0].pagesize = LINES/mbscr_divide-2;
  mb_layout[1].win = &mb_statwin;
  mb_layout[1].first_real_line = LINES/mbscr_divide-1;
  mb_layout[1].win_num_lines = 1;
  mb_layout[1].pagesize = 0;
  mb_layout[2].win = &mb_textwin;
  mb_layout[2].first_real_line = LINES/mbscr_divide;
  mb_layout[2].win_num_lines = LINES-LINES/mbscr_divide-1;
  mb_layout[2].pagesize = LINES-LINES/mbscr_divide-4;
  mb_layout[3].win = &statlin;
  mb_layout[3].first_real_line = LINES-1;
  mb_layout[3].win_num_lines = 1;
  mb_layout[3].pagesize = 0;
  mb_layout[4].win = NULL;
  mb_layout[4].first_real_line = 0;
  mb_layout[4].win_num_lines = 0;
  mb_layout[4].pagesize = 0;
 
  mb_stat_display();
}

void exit_mb()
{
  close_iface_con(0,1);
  close_window(&mb_prewin);
  close_window(&mb_statwin);
  close_window(&mb_textwin);
}

#endif
