/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for initialization (init.c)
   created: Mark Wahl DL4YBG 94/01/11
   updated: Mark Wahl DL4YBG 95/01/05
*/

#include "tnt.h"
#include "init.h"

extern int scr_divide;
extern int blist_add_plus;
extern char rem_info_file[];
extern char rem_help_file[];
extern char news_file_name[];
extern char name_file_name[];
extern char tnt_help_file[];
extern char tnt_dir[];
extern char remote_dir[];
extern char ctext_dir[];
extern char abin_dir[];
extern char tnt_logbookfile[];
extern int pty_timeout;
extern int auto_newline;
extern int supp_hicntl;
#ifdef GEN_NEW_USER
extern int unix_new_user;
extern char unix_user_dir[];
extern int unix_first_uid;
extern int unix_user_gid;
#endif
#ifdef USE_IFACE
extern char box_socket[];
extern int tnt_box_ssid;
#endif
extern char tnt_pwfile[];

char tnt_initfile[80];
char tnt_logfile[80];

/* variables filled from init-file */
char device[80];
unsigned int speed;
int speedflag;
char tnt_upfile[80];
char tnt_downfile[80];
char attc_normal;
char attc_statline;
char attc_monitor;
char attc_cstatline;
char attc_controlchar;
char attc_remote;
char attm_normal;
char attm_statline;
char attm_monitor;
char attm_cstatline;
char attm_controlchar;
char attm_remote;
int lines_command;
int lines_monitor;
int lines_input;
int lines_output;
#ifdef USE_IFACE
int lines_mbinput;
int lines_mboutput;
int mbscr_divide;
#endif
extern int lines_xmon;
extern int lines_xmon_pre;
extern int xmon_scr_divide;
extern int num_heardentries;
int color;
int termcap;
char tnt_cookiefile[80];
char tnt_lockfile[80];
char tnt_ctextfile[80];
char remote_user[80];
char resy_log_file[80];
char upload_dir[80];
char download_dir[80];
char run_dir[80];
#ifdef USE_SOCKET
char sock_passfile[80];
#endif
char func_key_file[80];
char macrotext_dir[80];
#ifdef USE_IFACE
char newmaildir[80];
char autobox_dir[80];
#endif

static int analyse_value(str1,str2)
char str1[];
char str2[];
{
  int tmp;
  
  if (strcmp(str1,"device") == 0) {
    strcpy(device,str2);
    return(0);
  }
  else if (strcmp(str1,"speed") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    speedflag = 0;
    switch (tmp) {
    case 4800:
      speed = B4800;
      return(0);
    case 9600:
      speed = B9600;
      return(0);
    case 19200:
      speed = B19200;
      return(0);
    case 38400:
      speed = B38400;
      return(0);
#ifdef USE_HIBAUD
    case 57600:
      speed = B38400;
      speedflag = ASYNC_SPD_HI;
      return(0);
    case 115200:
      speed = B38400;
      speedflag = ASYNC_SPD_VHI;
      return(0);
#endif
    default:
      return(1);
    }
  }
  else if (strcmp(str1,"attc_normal") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_normal = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_statline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_statline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_monitor") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_monitor = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_cstatline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_cstatline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_controlchar") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_controlchar = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attc_remote") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attc_remote = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_normal") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_normal = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_statline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_statline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_monitor") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_monitor = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_cstatline") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_cstatline = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_controlchar") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_controlchar = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"attm_remote") == 0) {
    if (sscanf(str2,"%x",&tmp) != 1) return(1);
    attm_remote = (char) tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_command") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_command = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_monitor") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_monitor = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_input") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_input = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_output") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_output = tmp;
    return(0);
  }
  else if (strcmp(str1,"scr_divide") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    scr_divide = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_xmon") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_xmon = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_xmon_pre") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_xmon_pre = tmp;
    return(0);
  }
  else if (strcmp(str1,"xmon_scr_divide") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    xmon_scr_divide = tmp;
    return(0);
  }
  else if (strcmp(str1,"num_heardentries") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    num_heardentries = tmp;
    return(0);
  }
#ifdef USE_IFACE
  else if (strcmp(str1,"lines_mbinput") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_mbinput = tmp;
    return(0);
  }
  else if (strcmp(str1,"lines_mboutput") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    lines_mboutput = tmp;
    return(0);
  }
  else if (strcmp(str1,"mbscr_divide") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    mbscr_divide = tmp;
    return(0);
  }
#endif
  else if (strcmp(str1,"tnt_upfile") == 0) {
    strcpy(tnt_upfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_downfile") == 0) {
    strcpy(tnt_downfile,str2);
    return(0);
  }
  else if (strcmp(str1,"rem_info_file") == 0) {
    strcpy(rem_info_file,str2);
    return(0);
  }
  else if (strcmp(str1,"rem_help_file") == 0) {
    strcpy(rem_help_file,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_cookiefile") == 0) {
    strcpy(tnt_cookiefile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_lockfile") == 0) {
    strcpy(tnt_lockfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_help_file") == 0) {
    strcpy(tnt_help_file,str2);
    return(0);
  }
  else if (strcmp(str1,"termcap") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    termcap = tmp;
    return(0);
  }
  else if (strcmp(str1,"color") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    color = tmp;
    return(0);
  }
  else if (strcmp(str1,"remote_user") == 0) {
    strcpy(remote_user,str2);
    return(0);
  }
  else if (strcmp(str1,"news_file_name") == 0) {
    strcpy(news_file_name,str2);
    return(0);
  }
  else if (strcmp(str1,"name_file_name") == 0) {
    strcpy(name_file_name,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_ctextfile") == 0) {
    strcpy(tnt_ctextfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_dir") == 0) {
    strcpy(tnt_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"remote_dir") == 0) {
    strcpy(remote_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"ctext_dir") == 0) {
    strcpy(ctext_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"abin_dir") == 0) {
    strcpy(abin_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_logbookfile") == 0) {
    strcpy(tnt_logbookfile,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_pwfile") == 0) {
    strcpy(tnt_pwfile,str2);
    return(0);
  }
  else if (strcmp(str1,"blist_add_plus") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    blist_add_plus = tmp;
    return(0);
  }
  else if (strcmp(str1,"pty_timeout") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    pty_timeout = tmp;
    return(0);
  }
  else if (strcmp(str1,"auto_newline") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    auto_newline = tmp;
    return(0);
  }
  else if (strcmp(str1,"supp_hicntl") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    supp_hicntl = tmp;
    return(0);
  }
  else if (strcmp(str1,"resy_log_file") == 0) {
    strcpy(resy_log_file,str2);
    return(0);
  }
  else if (strcmp(str1,"run_dir") == 0) {
    strcpy(run_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"upload_dir") == 0) {
    strcpy(upload_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"download_dir") == 0) {
    strcpy(download_dir,str2);
    return(0);
  }
#ifdef USE_SOCKET  
  else if (strcmp(str1,"sock_passfile") == 0) {
    strcpy(sock_passfile,str2);
    return(0);
  }
#endif
#ifdef GEN_NEW_USER
  else if (strcmp(str1,"unix_new_user") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    unix_new_user = tmp;
    return(0);
  }
  else if (strcmp(str1,"unix_user_dir") == 0) {
    strcpy(unix_user_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"unix_first_uid") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    unix_first_uid = tmp;
    return(0);
  }
  else if (strcmp(str1,"unix_user_gid") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    unix_user_gid = tmp;
    return(0);
  }
#endif
  else if (strcmp(str1,"func_key_file") == 0) {
    strcpy(func_key_file,str2);
    return(0);
  }
  else if (strcmp(str1,"macrotext_dir") == 0) {
    strcpy(macrotext_dir,str2);
    return(0);
  }
#ifdef USE_IFACE
  else if (strcmp(str1,"box_socket") == 0) {
    strcpy(box_socket,str2);
    return(0);
  }
  else if (strcmp(str1,"newmaildir") == 0) {
    strcpy(newmaildir,str2);
    return(0);
  }
  else if (strcmp(str1,"autobox_dir") == 0) {
    strcpy(autobox_dir,str2);
    return(0);
  }
  else if (strcmp(str1,"tnt_box_ssid") == 0) {
    if (sscanf(str2,"%d",&tmp) != 1) return(1);
    tnt_box_ssid = tmp;
    return(0);
  }
#endif
  else {
    return(1);
  }
}

int read_init_file(argc,argv,unlock)
int argc;
char *argv[];
int *unlock;
{
  FILE *init_file_fp;
  int file_end;
  int file_corrupt;
  char line[82];
  char str1[82];
  char str2[82];
  char tmp_str[80];
  int rslt;
  int warning;
  int wrong_usage;
  int init_file_flag;
  char *str_ptr;
  int scanned;

  strcpy(device,DEF_DEVICE);
  speed = DEF_SPEED;
  speedflag = DEF_SPEEDFLAG;
  attc_normal = DEF_ATTC_NORMAL;
  attc_statline = DEF_ATTC_STATLINE;
  attc_monitor = DEF_ATTC_MONITOR;
  attc_cstatline = DEF_ATTC_CSTATLINE;
  attc_controlchar = DEF_ATTC_CONTROLCHAR;
  attc_remote = DEF_ATTC_REMOTE;
  attm_normal = DEF_ATTM_NORMAL;
  attm_statline = DEF_ATTM_STATLINE;
  attm_monitor = DEF_ATTM_MONITOR;
  attm_cstatline = DEF_ATTM_CSTATLINE;
  attm_controlchar = DEF_ATTM_CONTROLCHAR;
  attm_remote = DEF_ATTM_REMOTE;
  lines_command = DEF_LINES_COMMAND;
  lines_monitor = DEF_LINES_MONITOR;
  lines_input = DEF_LINES_INPUT;
  lines_output = DEF_LINES_OUTPUT;
  scr_divide = DEF_SCR_DIVIDE;
  blist_add_plus = DEF_BLIST_ADD_PLUS;
  strcpy(tnt_upfile,UP_FILE);
  strcpy(tnt_downfile,DWN_FILE);
  strcpy(rem_info_file,DEF_INFO_FILE);
  strcpy(rem_help_file,DEF_HELP_FILE);
  strcpy(tnt_help_file,DEF_TNT_HELP_FILE);
  strcpy(tnt_cookiefile,COOKIE_FILE);
  strcpy(tnt_lockfile,DEF_LOCK_FILE);
  color = DEF_COLOR;
  termcap = DEF_TERMCAP;
  strcpy(remote_user,DEF_REMOTE_USER);
  strcpy(news_file_name,DEF_NEWS_FILE_NAME);
  strcpy(name_file_name,DEF_NAME_FILE_NAME);
  strcpy(tnt_ctextfile,DEF_TNT_CTEXTFILE);
  strcpy(tnt_dir,DEF_TNT_DIR);
  strcpy(remote_dir,DEF_REMOTE_DIR);
  strcpy(ctext_dir,DEF_CTEXT_DIR);
  strcpy(abin_dir,DEF_ABIN_DIR);
  strcpy(tnt_logbookfile,DEF_TNT_LOGBOOKFILE);
  strcpy(tnt_pwfile,DEF_TNT_PWFILE);
  pty_timeout = DEF_PTY_TIMEOUT;
  auto_newline = DEF_AUTO_NEWLINE;
  supp_hicntl = DEF_SUPP_HICNTL;
  strcpy(resy_log_file,DEF_RESY_LOG_FILE);
  strcpy(run_dir,DEF_RUN_DIR);
  strcpy(upload_dir,DEF_UPLOAD_DIR);
  strcpy(download_dir,DEF_DOWNLOAD_DIR);
#ifdef USE_SOCKET
  strcpy(sock_passfile,DEF_SOCKPASS_FILE);
#endif
#ifdef GEN_NEW_USER
  unix_new_user = DEF_UNIX_NEW_USER;
  strcpy(unix_user_dir,DEF_UNIX_USER_DIR);
  unix_first_uid = DEF_UNIX_FIRST_UID;
  unix_user_gid = DEF_UNIX_USER_GID;
#endif
  strcpy(func_key_file,DEF_FUNC_KEY_FILE);
  strcpy(macrotext_dir,DEF_MACROTEXT_DIR);
#ifdef USE_IFACE
  strcpy(box_socket,DEF_BOX_SOCKET);
  strcpy(newmaildir,DEF_NEWMAILDIR);
  strcpy(autobox_dir,DEF_AUTOBOX_DIR);
  lines_mbinput = DEF_LINES_MBINPUT;
  lines_mboutput = DEF_LINES_MBOUTPUT;
  mbscr_divide = DEF_MBSCR_DIVIDE;
  tnt_box_ssid = DEF_TNT_BOX_SSID;
#endif
  lines_xmon = DEF_LINES_XMON;
  lines_xmon_pre = DEF_LINES_XMON_PRE;
  xmon_scr_divide = DEF_XMON_SCR_DIVIDE;
  num_heardentries = DEF_NUM_HEARDENTRIES;
  
  strcpy(tnt_initfile,INIT_FILE);
  strcpy(tnt_logfile,"");
  wrong_usage = 0;
  scanned = 1;
  while ((scanned < argc) && (!wrong_usage)) {
    if (strcmp(argv[scanned],"-i") == 0) {
      scanned++;
      if (scanned < argc) {
        strcpy(tnt_initfile,argv[scanned]);
      }
      else wrong_usage = 1;
    }
    else if (strcmp(argv[scanned],"-l") == 0) {
      scanned++;
      if (scanned < argc) {
        strcpy(tnt_logfile,argv[scanned]);
      }
      else wrong_usage = 1;
    }
    else if (strcmp(argv[scanned],"-u") == 0) {
      *unlock = 1;
    }
    else {
      wrong_usage = 1;
    }
    scanned++;
  }
  if (wrong_usage) {
    printf("Usage : tnt [-i <init-file>] [-l <log-file>] [-u]\n");
    return(1);
  }
  
  warning = 0;
  if (!(init_file_fp = fopen(tnt_initfile,"r"))) {
    str_ptr = getenv("HOME");
    if (str_ptr != NULL) {
      strcpy(tmp_str,str_ptr);
      strcat(tmp_str,"/");
      strcat(tmp_str,tnt_initfile);
      if (!(init_file_fp = fopen(tmp_str,"r"))) {
        warning = 1;
      }
    }
    else warning = 1;
  }
  if (warning) {
    printf("WARNING: %s not found\n\n",tnt_initfile);
    sleep(2);
    return(0);
  }
  file_end = 0;
  file_corrupt = 0;
  while (!file_end) {
    if (fgets(line,82,init_file_fp) == NULL) {
      file_end = 1;
    }
    else {
      if (strlen(line) == 82) {
        file_end = 1;
        file_corrupt = 1;
      }
      else {
        if (line[0] != '#') { /* ignore comment-lines */
          rslt = sscanf(line,"%s %s",str1,str2);
          switch (rslt) {
          case EOF: /* ignore blank lines */
            break;
          case 2:
            if (analyse_value(str1,str2)) {
              file_end = 1;
              file_corrupt = 1;
            }
            break;
          default:
            file_end = 1;
            file_corrupt = 1;
            break;
          }
        }
      }
    }
  }
  fclose(init_file_fp);
  if (file_corrupt) {
    if (line == NULL) line[0] = '\0';
    printf("WARNING: tnt.ini is in wrong format, wrong line:\n%s\n\n",line);
    sleep(5);
    return(0);
  }
  else {
    return(0);
  }
}
