/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for keyboard input (keyboard.c)
   created: Mark Wahl DL4YBG 93/07/24
   updated: Mark Wahl DL4YBG 95/01/06
*/

#include "tnt.h"
#include "keys.h"
#include "boxlist.h"

extern int act_channel;
extern int act_mode;
extern int xmon_screen;
extern int pause_flag[CHANNELS];
extern int tx_hiwater;
extern struct channel_stat ch_stat[CHANNELS];
extern char *term;
extern char ok_text[];
extern int resync_count;
extern int script_channel;
extern char tnt_dir[];
extern char func_key_file[];

int chan_flag;
int tnc_command;

static int esc_flag;
static int pass_flag;
#ifdef USE_IFACE
static int mb_pass_flag;
#endif
static int exit_flag;
static struct termios okbd_termios;
static struct termios nkbd_termios;
static char func_str[11];
static char input_str[MAXCOLS+1];
static char tnc_string[257];

#define MACRO_MIN 1
#define MACRO_MAX 12
struct macro_string { char string[100]; };
static struct macro_string macro_strings[MACRO_MAX+1];

extern void open_logfile();
extern void close_file();
extern void open_comscript();
extern void open_sendfile();
extern void break_send();
extern void rem_time();
extern void rem_cookie();
extern void cmd_shell();
extern void cmd_endshell();
extern void cmd_redir();
static void exit_tnt();
static void cmd_settime();
static void cmd_setdate();
static void cmd_flag();
static void cmd_mycall();
static void cmd_tnc();
extern void rem_ctext();
extern void rem_name();
static void cmd_chdir();
static void cmd_cwdir();
static void cmd_resync();
extern void cmd_xconnect();
extern void cmd_run();
#ifdef USE_SOCKET
extern void cmd_socket();
#endif
#ifdef USE_IFACE
extern void cmd_iface();
extern void cmd_endiface();
extern void cmd_actiface();
extern void cmd_deactiface();
extern void cmd_finiface();
extern void cmd_box();
extern void cmd_endbox();
extern void cmd_actbox();
extern void cmd_deactbox();
extern void cmd_finbox();
extern void cmd_monboxlist();
#endif
static void cmd_kmacro();
extern void cmd_msend();
extern void cmd_extmon();
extern void cmd_endextmon();
extern void cmd_logblist();
extern void cmd_blist();
extern void cmd_xblist();
extern void open_monfile();
extern void close_monfile();
extern void open_xmonfile();
extern void close_xmonfile();
extern void cmd_concall();
extern void cmd_priv();
extern void cmd_loadpriv();
extern void cmd_listpriv();

/* values for flag command */
#define F_INFOBELL 0
#define F_CBELL 1
#define F_APPEND 2
#define F_COOKIE 3
#define F_REMOTE 4
#define F_UMLAUT 5
#define F_TXECHO 6
#define F_CTEXT 8
#define F_PTYECHO 9
#define F_AUTOBIN 10
#define F_XMON 11
#define F_MONBOX 12
#define F_AUTOBOX 13
#define F_HEARD 14
#define F_LAYER3 15

/* variables for flag command */
int infobell_flag;
int cbell_flag;
int append_flag;
int cookie_flag;
int remote_flag;
extern int umlaut;
int txecho_flag;
int ctext_flag;
int ptyecho_flag;
int autobin_flag;
extern int xmon_flag;
extern int mheard_flag;
#ifdef USE_IFACE
extern int monbox_flag;
extern int autobox_flag;
#endif
int layer3_flag;

static struct com_list gen_com_list[] = {
  {"ALFDISP" ,4,0,NULL,"A"  ,0,0},
  {"A"       ,1,0,NULL,"A"  ,0,0},
  {"DAMADIS" ,4,0,NULL,"B"  ,0,0},
  {"B"       ,1,0,NULL,"B"  ,0,0},
  {"VERSION" ,1,0,NULL,"V"  ,0,0},
  {"CHECK"   ,2,0,NULL,"@T3",0,0},
  {"@T3"     ,3,0,NULL,"@T3",0,0},
  {"CONNECT" ,1,0,NULL,"C"  ,0,0},
  {"CTEXT"   ,2,0,NULL,"U"  ,0,0},
  {"U"       ,1,0,NULL,"U"  ,0,0},
  {"DIGIPEAT",4,0,NULL,"R"  ,0,0},
  {"R"       ,1,0,NULL,"R"  ,0,0},
  {"DISCONNE",1,0,NULL,"D"  ,0,0},
  {"DAYTIME" ,5,0,NULL,"K"  ,0,0},
  {"K"       ,1,0,NULL,"K"  ,0,0},
  {"ECHO"    ,1,0,NULL,"E"  ,0,0},
  {"FLOW"    ,2,0,NULL,"Z"  ,0,0},
  {"Z"       ,1,0,NULL,"Z"  ,0,0},
  {"FRACK"   ,1,0,NULL,"F"  ,0,0},
  {"FULLDUP" ,2,0,NULL,"@D" ,0,0},
  {"@D"      ,2,0,NULL,"@D" ,0,0},
  {"MAXFRAME",3,0,NULL,"O"  ,0,0},
  {"O"       ,1,0,NULL,"O"  ,0,0},
  {"MONITOR" ,1,0,NULL,"M"  ,0,0},
  {"MYCALL"  ,2,1,cmd_mycall,"I"  ,0,0},
  {"I"       ,1,0,NULL,"I"  ,0,0},
  {"MHEARD"  ,2,0,NULL,"H"  ,0,0},
  {"H"       ,1,0,NULL,"H"  ,0,0},
  {"PERSIST" ,1,0,NULL,"P"  ,0,0},
  {"RESPTIME",3,0,NULL,"@T2",0,0},
  {"@T2"     ,3,0,NULL,"@T2",0,0},
  {"RETRY"   ,2,0,NULL,"N"  ,0,0},
  {"N"       ,1,0,NULL,"N"  ,0,0},
  {"SLOTTIME",2,0,NULL,"W"  ,0,0},
  {"W"       ,1,0,NULL,"W"  ,0,0},
  {"TXDELAY" ,1,0,NULL,"T"  ,0,0},
  {"USERS"   ,2,0,NULL,"Y"  ,0,0},
  {"Y"       ,1,0,NULL,"Y"  ,0,0},
  {"XMITOK"  ,1,0,NULL,"X"  ,0,0},
  {"BUFFERS" ,3,0,NULL,"@B" ,0,0},
  {"@B"      ,2,0,NULL,"@B" ,0,0},
  {"A1SRTT"  ,6,0,NULL,"@A1",0,0},
  {"@A1"     ,3,0,NULL,"@A1",0,0},
  {"A2SRTT"  ,6,0,NULL,"@A2",0,0},
  {"@A2"     ,3,0,NULL,"@A2",0,0},
  {"A3SRTT"  ,6,0,NULL,"@A3",0,0},
  {"@A3"     ,3,0,NULL,"@A3",0,0},
  {"IPOLL"   ,3,0,NULL,"@I" ,0,0},
  {"@I"      ,2,0,NULL,"@I" ,0,0},
  {"8BIT"    ,3,0,NULL,"@M" ,0,0},
  {"@M"      ,4,0,NULL,"@M" ,0,0},
  {"VALCALL" ,3,0,NULL,"@V" ,0,0},
  {"@V"      ,4,0,NULL,"@V" ,0,0},
  {"CHANNEL" ,4,2,NULL,"" ,0,0},
  {"S"       ,1,2,NULL,"" ,0,0},
  {"TNC"     ,3,1,cmd_tnc,"",0,0},
  {"LOGQSO"  ,4,1,open_logfile,"",RX_NORM,RX_RCV | RX_SND},
  {"LOGREC"  ,4,1,open_logfile,"",RX_NORM,RX_RCV},
  {"LOGSND"  ,4,1,open_logfile,"",RX_NORM,RX_SND},
  {"READBIN" ,5,1,open_logfile,"",RX_BIN,RX_RCV},
  {"READ"    ,4,1,open_logfile,"",RX_PLAIN,RX_RCV},
  {"READABIN",5,1,open_logfile,"",RX_ABIN,RX_RCV},
  {"LOGABIN" ,4,1,open_logfile,"",RX_ABIN_Q,RX_RCV},
  {"LOGMON"  ,4,1,open_monfile,"",RX_NORM,0},
  {"RDMON"   ,5,1,open_monfile,"",RX_PLAIN,0},
  {"RDMONBIN",6,1,open_monfile,"",RX_BIN,0},
  {"LOGXMON" ,4,1,open_xmonfile,"",RX_NORM,0},
  {"RDXMON"  ,6,1,open_xmonfile,"",RX_PLAIN,0},
  {"RDXMONBI",7,1,open_xmonfile,"",RX_BIN,0},
  {"CLOSE"   ,2,1,close_file,"",0,0},
  {"CLOSEMON",6,1,close_monfile,"",0,0},
  {"CLOSEXMO",6,1,close_xmonfile,"",0,0},
  {"SENDLOG" ,5,1,open_sendfile,"",TX_NORM,0},
  {"SENDBIN" ,5,1,open_sendfile,"",TX_BIN,0},
  {"SENDABIN",5,1,open_sendfile,"",TX_ABIN,0},
  {"SEND"    ,4,1,open_sendfile,"",TX_PLAIN,0},
  {"SENDCOM" ,5,1,open_comscript,"",0,0},
  {"BREAK"   ,2,1,break_send,"",0,0},
  {"TIMESET" ,7,1,cmd_settime,"",0,0},
  {"DATESET" ,7,1,cmd_setdate,"",0,0},
  {"EXIT"    ,2,1,exit_tnt,"",0,0},
  {"INFOBELL",4,1,cmd_flag,"",F_INFOBELL,0},
  {"CBELL"   ,2,1,cmd_flag,"",F_CBELL,0},
  {"APPEND"  ,3,1,cmd_flag,"",F_APPEND,0},
  {"COOKIE"  ,4,1,cmd_flag,"",F_COOKIE,0},
  {"REMOTE"  ,3,1,cmd_flag,"",F_REMOTE,0},
  {"UMLAUT"  ,3,1,cmd_flag,"",F_UMLAUT,0},
  {"TXECHO"  ,3,1,cmd_flag,"",F_TXECHO,0},
  {"CONTEXT" ,4,1,cmd_flag,"",F_CTEXT,0},
  {"PTYECHO" ,4,1,cmd_flag,"",F_PTYECHO,0},
  {"AUTOBIN" ,6,1,cmd_flag,"",F_AUTOBIN,0},
  {"XMON"    ,4,1,cmd_flag,"",F_XMON,0},
  {"HEARD"   ,3,1,cmd_flag,"",F_HEARD,0},
#ifdef USE_IFACE
  {"MONBOX"  ,4,1,cmd_flag,"",F_MONBOX,0},
  {"AUTOBOX" ,6,1,cmd_flag,"",F_AUTOBOX,0},
#endif
  {"LAYER3"  ,3,1,cmd_flag,"",F_LAYER3,0},
#ifdef USE_IFACE
  {"LMONBOX" ,4,1,cmd_monboxlist,"",0,0},
#endif
  {"SCOOKIE" ,3,1,rem_cookie,"",0,0},
  {"SCTEXT"  ,3,1,rem_ctext,"",0,0},
  {"NAME"    ,4,1,rem_name,"",0,0},
  {"STIME"   ,3,1,rem_time,"",0,0},
  {"SHELL"   ,2,1,cmd_shell,"",0,0},
  {"ENDSHELL",4,1,cmd_endshell,"",0,0},
  {"RUN"     ,3,1,cmd_run,"",0,0},
  {"RUNT"    ,4,1,cmd_run,"",1,0},
  {"ENDRUN"  ,4,1,cmd_endshell,"",0,0},
  {"REDIR"   ,3,1,cmd_redir,"",0,0},
  {"ENDREDIR",4,1,cmd_endshell,"",0,0},
#ifdef USE_SOCKET
  {"SOCKET"  ,3,1,cmd_socket,"",0,0},
  {"ENDSOCK" ,4,1,cmd_endshell,"",0,0},
#endif
#ifdef USE_IFACE
  {"IFACE"   ,3,1,cmd_iface,"",0,0},
  {"ENDIFACE",5,1,cmd_endiface,"",0,0},
  {"ACTIF"   ,4,1,cmd_actiface,"",0,0},
  {"DEACTIF" ,4,1,cmd_deactiface,"",0,0},
  {"FINIFACE",5,1,cmd_finiface,"",0,0},
  {"BOX"     ,3,1,cmd_box,"",0,0},
  {"ENDBOX"  ,4,1,cmd_endbox,"",0,0},
  {"ACTBOX"  ,4,1,cmd_actbox,"",0,0},
  {"DEACTBOX",6,1,cmd_deactbox,"",0,0},
  {"FINBOX"  ,4,1,cmd_finbox,"",0,0},
#endif
  {"CD"      ,2,1,cmd_chdir,"",0,0},
  {"CWD"     ,3,1,cmd_cwdir,"",0,0},
  {"RESYNC"  ,4,1,cmd_resync,"",0,0},
  {"XCONNECT",2,1,cmd_xconnect,"",0,0},
  {"KMACRO"  ,3,1,cmd_kmacro,"",0,0},
  {"MSEND"   ,3,1,cmd_msend,"",0,0},
  {"EXTMON"  ,3,1,cmd_extmon,"",0,0},
  {"EXTAMON" ,4,1,cmd_extmon,"",1,0},
  {"ENDEXTM" ,5,1,cmd_endextmon,"",0,0},
  {"LOGBLIST",4,1,cmd_logblist,"",0,0},
  {"BLIST"   ,3,1,cmd_blist,"",0,0},
  {"XBLIST"  ,3,1,cmd_xblist,"",0,0},
  {"CONCALL" ,4,1,cmd_concall,"",0,0},
  {"PRIV"    ,4,1,cmd_priv,"",0,0},
  {"LOADPRIV",8,1,cmd_loadpriv,"",0,0},
  {"LISTPRIV",8,1,cmd_listpriv,"",0,0},
  {""        ,0,-1,NULL,""   ,0,0}
};

#ifdef USE_IFACE
/* commands allowed in mailbox screen */
static struct com_list mb_com_list[] = {
  {"LOGREC"  ,4,1,open_logfile,"",RX_NORM,RX_RCV},
  {"READBIN" ,5,1,open_logfile,"",RX_BIN,RX_RCV},
  {"READ"    ,4,1,open_logfile,"",RX_PLAIN,RX_RCV},
  {"CLOSE"   ,2,1,close_file,"",0,0},
  {"EXIT"    ,2,1,exit_tnt,"",0,0},
  {"APPEND"  ,3,1,cmd_flag,"",F_APPEND,0},
  {"UMLAUT"  ,3,1,cmd_flag,"",F_UMLAUT,0},
  {"TXECHO"  ,3,1,cmd_flag,"",F_TXECHO,0},
  {"XMON"    ,4,1,cmd_flag,"",F_XMON,0},
  {"HEARD"   ,3,1,cmd_flag,"",F_HEARD,0},
  {"MONBOX"  ,4,1,cmd_flag,"",F_MONBOX,0},
  {"AUTOBOX" ,6,1,cmd_flag,"",F_AUTOBOX,0},
  {"LMONBOX" ,4,1,cmd_monboxlist,"",0,0},
  {"IFACE"   ,3,1,cmd_iface,"",0,0},
  {"ENDIFACE",5,1,cmd_endiface,"",0,0},
  {"ACTIF"   ,4,1,cmd_actiface,"",0,0},
  {"DEACTIF" ,4,1,cmd_deactiface,"",0,0},
  {"FINIFACE",5,1,cmd_finiface,"",0,0},
  {"BOX"     ,3,1,cmd_box,"",0,0},
  {"ENDBOX"  ,4,1,cmd_endbox,"",0,0},
  {"ACTBOX"  ,4,1,cmd_actbox,"",0,0},
  {"DEACTBOX",6,1,cmd_deactbox,"",0,0},
  {"FINBOX"  ,4,1,cmd_finbox,"",0,0},
  {"CD"      ,2,1,cmd_chdir,"",0,0},
  {"CWD"     ,3,1,cmd_cwdir,"",0,0},
  {"RESYNC"  ,4,1,cmd_resync,"",0,0},
  {"KMACRO"  ,3,1,cmd_kmacro,"",0,0},
  {"MSEND"   ,3,1,cmd_msend,"",0,0},
  {"LOGBLIST",4,1,cmd_logblist,"",0,0},
  {"BLIST"   ,3,1,cmd_blist,"",0,0},
  {"XBLIST"  ,3,1,cmd_xblist,"",0,0},
  {""        ,0,-1,NULL,""   ,0,0}
};
#endif

/* commands allowed in extended monitor screen */
static struct com_list xm_com_list[] = {
  {"LOGXMON" ,4,1,open_xmonfile,"",RX_NORM,0},
  {"RDXMON"  ,6,1,open_xmonfile,"",RX_PLAIN,0},
  {"RDXMONBI",7,1,open_xmonfile,"",RX_BIN,0},
  {"CLOSEXMO",6,1,close_xmonfile,"",0,0},
  {"EXIT"    ,2,1,exit_tnt,"",0,0},
  {"APPEND"  ,3,1,cmd_flag,"",F_APPEND,0},
  {"XMON"    ,4,1,cmd_flag,"",F_XMON,0},
  {"CD"      ,2,1,cmd_chdir,"",0,0},
  {"CWD"     ,3,1,cmd_cwdir,"",0,0},
  {"KMACRO"  ,3,1,cmd_kmacro,"",0,0},
  {"MSEND"   ,3,1,cmd_msend,"",0,0},
  {"EXTMON"  ,3,1,cmd_extmon,"",0,0},
  {"EXTAMON" ,4,1,cmd_extmon,"",1,0},
  {"ENDEXTM" ,5,1,cmd_endextmon,"",0,0},
  {""        ,0,-1,NULL,""   ,0,0}
};

void init_keyboard()
{
  txecho_flag = 1;
  ptyecho_flag = 1;
  esc_flag = 0;
  pass_flag = 0;
#ifdef USE_IFACE
  mb_pass_flag = 0;
#endif
  chan_flag = 0;
  exit_flag = 0;
  tcgetattr(0,&okbd_termios);
  nkbd_termios = okbd_termios;
#ifdef USE_SELECT
  nkbd_termios.c_cc[VTIME] = 0;
  nkbd_termios.c_cc[VMIN] = 1;
#else
  nkbd_termios.c_cc[VTIME] = 0;
  nkbd_termios.c_cc[VMIN] = 0;
#endif
  nkbd_termios.c_cc[VSTART] = -1;
  nkbd_termios.c_cc[VSTOP] = -1;
  nkbd_termios.c_iflag = 0;
  nkbd_termios.c_iflag |= (IGNBRK|ICRNL);
  nkbd_termios.c_oflag = 0;
  nkbd_termios.c_lflag = 0;
  nkbd_termios.c_cflag |= (CS8|CREAD|CLOCAL);
#ifdef HAS_CRTSCTS
  nkbd_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|CRTSCTS|HUPCL);
#else
  nkbd_termios.c_cflag &= ~(CSTOPB|PARENB|PARODD|HUPCL);
#endif
  tcsetattr(0,TCSADRAIN,&nkbd_termios);
  cmd_kmacro(0,0,0,0,M_CMDSCRIPT,NULL); /* init keyboard macros */ 
}

void exit_keyboard()
{
  tcsetattr(0,TCSADRAIN,&okbd_termios);
}

/* analyse input-string (command) */
void cmd_input(channel,mode,str,len,cscript)
int channel;
int mode;
char *str;
int len;
int cscript;
{
  char com_string[9];
  int found;
  int i,j;
  int invalid;
  int chan;
  struct com_list *com_list;
  
  if (mode == M_EXTMON)
    com_list = xm_com_list;
#ifdef USE_IFACE
  else if (mode == M_MAILBOX)
    com_list = mb_com_list;
#endif
  else
    com_list = gen_com_list;
  tnc_command = 0;
  invalid = 0;
  /* delete CR at end of string and return if empty */
  while ((str[len-1] == '\n') || (str[len-1] == CR)) {
    str[len-1] = '\0';
    len--;
  }
  if (len == 0) return;
  i = 0;
  while ((i < 8) && (i < len) && (*str != SPACE)) {
    com_string[i] = *str++;
    if ((com_string[i] > 0x60) && (com_string[i] < 0x7E))
      com_string[i] &= 0xDF;
    i++;
  }
  com_string[i]= '\0';
  if ((i != len) && (*str != SPACE)) {
    invalid = 1;
  }
  len -= i;
  if (!invalid) {
    /* find string in command-list */
    found = 0;
    j = 0;
    while ((com_list[j].ext_com != -1) && !found) {
      if (strstr(com_list[j].string,com_string) == com_list[j].string) {
        /* string found, now check if length valid */
        if (strlen(com_string) >= com_list[j].len) {
          found = 1;
        }
      }
      if (!found) j++;
    }
    if (found) {
      /* delete spaces before parameters */
      while (*str == SPACE) {
        str++;
        len--;
      }
      switch (com_list[j].ext_com) {
      case 0: /* tnc command */
        tnc_command = 1;
        strcpy(tnc_string,com_list[j].new_string);
        if (len)
          strcat(tnc_string,str);
        len = strlen(tnc_string);
        queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
        break;
      case 1: /* external command */
        (*com_list[j].func)(com_list[j].par1,com_list[j].par2,
                            channel,len,mode,str);
        break;
      case 2: /* change of channel */
        chan = atoi(str);
        if ((chan >= 0) && (chan < CHANNELS)) {
          if (cscript) { /* change command script channel */
            script_channel = chan;
          }
          else { /* change channel */
            sel_connect(chan);
            statlin_update();
          }
        }
        break;
      }
      return;
    }
  }
  cmd_display(mode,channel,"Invalid command",1);
  return;
}

void exit_tnt(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  cmd_display(mode,channel,"Do you really want to exit (y/n)? ",0);
  exit_flag = 1;
}

/* analyse input-string (data) */
void dat_input(channel,str,len)
int channel;
char *str;
int len;
{
  int flag;
  
  flag = 0;
  queue_cmd_data(channel,X_DATA,len,flag,str);
  write_file(channel,len,RX_SND,str);
}

void keyboard_server(state,ch)
int *state;
char *ch;
{
  int i;
  int found;
  int analysed;
  int input_len;
  char temp_str[100];
  int num_func;
  struct func_keys *spec_keys;
  char msgs[100];
  

  if (exit_flag) {
    switch (*ch) {
      case 'Y':
      case 'y':
        *state = S_TNT_DWN;
        open_dwnfile(state);
        exit_flag = 0;
        if (act_mode != M_EXTMON)
          cmd_display(act_mode,act_channel,"Y",1);
        else
          cmd_display(act_mode,xmon_screen,"Y",1);
        break;
      case 'N':
      case 'n':
        exit_flag = 0;
        if (act_mode != M_EXTMON)
          cmd_display(act_mode,act_channel,"N",1);
        else
          cmd_display(act_mode,xmon_screen,"N",1);
        break;
    }
    return;
  }
  else if (chan_flag) {
    if ((*ch >= '0') && (*ch <= '9')) {
      if (act_mode == M_EXTMON) {
        if (*ch >= '5') return;
        sel_xmon((int)(*ch-'0'));
      }
      else
        sel_connect((int)(*ch-'0'));
      chan_flag = 0;
      statlin_update();
      return;
    }
    else if (*ch == '@') {
      chan_flag = 0;
      sel_connect((int)(*ch-'@'+10));
      statlin_update();
      return;
    }
    else return; /* wait for valid channel number */
  }
  if (esc_flag) {
    found = 0;
    if (strcmp(term,"xterm") == 0) {
      spec_keys = special_keys_xterm;
      num_func = NUMFUNC_XTERM;
    }
    else {
      spec_keys = special_keys;
      num_func = NUMFUNC;
    }
    func_str[esc_flag] = *ch;
    func_str[esc_flag + 1] = '\0';
    for (i = 0; (i < num_func) && (!found); i++) {
      if (strstr(spec_keys[i].code,func_str) == spec_keys[i].code) {
        found = 1;
        if (strlen(spec_keys[i].code) == (esc_flag + 1)) {
          esc_flag = 0;
          if (spec_keys[i].sel_channel == 2) {
#ifdef USE_IFACE
            if ((act_mode != M_CONNECT) && (act_mode != M_COMMAND) &&
                (act_mode != M_MAILBOX))
#else
            if ((act_mode != M_CONNECT) && (act_mode != M_COMMAND))
#endif
              return;
            if ((tx_hiwater) || (!data_allowed(act_channel))) {
              beep();
              return;
            }
            strcpy(msgs,macro_strings[(int)spec_keys[i].res_code].string);
            if(strcmp(msgs,"")) {
              if(msgs[0]==':') {
	        if (act_mode == M_COMMAND)
	          cmd_display(act_mode,act_channel,&msgs[1],1);
                else
	          cmd_display(act_mode,act_channel,msgs,1);
	        cmd_input(act_channel,act_mode,&msgs[1],strlen(msgs)-1,0);
	        return;
	      }
	      else {
	        if (act_mode != M_CONNECT)
	          return;
	        if(ch_stat[act_channel].state < 4)
	          return;
	        if(msgs[strlen(msgs)-1]=='*') {
	          msgs[strlen(msgs)-1]=CR;
	          dat_input(act_channel,msgs,strlen(msgs));
	          if(txecho_flag)
	            rem_data_display(act_channel,msgs);
	        }
	        cmd_display(act_mode,act_channel,msgs,0);
	        return;
	      }
	    }
	    return;
          }
          if (spec_keys[i].sel_channel == 1) {
            sel_connect((int)spec_keys[i].res_code);
            return;
          }
          else {
            *ch = spec_keys[i].res_code;
          }
        }
        else {
          if (++esc_flag > 10) esc_flag = 0;
          return;
        }
      }
    }
    if (!found) {
      esc_flag = 0;
      return;
    }
  }
  if ((*ch < SPACE) || ((*ch >= '\240') && (*ch <= '\247'))) {
    analysed = 1;
    switch (*ch) {
      case ESC:
        func_str[0] = ESC;
        esc_flag = 1;
        break;
      case C_MONITOR:
        sel_monitor();
        break;
      case C_COMMAND:
        sel_command();
        break;
      case C_CONNECT:
        sel_connect(act_channel);
        break;
      case C_EXTMON:
        sel_xmon(xmon_screen);
        break;
      case C_MHEARD:
        sel_heardlist();
        break;
      case C_BOXLIST:
        sel_boxlist();
        break;
      case C_HELP:
        sel_help();
        break;
#ifdef USE_IFACE
      case C_MAILBOX:
        sel_mailbox();
        break;
#endif    
      case C_CHANNEL:
        chan_flag = 1;
        statlin_update();
        break;
      default:
        analysed = 0;
        break;
    }
    if (analysed) return;
  }
  if (*state == S_NORMAL) {
    switch (act_mode) {
    case M_CONNECT:
      if (pass_flag) {
        pre_charout_cntl(act_channel,*ch);
        pass_flag = 0;
      }
      else if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            if ((tx_hiwater) || (!data_allowed(act_channel))) {
              beep();
            }
            else {
              input_len = pre_newline(act_channel,input_str);
              if ((input_len > 1) && (input_str[0] == ':'))
                cmd_input(act_channel,act_mode,input_str + 1,input_len - 1,0);
              else {
                dat_input(act_channel,input_str,input_len);
                if (txecho_flag) {
                  insert_cr_rx(act_channel);
                  rem_data_display_buf(act_channel,input_str,input_len);
                }
              }
            }
            break;
          case C_PASS:
            pass_flag = 1;
            break;
          case C_PAUSE:
            pause_flag[act_channel] ^= 1;
            statlin_update();
            break;
          case C_WINUP:
            real_window_up(2,1);
            break;
          case C_WINDWN:
            real_window_down(2,1);
            break;
          case C_LINEUP:
            real_window_up(2,0);
            break;
          case C_LINEDWN:
            real_window_down(2,0);
            break;
          case C_CUTOP:
            real_window_top(2);
            break;
          case C_CUBOT:
            real_window_bot(2);
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            pre_charout_nobnd(act_channel,*ch);
            break;
          case C_CURIGHT:
            pre_charout(act_channel,*ch);
            break;
          default:
            pre_charout_cntl(act_channel,*ch);
            break;
        }
      }
      else {
        pre_charout(act_channel,*ch);
      }
      break;
    case M_COMMAND:
      if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            if (tx_hiwater)
              beep();
            else {  
              input_len = cmd_newline(act_channel,input_str);
              cmd_input(act_channel,act_mode,input_str,input_len,0);
            }
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_WINUP:
          case C_WINDWN:
          case C_CUTOP:
          case C_CUBOT:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            cmd_charout_nobnd(act_channel,*ch);
            break;
          case C_CURIGHT:
            cmd_charout(act_channel,*ch);
            break;
        }
      }
      else {
        cmd_charout(act_channel,*ch);
      }
      break;
    case M_MONITOR:
      switch (*ch) {
        case C_PAUSE:
          pause_flag[0] ^= 1;
          statlin_update();
          break;
        case C_WINUP:
          real_window_up(0,1);
          break;
        case C_WINDWN:
          real_window_down(0,1);
          break;
        case C_LINEUP:
        case C_CUUP:
          real_window_up(0,0);
          break;
        case C_LINEDWN:
        case C_CUDWN:
          real_window_down(0,0);
          break;
        case C_CUTOP:
          real_window_top(0);
          break;
        case C_CUBOT:
          real_window_bot(0);
          break;
      }
      break;
    case M_EXTMON:
      if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            if (tx_hiwater)
              beep();
            else {  
              input_len = xm_newline(xmon_screen,input_str);
              cmd_input(xmon_screen,act_mode,input_str,input_len,0);
            }
            break;
          case C_WINUP:
            real_window_up(2,1);
            break;
          case C_WINDWN:
            real_window_down(2,1);
            break;
          case C_LINEUP:
            real_window_up(2,0);
            break;
          case C_LINEDWN:
            real_window_down(2,0);
            break;
           case C_CUTOP:
            real_window_top(2);
            break;
          case C_CUBOT:
            real_window_bot(2);
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            xm_charout_nobnd(xmon_screen,*ch);
            break;
          case C_CURIGHT:
            xm_charout(xmon_screen,*ch);
            break;
        }
      }
      else {
        xm_charout(xmon_screen,*ch);
      }
      break;
    case M_HEARD:
      switch (*ch) {
        case C_WINUP:
        case C_WINDWN:
        case C_CUUP:
        case C_CUDWN:
        case C_CUTOP:
        case C_CUBOT:
          heard_charout_cntl(*ch);
          break;
      }
      break;
    case M_HELP:
      switch (*ch) {
        case C_WINUP:
        case C_WINDWN:
        case C_CUUP:
        case C_CUDWN:
        case C_CUTOP:
        case C_CUBOT:
          help_charout_cntl(*ch);
          break;
      }
      break;
    case M_BOXLIST:
      switch (*ch) {
        case LF:
          blist_analyse_line(act_channel,FUNC_READ);
          break;
        case 'l':
        case 'L':
          blist_analyse_line(act_channel,FUNC_LIST);
          break;
        case 'e':
        case 'E':
          blist_analyse_line(act_channel,FUNC_ERASE);
          break;
        case 'k':
        case 'K':
          blist_analyse_line(act_channel,FUNC_KILL);
          break;
        case C_CUUP:
        case C_CUDWN:
        case C_WINUP:
        case C_WINDWN:
        case C_CUTOP:
        case C_CUBOT:
          blist_charout_cntl(act_channel,*ch);
          break;
      }
      break;
#ifdef USE_IFACE
    case M_MAILBOX :
      if (mb_pass_flag) {
        mb_charout_cntl(*ch);
        mb_pass_flag = 0;
      }
      else if ((*ch < SPACE) || (*ch == ESC2) || (*ch == DEL)) {
        switch (*ch) {
          case LF:
            input_len = mb_newline(input_str);
            if ((input_len > 1) && (input_str[0] == ':'))
              cmd_input(act_channel,act_mode,input_str + 1,input_len - 1,0);
            else {
              mb_input(input_str,input_len);
              if (txecho_flag) {
                rem_mb_display_buf(input_str,input_len);
              }
            }
            break;
          case C_PASS:
            mb_pass_flag = 1;
            break;
          case C_PAUSE:
            break;
          case C_WINUP:
            real_window_up(2,1);
            break;
          case C_WINDWN:
            real_window_down(2,1);
            break;
          case C_LINEUP:
            real_window_up(2,0);
            break;
          case C_LINEDWN:
            real_window_down(2,0);
            break;
          case C_CUTOP:
            real_window_top(2);
            break;
          case C_CUBOT:
            real_window_bot(2);
            break;
          case C_CULEFT:
          case C_CUUP:
          case C_CUDWN:
          case C_STLINE:
          case C_EOLINE:
          case C_INSERT:
          case C_DELLINE:
          case C_DELCHAR:
          case BS:
          case DEL:
            mb_charout_nobnd(*ch);
            break;
          case C_CURIGHT:
            mb_charout(*ch);
            break;
          default:
            mb_charout_cntl(*ch);
            break;
        }
      }
      else {
        mb_charout(*ch);
      }
      break;
#endif
    }
  }
}

static void cmd_settime(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  struct tm timestr;
  time_t timeval;
  char tmpstr[10];
    
  timeval = time(&timeval);
  timestr = *localtime(&timeval);
  sprintf(tmpstr,"K %2.2u:%2.2u:%2.2u",
          timestr.tm_hour,timestr.tm_min,timestr.tm_sec);
  queue_cmd_data(channel,X_COMM,strlen(tmpstr),mode,tmpstr);
  cmd_display(mode,channel,tmpstr,1);
  tnc_command = 1;
}

static void cmd_setdate(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  struct tm timestr;
  time_t timeval;
  char tmpstr[10];
    
  timeval = time(&timeval);
  timestr = *localtime(&timeval);
  sprintf(tmpstr,"K %2.2u.%2.2u.%2.2u",
          timestr.tm_mday,timestr.tm_mon+1,timestr.tm_year);
  queue_cmd_data(channel,X_COMM,strlen(tmpstr),mode,tmpstr);
  cmd_display(mode,channel,tmpstr,1);
  tnc_command = 1;
}

static void cmd_flag(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int *flag;
  int i;
  char ch;
  
  switch (par1) {
  case F_INFOBELL:
    flag = &infobell_flag;
    break;
  case F_CBELL:
    flag = &cbell_flag;
    break;
  case F_APPEND:
    flag = &append_flag;
    break;
  case F_COOKIE:
    flag = &cookie_flag;
    break;
  case F_REMOTE:
    flag = &remote_flag;
    break;
  case F_UMLAUT:
    flag = &umlaut;
    break;
  case F_TXECHO:
    flag = &txecho_flag;
    break;
  case F_CTEXT:
    flag = &ctext_flag;
    break;
  case F_PTYECHO:
    flag = &ptyecho_flag;
    break;
  case F_AUTOBIN:
    flag = &autobin_flag;
    break;
  case F_XMON:
    flag = &xmon_flag;
    break;
#ifdef USE_IFACE
  case F_MONBOX:
    flag = &monbox_flag;
    break;
  case F_AUTOBOX:
    flag = &autobox_flag;
    break;
#endif
  case F_HEARD:
    flag = &mheard_flag;
    break;
  case F_LAYER3:
    flag = &layer3_flag;
    break;
  }
  if (len == 0) {
    if (*flag == 0)
      cmd_display(mode,channel,"OFF",1);
    else
      cmd_display(mode,channel,"ON",1);
    return;
  }
  for (i = 0; i < len; i++) {
    ch = str[i];
    if ((ch > 0x40) && (ch < 0x5b)) {
      ch |= 0x20;
      str[i] = ch;
    }
  }    
  if (len >= 2) {
    if (strncmp(str,"on",2) == 0) {
      *flag = 1;
      cmd_display(mode,channel,"Ok",1);
      return;
    }
  } 
  if (len >= 3) {
    if (strncmp(str,"off",3) == 0) {
      *flag = 0;
      cmd_display(mode,channel,"Ok",1);
      return;
    }
  }
  cmd_display(mode,channel,"INVALID VALUE",1);
}

/* Set permanent mycall */
static void cmd_mycall(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int i;
  
  strcpy(tnc_string,"I");
  if (len) {
    if (strcmp(str,"$") == 0) {
      ch_stat[channel].mycall[0] = '\0';
    }
    else {
      tnc_command = 1;
      for (i=0;i<=9;i++) {
        ch_stat[channel].mycall[i] = toupper(str[i]);
      }
      ch_stat[channel].mycall[9] = '\0';
      strcat(tnc_string,str);
      len = strlen(tnc_string);
      queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
    }
  }
  else {
    if (ch_stat[channel].mycall[0] == '\0') 
      cmd_display(mode,channel,"<empty>",1);
    else
      cmd_display(mode,channel,ch_stat[channel].mycall,1);
  }
}

/* Send TNC command */
static void cmd_tnc(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int i;
  
  if (len) {
    tnc_command = 1;
    strcpy(tnc_string,str);
    len = strlen(tnc_string);
    queue_cmd_data(channel,X_COMM,len,mode,tnc_string);
  }
  else {
    cmd_display(mode,channel,"INVALID TNC STRING",1);
  }
}

/* Change directory */
static void cmd_chdir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int result;
  char *dir;

  dir = str;
  if (strlen(dir) == 0) {
    dir = getenv("HOME");
  }  
  if ((result = chdir(dir)) == -1) {
    cmd_display(mode,channel,"Can't change directory",1);
  }
  else {
    cmd_display(mode,channel,ok_text,1);
  }
}

/* Display working directory */
static void cmd_cwdir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  char dir[80];
  
  if (getcwd(dir,80) == NULL) {
    cmd_display(mode,channel,"Can't get directory name",1);
  }
  else {
    cmd_display(mode,channel,dir,1);
  }
}

/* Display number of resyncs up to now */
static void cmd_resync(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  char number[20];
  
  sprintf(number,"%d",resync_count);
  cmd_display(mode,channel,number,1);
}

/* Read in keyboard-macros file */
static void cmd_kmacro(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  int x,y;
  FILE *fp;
  char tmp[100], tmp2[100];

  /* first delete all macros */
  for (x = MACRO_MIN;x <= MACRO_MAX;x++) {
    strcpy(macro_strings[x].string,"");
  }
  /* now open file and copy all macro-strings */
  sprintf(tmp,"%s/%s",tnt_dir,func_key_file);
  fp=fopen(tmp,"r");
  if(fp!=0) {
    while(!feof(fp)) {
      fgets(tmp,99,fp);
      if(feof(fp)) break;
      if(tmp[0]=='#') continue;
      tmp[strlen(tmp)-1]='\0';
      strcpy(tmp2,tmp);
      tmp2[2]='\0';
      x=atoi(tmp2);
      if ((x >= MACRO_MIN) && (x <= MACRO_MAX)) {
        strcpy(macro_strings[x].string,&tmp[3]);
      }
    }
    fclose(fp);
    cmd_display(mode,channel,ok_text,1);
    return;
  }
  cmd_display(mode,channel,"Can't load keyboard-macro-file",1);
}
