/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Main procedure (main.c)
   created: Mark Wahl DL4YBG 93/08/07
   updated: Mark Wahl DL4YBG 95/01/09
*/
#define TNT_DATE "95/01/09"
#define TNT_VERSION "V0.9"
#include "tnt.h"

extern int serial;
extern char device[];
extern unsigned int speed;
extern int speedflag;
extern char tnt_logfile[];

int act_channel;
int act_mode;
int act_state;
int is_root;

char signon[] =
"TNT "TNT_VERSION" Packet Radio Hostmode Terminal\015"
#ifdef TNT_LINUX
"LINUX"
#endif
#ifdef TNT_ISC
"ISC"
#endif
"-Version, Created: "TNT_DATE"\015"
"Copyright (C) 1993-95 by Mark Wahl, DL4YBG\015"
"For license details see documentation.\015"
"To get help: <ALT> H or <Esc> H\015";

char add_signon[] =
"No root permissions: SHELL command disabled\015";

char rem_ver_str[] =
"TNT/"
#ifdef TNT_LINUX
"LINUX"
#endif
#ifdef TNT_ISC
"ISC"
#endif
" "TNT_VERSION" ("TNT_DATE", DL4YBG)";


int 
main(argc,argv)
int argc;
char *argv[];
{
  int len;
  char rdbuffer[259];
  char ch;
  int unlock;
#ifdef USE_SELECT
  struct fd_set fdmask;
  struct timeval timevalue;
  int served;
  int max_fd;
  int count;
#else
  int kbd_res;
  int ser_res;
  time_t ser_resp_time;
  int shell_res;
  int empty_rounds;
#ifdef USE_IFACE
  int iface_res;
#endif
#ifndef HAS_USLEEP
  struct timeval timevalue;
#endif
#endif

  umask(0); /* don't filter file-permissions */

  is_root = (geteuid() == 0);
  
  if (read_init_file(argc,argv,&unlock))
    exit(1);
  if (term_init())
    exit(1);
  
  if (init_serial(device,speed,speedflag,unlock))
    exit(1);

  signal(SIGHUP, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGTTIN, SIG_IGN);
  signal(SIGTTOU, SIG_IGN);

  init_priv();
  init_file();
  init_screen();  
  init_keyboard();
  init_help();
  init_xconnect();
  init_mheard();
  init_xmon();
  init_conv();
  init_remote();
  init_cookie();
  init_crc();
  init_shell();
#ifdef USE_IFACE
  init_iface();
  init_mb();
#endif
  init_blist();
  init_log();
  
  if (strcmp(tnt_logfile,"") != 0) {
    open_logfile(RX_NORM,RX_RCV,1,strlen(tnt_logfile),
                 M_CMDSCRIPT,tnt_logfile);
  }
  
#ifdef USE_SELECT

  act_state = S_INIT;
  len = 1;
  serial_server(&act_state,rdbuffer,len);
  
  while (act_state) {
    check_xmon_timeout();
    check_xconnect_timeout();
    FD_ZERO(&fdmask);
    FD_SET(0,&fdmask);
    FD_SET(serial,&fdmask);
    max_fd = serial + 1;
    shell_fdset(&max_fd,&fdmask);
#ifdef USE_IFACE
    iface_fdset(&max_fd,&fdmask);
#endif
    timevalue.tv_usec = 0;
    timevalue.tv_sec = RESY_TIME;
    count = select(max_fd,&fdmask,
                  (struct fd_set *) 0,(struct fd_set *) 0,&timevalue);
    if (count == -1) {
      continue;
    }
    served = 0;
    if (FD_ISSET(0,&fdmask)) {
      if (read(0,&ch,1) == 1) {
        keyboard_server(&act_state,&ch);
      }
      served = 1;
    }
    if (FD_ISSET(serial,&fdmask)) {
      if (len = read(serial,rdbuffer,259)) {
        serial_server(&act_state,rdbuffer,len);
      }
      served = 1;
    }
    if (!shell_receive(&fdmask)) {
      served = 1;
    }
#ifdef USE_IFACE
    if (!iface_receive(&fdmask)) {
      served = 1;
    }
#endif 
    if (!served) {
      resync_tnc(&act_state);
    }
  }

#else

  act_state = S_INIT;
  ser_resp_time = time(NULL);
  len = 1;
  empty_rounds = 0;
  serial_server(&act_state,rdbuffer,len);

  while (act_state) {
    ser_res = 1;
    if (len = read(serial,rdbuffer,259)) {
      serial_server(&act_state,rdbuffer,len);
      ser_resp_time = time(NULL);
    }
    kbd_res = 1;
    if (read(0,&ch,1)) {
      keyboard_server(&act_state,&ch);
      kbd_res = 0;
    }
    shell_res = shell_receive();
#ifdef USE_IFACE
    iface_res = iface_receive();
#endif
    /* if no serial processing and no key pressed: sleep */
#ifdef USE_IFACE
    if (ser_res && kbd_res && shell_res && iface_res) {
#else
    if (ser_res && kbd_res && shell_res) {
#endif
      if (empty_rounds > 10) {
        empty_rounds = 0;
#ifdef HAS_USLEEP
        usleep(10000); /* sleep for 10ms */
#else
        timevalue.tv_usec = 10000; /* sleep for 10ms */
        timevalue.tv_sec = 0;
        select(0,(struct fd_set *) 0,
                 (struct fd_set *) 0,
                 (struct fd_set *) 0,
                 &timevalue);
#endif
        check_xmon_timeout();
        check_xconnect_timeout();
      }
      else {
        empty_rounds++;
      }
      if ((time(NULL) - ser_resp_time) >= RESY_TIME) {
        resync_tnc(&act_state);
        ser_resp_time = time(NULL);
      }
    }
  }
  
#endif  

  exit_log();
  exit_blist();
#ifdef USE_IFACE
  exit_mb();
  exit_iface();
#endif
  exit_shell();
  exit_file();
  exit_priv();
  exit_xmon();
  exit_mheard();
  exit_xconnect();
  exit_help();
  exit_keyboard();
  exit_screen();
  if (exit_serial())
    exit(1);

  exit(0);
}
