/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for remote passwords (priv.c)
   created: Mark Wahl DL4YBG 94/12/21
   updated: Mark Wahl DL4YBG 95/01/04
*/

#include "tnt.h"
#include "priv.h"

extern char ok_text[];
extern char tnt_dir[];
extern char rem_newlin_str[];
extern struct channel_stat ch_stat[CHANNELS];

char tnt_pwfile[80];

static struct password_stat pw_stat[CHANNELS];
static struct calllist *calllist_root;

static struct pwmodes pwmodes[] = {
  {"DIEBOX", PW_DIEBOX},
  {"",PW_NONE}
};


/* analyse one line of password file */
static int pwline_analyse(str1,str2,str3,calllist_ptr)
char *str1;
char *str2;
char *str3;
struct calllist **calllist_ptr;
{
  struct calllist *calllist_wrk;
  int found;
  int i;
  int pwmode;
  
  if (strlen(str1) > 9) return(1); /* callsign maximum 9 characters */
  i = 0;
  found = 0;
  while ((!found) && (pwmodes[i].pwtype != NULL)) {
    if (strcasecmp(str2,pwmodes[i].pwtype) == 0) {
      found = 1;
      pwmode = pwmodes[i].pwmode;
    }
    i++;
  }
  if (!found) return(1);
  calllist_wrk = (struct calllist *)malloc(sizeof(struct calllist));
  for (i=0;i<strlen(str1);i++) {
    calllist_wrk->callsign[i] = toupper(str1[i]);
  }
  calllist_wrk->callsign[i] = '\0';
  calllist_wrk->pwmode = pwmode;
  strcpy(calllist_wrk->pw_file,str3);
  calllist_wrk->next = NULL;
  if (*calllist_ptr == NULL) {
    calllist_root = calllist_wrk;
    *calllist_ptr = calllist_wrk;
  }
  else {
    (*calllist_ptr)->next = calllist_wrk;
    *calllist_ptr = calllist_wrk;
  }
  return(0);
} 

/* load password file */
static void load_pwfile()
{
  int file_end;
  int file_corrupt;
  char file_str[160];
  char line[83];
  char str1[83];
  char str2[83];
  char str3[83];
  int rslt;
  FILE *pw_file_fp;
  struct calllist *calllist_cur;

  strcpy(file_str,tnt_dir);
  strcat(file_str,tnt_pwfile);
  if (!(pw_file_fp = fopen(file_str,"r"))) {
    /* no file present, exit */
    return;
  }
  file_end = 0;
  file_corrupt = 0;
  calllist_cur = NULL;
  while(!file_end) {
    if (fgets(line,82,pw_file_fp) == NULL) {
      file_end = 1;
    }
    else {
      if (strlen(line) == 82) {
        file_corrupt = 1;
        file_end = 1;
      }
      else {
        if (line[0] != '#') {
          rslt = sscanf(line,"%s %s %s",str1,str2,str3);
          switch (rslt) {
          case EOF:
            break;
          case 3:
            if (pwline_analyse(str1,str2,str3,&calllist_cur)) {
              file_corrupt = 1;
              file_end = 1;
            }
            break;
          default:
            file_corrupt = 1;
            file_end = 1;
            break;
          }
        }
      }
    }
  }
  fclose(pw_file_fp);
  if (file_corrupt) {
    if (line == NULL) line[0] = '\0';
    printf("WARNING: pwfile is in wrong format, wrong line:\n%s\n\n",line);
    sleep(5);
    return;
  }
}

static void clear_pwfile()
{
  struct calllist *calllist_wrk;
  struct calllist *calllist_tmp;
  
  calllist_wrk = calllist_root;
  while (calllist_wrk != NULL) {
    calllist_tmp = calllist_wrk;
    calllist_wrk = calllist_tmp->next;
    free(calllist_tmp);
  }
  calllist_root = NULL;
}

/* the callsign has changed on channel, check if callsign in calllist
   and if found set pwmode */
void set_pwmode(channel)
int channel;
{
  char callsign[10];
  struct calllist *calllist_wrk;
  int found;
  
  strip_call_log(callsign,channel);
  pw_stat[channel].pwmode = PW_NONE;
  calllist_wrk = calllist_root;
  found = 0;
  while ((!found) && (calllist_wrk != NULL)) {
    if (strcmp(callsign,calllist_wrk->callsign) == 0) {
      found = 1;
      pw_stat[channel].pwmode = calllist_wrk->pwmode;
      pw_stat[channel].entry = calllist_wrk;
      pw_stat[channel].box_login[0] = '\0';
    }
    else {
      calllist_wrk = calllist_wrk->next;
    }
  }
}

/* clear password information on disconnect */
void clear_pwmode(channel)
int channel;
{
  pw_stat[channel].pwmode = PW_NONE;
  pw_stat[channel].entry = NULL;
  pw_stat[channel].box_login[0] = '\0';
}

/* try to find login time for DieBox */
void scan_login_received(channel,buffer)
int channel;
char *buffer;
{
  char *box_logpos;
  char buffer_tmp[257];
  
  if ((ch_stat[channel].state >= 4) && 
      (pw_stat[channel].pwmode == PW_DIEBOX) &&
      (pw_stat[channel].box_login[0] == '\0')) {
    strncpy(buffer_tmp,buffer+1,(*buffer)+1);
    buffer_tmp[(*buffer)+1] = '\0';
    box_logpos = strstr(buffer_tmp,"Login: ");
    if (box_logpos != NULL) {
      strncpy(pw_stat[channel].box_login,box_logpos+7,14);
      pw_stat[channel].box_login[14] = '\0';
    }
  }
}

/* password generation for DieBox */
/**************************************************/
/* Command PRIV  (for THEBOX Remote Sysops) DK3NY */
/**************************************************/
static void priv_diebox(channel,mode)
int channel;
int mode;
{
  char priv[5];
  char privstr[15];
  char passfile[80];
  FILE *fd;
  int ofs, min, day, hour;
  int len;

  if(pw_stat[channel].box_login[0] != '\0') /* if logintime recorded */
  {
    if (pw_stat[channel].entry->pw_file[0] == '/') {
      strcpy(passfile,pw_stat[channel].entry->pw_file);
    }
    else {
      strcpy(passfile,tnt_dir);
      strcat(passfile,pw_stat[channel].entry->pw_file);
    }
    if(fd = fopen(passfile,"r"))      /* open box_passfile */
    {
      /* Get values of day, hour and minute as integers */
      /* Format of box_login string is "22.07.94 16:23" */
      day  = (pw_stat[channel].box_login[0]-'0') * 10;
      day += (pw_stat[channel].box_login[1]-'0');
      hour = (pw_stat[channel].box_login[9]-'0') * 10;
      hour+= (pw_stat[channel].box_login[10]-'0');
      min  = (pw_stat[channel].box_login[12]-'0') * 10;
      min += (pw_stat[channel].box_login[13]-'0');

      ofs = min + day;                /* calculate offset for box_passfile */
      if(ofs > 59)
        ofs = ofs - 60;
      ofs = ofs * 29 + hour;
      fseek(fd,ofs,0);                /* seek to valid priv position */
      fscanf(fd,"%4s",priv);          /* get the 4 priv characters */
      fclose(fd);
      sprintf(privstr,"PRIV %s%s",priv,rem_newlin_str);
      len = strlen(privstr);
      rem_data_display(channel,privstr);
      queue_cmd_data(channel,X_DATA,len,mode,privstr);
    }
    else
      cmd_display(mode,channel,"Sorry, passwordfile not found",1);
  }
  else {
    cmd_display(mode,channel,"Sorry, no Logintime recorded",1);
  }
}

/* reload password file */
void cmd_loadpriv(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  clear_pwfile();  
  load_pwfile();
  cmd_display(mode,channel,ok_text,1);
  return;
}

/* list password file data */
void cmd_listpriv(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  struct calllist *calllist_wrk;
  int i;
  int found;
  char pwtype[20];
  char disp_line[83];
  
  if (calllist_root == NULL) {
    cmd_display(mode,channel,"No lines found",1);
    return;
  }
  calllist_wrk = calllist_root;
  while (calllist_wrk != NULL) {
    i = 0;
    found = 0;
    while ((!found) && (pwmodes[i].pwtype != NULL)) {
      if (pwmodes[i].pwmode == calllist_wrk->pwmode) {
        found = 1;
        strcpy(pwtype,pwmodes[i].pwtype);
      }
      i++;
    }
    if (!found) {
      cmd_display(mode,channel,"Calllist corrupt!",1);
      return;
    }
    sprintf(disp_line,"%s %s %s",
            calllist_wrk->callsign, pwtype, calllist_wrk->pw_file);
    cmd_display(mode,channel,disp_line,1);
    calllist_wrk = calllist_wrk->next;
  }
}

/* generate password */
void cmd_priv(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
  if (ch_stat[channel].state < 4) {
    cmd_display(mode,channel,"Only while connected",1);
    return;
  }
  switch (pw_stat[channel].pwmode) {
  case PW_DIEBOX:
    priv_diebox(channel,mode);
    return;
  default:
    cmd_display(mode,channel,"No password available",1);
    return;
  }
}

/* init of priv data */
void init_priv()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    clear_pwmode(i);
  }
  calllist_root = NULL;
  load_pwfile();
}

/* exit of priv data */
void exit_priv()
{
  clear_pwfile();
}

