/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993-1995 by Mark Wahl
   For license details see documentation
   Procedures for shell-login (shell.c)
   created: Mark Wahl DL4YBG 94/01/22
   updated: Mark Wahl DL4YBG 95/01/03
*/

#include "tnt.h"
#include "shell.h"
#include <pwd.h>
#include <utmp.h>

#ifdef USE_SOCKET
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#ifdef TNT_CHAMBER
#define BINTRX "/tmp"
#ifndef _NFILE
#define _NFILE FD_SETSIZE
#endif
#endif

#define BUFFER_LEN 256

struct shell_stat sh_stat[CHANNELS];
extern struct channel_stat ch_stat[CHANNELS];
extern char remote_user[];
extern char run_dir[];
extern char tnt_dir[];
extern int ptyecho_flag;
int pty_timeout;
extern char rem_tnt_str[];
extern is_root;

static char rem_not_conn_txt[] = "Only while connected";
static char only_root_txt[] = "Only available if TNT is started as root";
static char rem_only_root_txt[] = "Not available";

#ifdef USE_SOCKET
static struct sockaddr_in cli_addr;
static struct sockaddr_in serv_addr;
static int clilen;
struct sock_data sd[CHANNELS];
extern char sock_passfile[];
#endif

#ifdef GEN_NEW_USER
int unix_new_user;
char unix_user_dir[80];
int unix_first_uid;
int unix_user_gid;
#endif

#ifdef USE_SOCKET
char sh_active_txt[] = "Shell, redirection or socket already active";
#else
char sh_active_txt[] = "Shell or redirection already active";
#endif
#ifdef USE_IFACE
extern char iface_active_txt[];
#endif

static long secclock()
{
  struct timeval tv;
  struct timezone tz;
  
  gettimeofday(&tv,&tz);
  return(tv.tv_sec);
}

static int find_pty(numptr, slave)
int *numptr;
char *slave;
{

  char master[80];
  int fd;
  int num;

  for (num = 0; num < NUMPTY; num++) {
    pty_name(master, MASTERPREFIX, num);
#ifdef USE_SELECT
    if ((fd = open(master, O_RDWR, 0600)) >= 0) {
#else
    if ((fd = open(master, O_RDWR | O_NONBLOCK, 0600)) >= 0) {
#endif
      *numptr = num;
      pty_name(slave, SLAVEPREFIX, num);
      return fd;
    }
  }
  return (-1);
}

static void restore_pty(id)
const char *id;
{
  char filename[80];

  sprintf(filename, "%s%s", MASTERPREFIX, id);
  chown(filename, 0, 0);
  chmod(filename, 0666);
  sprintf(filename, "%s%s", SLAVEPREFIX, id);
  chown(filename, 0, 0);
  chmod(filename, 0666);
}

void fixutmpfile()
{

#ifdef USER_PROCESS

  struct utmp *up;

  while (up = getutent())
    if (up->ut_type == USER_PROCESS && kill(up->ut_pid, 0)) {
      restore_pty(up->ut_id);
      up->ut_user[0] = 0;
      up->ut_type = DEAD_PROCESS;
#ifdef HAS_UTEXIT
      up->ut_exit.e_termination = 0;
      up->ut_exit.e_exit = 0;
#endif
      up->ut_time = secclock();
      pututline(up);
    }
  endutent();

#endif
}

static int callvalid(call)
const char *call;
{
  int d, l;

  l = strlen(call);
  if (l < 3 || l > 6) return 0;
  if (isdigit(uchar(call[0])) && isdigit(uchar(call[1]))) return 0;
  if (!(isdigit(uchar(call[1])) || isdigit(uchar(call[2])))) return 0;
  if (!isalpha(uchar(call[l-1]))) return 0;
  d = 0;
  for (; *call; call++) {
    if (!isalnum(uchar(*call))) return 0;
    if (isdigit(uchar(*call))) d++;
  }
  if (d < 1 || d > 2) return 0;
  return 1;
}

static char *find_user_name(name)
const char *name;
{

  char *cp;
  static char username[128];

  for (; ; ) {
    while (*name && !isalnum(uchar(*name))) name++;
    for (cp = username; isalnum(uchar(*name)); *cp++ = tolower(uchar(*name++))) ;
    *cp = 0;
    if (!*username) return remote_user;
    if (callvalid(username)) return username;
  }
}

struct passwd *getpasswdentry(name, create)
const char *name;
int create;
{
#ifdef GEN_NEW_USER
  FILE *fp;
  char bitmap[MAXUID+1];
  char homedir[80];
  char homedirparent[80];
  int fd;
  int uid;
#endif
  struct passwd *pw;

  /* Search existing passwd entry */

  if (pw = getpwnam(name)) return pw;
#ifdef GEN_NEW_USER
  if (!create) return 0;

  /* Find free user id */

  if ((fd = open(PWLOCKFILE, O_WRONLY | O_CREAT | O_EXCL, 0644)) < 0) return 0;
  close(fd);
  memset(bitmap, 0, sizeof(bitmap));
  while (pw = getpwent()) {
    if (!strcmp(name, pw->pw_name)) break;
    if (pw->pw_uid <= MAXUID) bitmap[pw->pw_uid] = 1;
  }
  endpwent();
  if (pw) {
    unlink(PWLOCKFILE);
    return pw;
  }
  for (uid = unix_first_uid; uid <= MAXUID && bitmap[uid]; uid++) ;
  if (uid > MAXUID) {
    unlink(PWLOCKFILE);
    return 0;
  }

  /* Add user to passwd file */

  sprintf(homedirparent, "%s%.3s...", unix_user_dir, name);
  sprintf(homedir, "%s/%s", homedirparent, name);
  if (!(fp = fopen(PASSWDFILE, "a"))) {
    unlink(PWLOCKFILE);
    return 0;
  }
#ifdef PWD_NOT_EMPTY
  fprintf(fp, "%s:,./:%d:%d::%s:\n", name, uid, unix_user_gid, homedir);
#else
  fprintf(fp, "%s::%d:%d::%s:\n", name, uid, unix_user_gid, homedir);
#endif
  fclose(fp);
  pw = getpwuid(uid);
  unlink(PWLOCKFILE);

  /* Create home directory */

  mkdir(homedirparent, 0755);
  mkdir(homedir, 0755);
  chown(homedir, uid, unix_user_gid);
  return pw;
#else
  else return 0;
#endif
}

#ifdef TNT_CHAMBER
void disc_frontend(channel)
int channel;
{
  int flag;
  char ans_str[10];
  int ans_len;
  
  flag = 0;
  strcpy(ans_str,"D");
  ans_len = strlen(ans_str);
  queue_cmd_data(channel,X_COMM,ans_len,M_CMDSCRIPT,ans_str);
}
#endif

/* open a shell on the current channel */
void cmd_shell(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char *env = 0;
  char slave[80];
  int i;
  struct passwd *pw;
  struct termios termios;
  struct utmp utmpbuf;
  char envcall[30];

  if (!is_root) {
    if (mode != M_REMOTE)
      cmd_display(mode,channel,only_root_txt,1);
    else
      cmd_display(mode,channel,rem_only_root_txt,1);
    return;
  }
#ifdef USE_IFACE
  if (iface_active(channel)) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
#endif
  if (ch_stat[channel].state < 4) {
    cmd_display(mode,channel,rem_not_conn_txt,1);
    return;
  }
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if ((sh_stat[channel].pty = find_pty(&sh_stat[channel].num, slave)) < 0) {
    cmd_display(mode,channel,"Can't open pseudo terminal",1);
    return;
  }
  sh_stat[channel].active = A_SHELL;
  sh_stat[channel].add_flag = F_SHELL;
  sh_stat[channel].mode = mode;
  sh_stat[channel].buflen = 0;
  for (i=0;i < len;i++) {
    str[i] = tolower(str[i]);
  }
  sh_stat[channel].lfcrconv = !(strcmp(str,"trans") == 0);
  strcpy(sh_stat[channel].id, slave + strlen(slave) - 2);
  if (!(sh_stat[channel].pid = fork())) {
#ifndef TNT_CHAMBER
    pw = getpasswdentry(find_user_name(ch_stat[channel].call), unix_new_user);
    if (!pw || pw->pw_passwd[0]) {
      pw = getpasswdentry(remote_user, 0);
      if (!pw || pw->pw_passwd[0]) exit(1);
    }
#endif
    for (i = 0; i < FD_SETSIZE; i++) close(i);
    setsid();
    open(slave, O_RDWR, 0666);
    dup(0);
    dup(0);
    chmod(slave, 0622);
    if (sh_stat[channel].lfcrconv) {
      memset((char *) &termios, 0, sizeof(termios));
      termios.c_iflag = ICRNL | IXOFF;
      termios.c_oflag = OPOST | TAB3 | ONLRET;
      termios.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
      termios.c_cflag = CS8 | CREAD | CLOCAL;
      termios.c_lflag = ISIG | ICANON;
      termios.c_cc[VINTR]  = 127;
      termios.c_cc[VQUIT]  =  28;
      termios.c_cc[VERASE] =   8;
      termios.c_cc[VKILL]  =  24;
      termios.c_cc[VEOF]   =   4;
      cfsetispeed(&termios, B1200);
      cfsetospeed(&termios, B1200);
    }
    else {
      tcgetattr(0,&termios);
    }
    tcsetattr(0, TCSANOW, &termios);
    if (is_root) {
#ifdef LOGIN_PROCESS
      memset(&utmpbuf, 0, sizeof(utmpbuf));
      strcpy(utmpbuf.ut_user, "LOGIN");
      strcpy(utmpbuf.ut_id, sh_stat[channel].id);
      strcpy(utmpbuf.ut_line, slave + 5);
      utmpbuf.ut_pid = getpid();
      utmpbuf.ut_type = LOGIN_PROCESS;
      utmpbuf.ut_time = secclock();
#ifdef HAS_UTHOST 
      strncpy(utmpbuf.ut_host,"AX25", sizeof(utmpbuf.ut_host));
#endif
      pututline(&utmpbuf);
      endutent();
#endif
    }
#ifdef TNT_CHAMBER
    execl("/bin/login.pr", "login.pr", (char *) 0);
#else
    strcpy(envcall,"CALLSIGN=");
    strcat(envcall,find_user_name(ch_stat[channel].call));
    execle("/bin/login", "login", pw->pw_name, envcall, (char *) 0, &env);
#endif
    exit(1);
  }
  cmd_display(mode,channel,"Ok",1);
  return;
}

/* redirect the current channel */
void cmd_redir(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char ans_str[80];

#ifdef USE_IFACE
  if (iface_active(channel)) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
#endif
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
#ifdef USE_SELECT
  if ((sh_stat[channel].pty = open(str,O_RDWR,0600)) < 0) {
#else
  if ((sh_stat[channel].pty = open(str,O_RDWR|O_NONBLOCK,0600)) < 0) {
#endif
    sprintf(ans_str,"Can't redirect to %s",str);
    cmd_display(mode,channel,ans_str,1);
    return;
  }
  sh_stat[channel].active = A_REDIR;
  sh_stat[channel].mode = mode;
  sh_stat[channel].pid = 0;
  sh_stat[channel].lfcrconv = 0;
  sh_stat[channel].buflen = 0;
  sprintf(ans_str,"Redirect to %s",str);
  cmd_display(mode,channel,ans_str,1);
}

/* run a program on the current channel */
void cmd_run(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char *envp[3];
  char envcall[30];
  char logcall[30];
  char slave[80];
  int i;
  struct passwd *pw;
  struct termios termios;
  char command[256];

#ifdef USE_IFACE
  if (iface_active(channel)) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
#endif
  if (ch_stat[channel].state < 4) {
    cmd_display(mode,channel,rem_not_conn_txt,1);
    return;
  }
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if (len == 0) {
    rem_dir(1,1,channel,0,mode,"");
    return;
  }
  if ((sh_stat[channel].pty = find_pty(&sh_stat[channel].num, slave)) < 0) {
    cmd_display(mode,channel,"Can't open pseudo terminal",1);
    return;
  }
  sh_stat[channel].active = A_SHELL;
  sh_stat[channel].add_flag = F_RUN;
  sh_stat[channel].mode = mode;
  sh_stat[channel].buflen = 0;
  for (i=0;i < len;i++) {
    str[i] = tolower(str[i]);
  }
  sh_stat[channel].lfcrconv = (par1 == 0);
  strcpy(sh_stat[channel].id, slave + strlen(slave) - 2);
  if (!(sh_stat[channel].pid = fork())) {
    for (i = 0; i < FD_SETSIZE; i++) close(i);
    setsid();
    if (is_root) {
      pw = getpwnam(remote_user);
      if (pw == NULL) exit(1);
      if (setgid(pw->pw_gid) == -1) exit(1);
      if (setuid(pw->pw_uid) == -1) exit(1);
    }
    open(slave, O_RDWR, 0666);
    dup(0);
    dup(0);
    chmod(slave, 0622);
    if (sh_stat[channel].lfcrconv) {
      memset((char *) &termios, 0, sizeof(termios));
      termios.c_iflag = ICRNL | IXOFF;
      termios.c_oflag = OPOST | TAB3 | ONLRET;
      termios.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
      termios.c_cflag = CS8 | CREAD | CLOCAL;
      termios.c_lflag = ISIG | ICANON;
      termios.c_cc[VINTR]  = 127;
      termios.c_cc[VQUIT]  =  28;
      termios.c_cc[VERASE] =   8;
      termios.c_cc[VKILL]  =  24;
      termios.c_cc[VEOF]   =   4;
      cfsetispeed(&termios, B1200);
      cfsetospeed(&termios, B1200);
    }
    else {
      tcgetattr(0,&termios);
    }
    tcsetattr(0, TCSANOW, &termios);
    /* execute command using shell */
    if (is_root)
      chdir(pw->pw_dir);
    strcpy(command,run_dir);
    strcat(command,str);
    strcpy(envcall,"CALLSIGN=");
    strcat(envcall,find_user_name(ch_stat[channel].call));
    strcpy(logcall,"LOGNAME=");
    strcat(logcall,find_user_name(ch_stat[channel].call));
    envp[0] = envcall;
    envp[1] = logcall;
    envp[2] = NULL;
    execle("/bin/sh","sh","-c",command, (char *) 0, envp);
    exit(1);
  }
  cmd_display(mode,channel,"Ok",1);
  return;
}

#ifdef USE_SOCKET
/* redirect the current channel to a socket */
void cmd_socket(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char ans_str[80];
  int port;

#ifdef USE_IFACE
  if (iface_active(channel)) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
#endif
  if (sscanf(str,"%d",&port) != 1) {
    cmd_display(mode,channel,"Invalid port",1);
    return;
  }
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if ((sh_stat[channel].sockfd = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    cmd_display(mode,channel,"Can't open stream socket",1);
    return;
  }
#ifndef USE_SELECT  
  fcntl(sh_stat[channel].sockfd,F_SETFL,O_NONBLOCK);
#endif
  bzero((char *) &serv_addr,sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);
  if (bind(sh_stat[channel].sockfd, (struct sockaddr *) &serv_addr,
           sizeof (serv_addr)) < 0) {
    sprintf(ans_str,"Can't bind port %d",port);
    cmd_display(mode,channel,ans_str,1);
    close(sh_stat[channel].sockfd);
    return;
  }
  listen(sh_stat[channel].sockfd,5);

  sh_stat[channel].active = A_SOCKET;
  sh_stat[channel].mode = mode;
  sh_stat[channel].pid = 0;
  sh_stat[channel].lfcrconv = 1;
  sh_stat[channel].buflen = 0;
  sh_stat[channel].connected = 0;
  sd[channel].login_stat = 0;
  sprintf(ans_str,"Listening at port %d",port);
  cmd_display(mode,channel,ans_str,1);
}
#endif

void dat_input_pty(channel,buffer,len)
int channel;
char *buffer;
int len;
{
#ifdef USE_SOCKET
  if (sh_stat[channel].active == A_SOCKET) {
    if(!do_sock_cmd(channel,buffer,len)) {
      if (ptyecho_flag) rem_data_display_buf(channel,buffer,len);
      dat_input(channel,buffer,len);
    }
  }
  else {
    if (ptyecho_flag) rem_data_display_buf(channel,buffer,len);
    dat_input(channel,buffer,len);
  }
#else
  if (ptyecho_flag) rem_data_display_buf(channel,buffer,len);
  dat_input(channel,buffer,len);
#endif
}

void pty_send_data(channel,buffer,len)
int channel;
char *buffer;
int len;
{
  int tmp_len;
  
  if (len > PACKETSIZE) return; /* sanity check */
  if ((sh_stat[channel].buflen == 0) && (len == PACKETSIZE)) {
    dat_input_pty(channel,buffer,len);
  }
  else if ((tmp_len = sh_stat[channel].buflen + len) > PACKETSIZE) {
    if (sh_stat[channel].buflen) {
      dat_input_pty(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
    }
    memcpy(sh_stat[channel].buffer,buffer,len);
    sh_stat[channel].buflen = len;
    sh_stat[channel].bufupdate = time(NULL);
  }
  else {
    memcpy((sh_stat[channel].buffer + sh_stat[channel].buflen),
           buffer,len);
    sh_stat[channel].buflen = tmp_len;       
    sh_stat[channel].bufupdate = time(NULL);
  }
}

void pty_check_timeout(channel)
int channel;
{
#ifdef SPECIAL
  int flag = 0;
#endif

  if (!sh_stat[channel].buflen) return;
#ifdef USE_SOCKET
#ifdef SPECIAL
  if (sh_stat[channel].buffer[sh_stat[channel].buflen-1]=='\r' ||
      sh_stat[channel].buffer[sh_stat[channel].buflen-1]=='\n')
    flag=1;
#endif
#endif
#ifdef SPECIAL
  if (((time(NULL) - sh_stat[channel].bufupdate) >= pty_timeout) || flag) {
  flag = 0;
#else
  if ((time(NULL) - sh_stat[channel].bufupdate) >= pty_timeout) {
#endif
    dat_input_pty(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
    sh_stat[channel].buflen = 0;
  }
}

static int read_data_pty(channel,buffer)
int channel;
char buffer[];
{
  int len;
#ifdef READ_NONBLOCK
  int fc;
#endif
#ifdef TNT_CHAMBER
  char buffer2[BUFFER_LEN+2];
  char tmp[100];
  static long xx;
  FILE *fp;
#endif

#ifdef USE_SOCKET
  if ((sh_stat[channel].active == A_SOCKET) &&
      (!sh_stat[channel].connected)) return(0);
#endif
#ifdef READ_NONBLOCK
  fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
#endif
  len = read(sh_stat[channel].pty,buffer,BUFFER_LEN);
#ifdef READ_NONBLOCK
  fc=fcntl(sh_stat[channel].pty,F_GETFL,NULL);
  fcntl(sh_stat[channel].pty,F_SETFL,fc & ~O_NONBLOCK);
#endif
  if (len > 0) {
#ifdef TNT_CHAMBER
    if ((sh_stat[channel].active == A_FRONTEND) ||
        (sh_stat[channel].active == A_SOCKET)) {
      sprintf(tmp,"%s/.bin.%d.%d",BINTRX,getpid(),channel);
      if(access(tmp,0)!=0) {
        len=conv_lftonil(buffer,len);
        xx=0L;
        buffer2[0] = len-1;
        memcpy(&buffer2[1],buffer,len);
        rem_data_display_len(channel,buffer2);
      }
      else {
        xx+=(long)len;
        sprintf(buffer2,"RBIN-Frontend: %ld",xx);
        rem_stat_display(channel,buffer2);
      }
    }
    else {
      if (sh_stat[channel].lfcrconv) {
        conv_lftocr(buffer,len);
      }
    }
#else
#ifdef USE_SOCKET
    if (sh_stat[channel].active == A_SOCKET) {
      len=conv_lftonil(buffer,len);
    }
    else if (sh_stat[channel].lfcrconv) {
      conv_lftocr(buffer,len);
    }
#else
    if (sh_stat[channel].lfcrconv) {
      conv_lftocr(buffer,len);
    }
#endif
#endif
    pty_send_data(channel,buffer,len);
#ifdef TNT_CHAMBER
    if((sh_stat[channel].active==A_FRONTEND) && (access(tmp,0)!=0)) {
      fp=fopen("/pr/log/FRONTEND.log","a");
      if(fp!=0)  {
        strncpy(buffer2,buffer,len);
        buffer2[strlen(buffer2)]=0;
        conv_crtolf(buffer2,len);
        fprintf(fp,"%s",buffer2);
        fclose(fp);
      }
    }
#endif
    return(1);
  }
  else {
#ifdef USE_SOCKET  
    if ((len == 0) && (sh_stat[channel].active == A_SOCKET)){
      if (sh_stat[channel].connected) {
        close(sh_stat[channel].pty);
        sh_stat[channel].connected = 0;
        sd[channel].login_stat = 0;
      }
    }
#endif
    return(0);
  }
}

void close_shell(channel,report)
int channel;
int report;
{

#ifndef UTMP_FILE
#define UTMP_FILE       "/etc/utmp"
#endif

#ifndef WTMP_FILE
#define WTMP_FILE       "/var/adm/wtmp"
#endif

  char slave[80];
  int fdut = -1;
  struct utmp utmpbuf;
  char ans_str[80];
  int result;
  char buffer[BUFFER_LEN];
#ifndef USE_SELECT
  int nodata;
#endif

  if (!sh_stat[channel].active) return;
#ifdef USE_IFACE
  if (sh_stat[channel].active == A_IFACE) {
    sh_stat[channel].active = 0;
    cmd_display(sh_stat[channel].mode,channel,"Interface closed",1);
    return;
  }
#endif
#ifdef USE_SOCKET
  if (sh_stat[channel].active == A_SOCKET) {
    while (read_data_pty(channel,buffer) == 1);
    return;
  }
#endif
  /* read all remaining data from pty */
#ifdef USE_SELECT
  fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
#endif
  while (read_data_pty(channel,buffer) == 1);
  if (sh_stat[channel].buflen) {
    if (ptyecho_flag)
      rem_data_display_buf(channel,sh_stat[channel].buffer,
                                   sh_stat[channel].buflen);
    dat_input(channel,sh_stat[channel].buffer,sh_stat[channel].buflen);
  }
  if (sh_stat[channel].pty > 0) {
    ioctl(sh_stat[channel].pty, TCFLSH, 2);
    close(sh_stat[channel].pty);
#ifdef TNT_CHAMBER
    if ((sh_stat[channel].active == A_SHELL) ||
       (sh_stat[channel].active == A_FRONTEND))
#else
    if (sh_stat[channel].active == A_SHELL)
#endif
      restore_pty(sh_stat[channel].id);
  }
#ifdef TNT_CHAMBER
  if (((sh_stat[channel].active == A_SHELL) ||
       (sh_stat[channel].active == A_FRONTEND)) &&
      (sh_stat[channel].pid > 0)) {
#else
  if ((sh_stat[channel].active == A_SHELL) && (sh_stat[channel].pid > 0)) {
#endif
    kill(-sh_stat[channel].pid, SIGHUP);
    pty_name(slave, SLAVEPREFIX, sh_stat[channel].num);
#ifdef TNT_CHAMBER
    if (sh_stat[channel].active == A_FRONTEND) {
      char tmp[100];
      sprintf(tmp,"%s/.bin.%d.%d",BINTRX,getpid(),channel);
      if(access(tmp,0)==0) unlink(tmp);
      disc_frontend(channel);
    }
#endif
    if ((fdut = open(UTMP_FILE, O_RDWR, 0644)) >= 0) {
      while (read(fdut, &utmpbuf, sizeof(utmpbuf)) == sizeof(utmpbuf)) {
	if (!strncmp(utmpbuf.ut_line, slave + 5, sizeof(utmpbuf.ut_line))) {
	  utmpbuf.ut_name[0] = 0;
#ifdef HAS_UTHOST	  
	  utmpbuf.ut_host[0] = 0;
#endif
	  utmpbuf.ut_time = secclock();
#ifdef DEAD_PROCESS
	  utmpbuf.ut_type = DEAD_PROCESS;
#ifdef HAS_UTEXIT
	  utmpbuf.ut_exit.e_termination = 0;
	  utmpbuf.ut_exit.e_exit = 0;
#endif
#endif
	  lseek(fdut, -sizeof(utmpbuf), SEEK_CUR);
	  write(fdut, &utmpbuf, sizeof(utmpbuf));
	  close(fdut);
	  if ((fdut = open(WTMP_FILE, O_WRONLY | O_CREAT | O_APPEND, 0644)) >= 0)
	    write(fdut, &utmpbuf, sizeof(utmpbuf));
	  break;
	}
      }	
    }
    if (fdut >= 0) close(fdut);
  }
  if (report) {
    switch (sh_stat[channel].active) {
    case A_SHELL:
      switch (sh_stat[channel].add_flag) {
      case F_SHELL:
        strcpy(ans_str,"Shell closed");
        break;
      case F_RUN:
        strcpy(ans_str,"Program finished");
        break;
      }
      break;
    case A_REDIR:
      strcpy(ans_str,"Redirection closed");
      break;
#ifdef TNT_CHAMBER
    case A_FRONTEND:
      strcpy(ans_str,"Frontend closed");
      break;
#endif
    }
    cmd_display(sh_stat[channel].mode,channel,ans_str,1);
  }
  sh_stat[channel].active = 0;
}

#ifdef USE_SOCKET
static void close_socket(channel,report)
int channel;
int report;
{
  char ans_str[80];

  if (!sh_stat[channel].active) return;
  if (sh_stat[channel].active != A_SOCKET) return;
  if (sh_stat[channel].connected) {
    close(sh_stat[channel].pty);
  }
  close(sh_stat[channel].sockfd);
  if (report) {
    strcpy(ans_str,"Socket closed");
    cmd_display(sh_stat[channel].mode,channel,ans_str,1);
  }
  sh_stat[channel].active = 0;
}
#endif

/* close a shell or redir on the current channel */
void cmd_endshell(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{
#ifdef USE_SOCKET
  if (sh_stat[channel].active == A_SOCKET) {
    close_socket(channel,1);
  }
#endif
  close_shell(channel,1);
}

int write_pty(channel,len,str)
int channel;
int len;
char *str;
{
  char buffer[300];
#ifdef TNT_CHAMBER
  FILE *fp;
  int i;
#endif

  if (!sh_stat[channel].active) return(0);
  memcpy(buffer,str,len);
  if (sh_stat[channel].lfcrconv) {
    conv_crtolf(buffer,len);
  }
  if (write(sh_stat[channel].pty,buffer,len) < len) {
    close_shell(channel,1);
  }
#ifdef TNT_CHAMBER
  if (sh_stat[channel].active==A_FRONTEND ||
      sh_stat[channel].active == A_SOCKET) {
    
    if (sh_stat[channel].active==A_FRONTEND) {
      fp=fopen("/pr/log/FRONTEND.log","a");
      if (fp!=0) {
	for(i=0;i<len;i++) {
	  if(str[i]==CR) str[i]=LF;
	  if(isprint(str[i]))
	    fprintf(fp,"%c",str[i]);
	}
	fprintf(fp,"\n");
	fclose(fp);
      }
    }
  }
#endif
  return(1);
}

#ifdef USE_SELECT
void shell_fdset(max_fd,fdmask)
int *max_fd;
struct fd_set *fdmask;
{
  int channel;
  int fd;
  
  for (channel=0;channel<CHANNELS;channel++) {
    if ((sh_stat[channel].active) && senddata_allowed(channel)) {
#ifdef USE_SOCKET
      if ((sh_stat[channel].active == A_SOCKET) &&
          (!sh_stat[channel].connected)) {
        fd = sh_stat[channel].sockfd;
      }
      else {
        fd = sh_stat[channel].pty;
      }
#else
      fd = sh_stat[channel].pty;
#endif
      FD_SET(fd,fdmask);
      if (fd > ((*max_fd) - 1)) {
        *max_fd = fd + 1;
      }
    }
  }
}

int shell_receive(fdmask)
struct fd_set *fdmask;
{
  int channel;
  int result;
  char buffer[BUFFER_LEN];
  int len;
  int status;
  pid_t pid;
#ifdef USE_SOCKET
  char tmp[100];
#endif  
  
  result = 1;
  /* now handle data */
  for (channel=0;channel<CHANNELS;channel++) {
    if (sh_stat[channel].active) {
      pty_check_timeout(channel);
#ifdef USE_SOCKET
      if ((sh_stat[channel].active == A_SOCKET) &&
          (!sh_stat[channel].connected)) {
        if (FD_ISSET(sh_stat[channel].sockfd,fdmask)) {
          if ((sh_stat[channel].pty = accept(sh_stat[channel].sockfd,
              (struct sockaddr *) &cli_addr, &clilen)) < 0) {
          }
          else {
	    sd[channel].login_stat=0;
	    sprintf(tmp,"\nWelcome to %s remote AX25-server (Ch: %d).\n",
		    ch_stat[channel].curcall,channel);
	    strcat(tmp,"\nYou have to login to establish access permission\n");
	    strcat(tmp,"to remote ham-radio AX25-server via ip-network.\n\n");
	    strcat(tmp,"login: ");
	    write_pty(channel,strlen(tmp),tmp);
            sh_stat[channel].connected = 1;
            result = 0;
          }
        }
        continue;
      }
#endif
      if (FD_ISSET(sh_stat[channel].pty,fdmask)) {
        read_data_pty(channel,buffer);
        result = 0;
      }
    }
  }
  /* clean up all remaining zombies */
  while ((pid = waitpid(-1,&status,WNOHANG)) > 0) {
    for (channel=0;channel<CHANNELS;channel++) {
      if (sh_stat[channel].active) {
        if (sh_stat[channel].pid == pid) {
          close_shell(channel,1);
        }
      }
    }
  }
  return (result);
}
#else
int shell_receive()
{
  int channel;
  int result;
  char buffer[BUFFER_LEN];
  int len;
  int nodata;
  int status;
  pid_t pid;
#ifdef USE_SOCKET
  char tmp[100];
#endif  
  
  result = 1;
  /* now handle data */
  for (channel=0;channel<CHANNELS;channel++) {
    nodata = 0;
    if (sh_stat[channel].active) {
      pty_check_timeout(channel);
    }
#ifdef USE_SOCKET
    if ((sh_stat[channel].active == A_SOCKET) &&
        (!sh_stat[channel].connected)) {
      if ((sh_stat[channel].pty = accept(sh_stat[channel].sockfd,
              (struct sockaddr *) &cli_addr, &clilen)) < 0) {
        nodata = 1;
      }
      else {
        fcntl(sh_stat[channel].pty,F_SETFL,O_NONBLOCK);
        sd[channel].login_stat=0;
        sprintf(tmp,"\nWelcome to %s remote AX25-server.\n",
	        ch_stat[channel].curcall);
        strcat(tmp,"\nYou have to login to establish access permission\n");
        strcat(tmp,"to remote ham-radio AX25-server.\n\n");
        strcat(tmp,"login: ");
        write_pty(channel,strlen(tmp),tmp);
        sh_stat[channel].connected = 1;
        result = 0;
      }
      continue;
    }
#endif    
    while (sh_stat[channel].active && senddata_allowed(channel) && !nodata) {
      if (read_data_pty(channel,buffer)) {
        result = 0;
      }
      else {
        nodata = 1;
      }
    }
  }
  /* clean up all remaining zombies */
  while ((pid = waitpid(-1,&status,WNOHANG)) > 0) {
    for (channel=0;channel<CHANNELS;channel++) {
      if (sh_stat[channel].active) {
        if (sh_stat[channel].pid == pid) {
          close_shell(channel,1);
        }
      }
    }
  }
  return (result);
}
#endif

void init_shell()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    sh_stat[i].active = 0;
    sh_stat[i].pid = 0;
  }
  if (is_root) fixutmpfile();
}

void exit_shell()
{
  int i;
  
  for (i=0;i<CHANNELS;i++) {
    close_shell(i,0);
#ifdef USE_SOCKET
    close_socket(i,0);
#endif    
  }
}

#ifdef USE_SOCKET
int sock_status_out(channel,str)
int channel;
char *str;
{
  char tmpstring[20];
  
  if (sh_stat[channel].active == A_SOCKET) {
    if (write_pty(channel,strlen(str),str)) {
      if (sd[channel].login_stat==2) {
        strcpy(tmpstring,"\nCmd> ");
        write_pty(channel,strlen(tmpstring),tmpstring);
      }
      return(1);
    }
  }
  return(0);
}
#endif

#ifdef TNT_CHAMBER
/* open a shell on the current channel */
void cmd_frontend(par1,par2,channel,len,mode,str)
int par1;
int par2;
int channel;
int len;
int mode;
char *str;
{

  char *env = 0;
  char slave[80];
  int i;
  char ucall[100],tmp[10];
  struct passwd *pw;
  struct termios termios;
  struct utmp utmpbuf;

#ifdef USE_IFACE
  if (iface_active(channel)) {
    cmd_display(mode,channel,iface_active_txt,1);
    return;
  }
#endif
  strip_call(&ucall,channel);
  if (sh_stat[channel].active) {
    cmd_display(mode,channel,sh_active_txt,1);
    return;
  }
  if ((sh_stat[channel].pty = find_pty(&sh_stat[channel].num, slave)) < 0) {
    cmd_display(mode,channel,"Can't open pseudo terminal",1);
    return;
  }
  if(!strncmp(ucall,"DB0",3)) {
    cmd_display(mode,channel,"Uncorrect connect from Digipeater",1);
    disc_frontend(channel);
    return;
  }

  sh_stat[channel].active = A_FRONTEND;
  sh_stat[channel].mode = mode;
  sh_stat[channel].buflen = 0;
  strcpy(sh_stat[channel].id, slave + strlen(slave) - 2);
  if (!(sh_stat[channel].pid = fork())) {
    for (i = 0; i < _NFILE; i++) close(i);
    setpgrp();
    open(slave, O_RDWR, 0666);
    dup(0);
    dup(0);
    chmod(slave, 0622);
    memset((char *) &termios, 0, sizeof(termios));
    termios.c_iflag = ICRNL | IXOFF;
    termios.c_oflag = OPOST | TAB3 | ONLRET;
    termios.c_oflag &= ~(OCRNL | ONLCR | ONOCR);
    termios.c_cflag = CS8 | CREAD | CLOCAL;
    termios.c_lflag = ISIG | ICANON;
    termios.c_cc[VINTR]  = 127;
    termios.c_cc[VQUIT]  =  28;
    termios.c_cc[VERASE] =   8;
    termios.c_cc[VKILL]  =  24;
    termios.c_cc[VEOF]   =   4;
    cfsetispeed(&termios, B1200);
    cfsetospeed(&termios, B1200);
    tcsetattr(0, TCSANOW, &termios);
#ifdef LOGIN_PROCESS
    memset(&utmpbuf, 0, sizeof(utmpbuf));
    strcpy(utmpbuf.ut_user, "LOGIN");
    strcpy(utmpbuf.ut_id, sh_stat[channel].id);
    strcpy(utmpbuf.ut_line, slave + 5);
    utmpbuf.ut_pid = getpid();
    utmpbuf.ut_type = LOGIN_PROCESS;
    utmpbuf.ut_time = secclock(); 
#ifdef HAS_UTHOST
    strncpy(utmpbuf.ut_host,"AX25", sizeof(utmpbuf.ut_host));
#endif
    pututline(&utmpbuf);
    endutent();
#endif
    sprintf(tmp,"%d",channel);
    execl("/tcp/bin/sp-emu", "sp-emu", ucall, tmp, (char *) 0);
    exit(1);
  }
  cmd_display(mode,channel,"Ok",1);
  return;
}
#endif

#ifdef USE_SOCKET
/*
** Pty-Login via Telnet; by Pat, DL7AUC @ DB0BLO, root@chamber.in-berlin.de
*/

/* When disconnected from a station, TNT calls this procedure */

send_pty_goodbye(channel)
int channel;
{
  if(sh_stat[channel].active == A_SOCKET) {
    if (sd[channel].login_stat>=2)
      sd[channel].login_stat=2;
    else
      sd[channel].login_stat=0;
  }
}

/* Here now the main procedure for socket login/logout/commands */

int do_sock_cmd(channel,str,len)
int channel;
char str[];
int len;
{
  int i;
  char buffer[500];
  char tmp[500];
  FILE *fp;

  strncpy(buffer,str,len);
  buffer[len]=0;

  /* check for password file */
  if(sd[channel].login_stat==0) {
    sprintf(tmp,"%s/%s",tnt_dir,sock_passfile);
    fp=fopen(tmp,"r");
    if(fp==0) {
      sprintf(tmp,"\n\n* Can't open password file *\n\n");
      write_pty(channel,strlen(tmp),tmp);
      sh_stat[channel].connected=0;
      close(sh_stat[channel].pty);
      sd[channel].login_stat=0;
      return(1);
    }
    do {
      fgets(tmp,99,fp);
      tmp[strlen(tmp)-1]=0;
      if(tmp[0]=='#' || tmp[0]==' ' || tmp[0]==9 || tmp[0]==LF) continue;
      sscanf(tmp,"%s%s%s%d",sd[channel].user_id,sd[channel].user_pw,
	     sd[channel].connect,&sd[channel].level);
      if(!strncmp(sd[channel].user_id,buffer,strlen(buffer)-1))
	break;
    } while(!feof(fp));

    if(feof(fp)) {
      strcpy(sd[channel].user_id,"");
      strcpy(sd[channel].user_pw,"");
      strcpy(sd[channel].connect,"");
      sd[channel].level=0;
    }
    fclose(fp);
    if(!strcmp(sd[channel].user_pw,""))
      strcpy(sd[channel].user_pw,"nonepwd-invalid");
    for(i=0;i<strlen(sd[channel].user_id);i++)
      if(sd[channel].user_id[i]=='-') {sd[channel].user_id[i]=0;break;}
    sprintf(tmp,"password: ");
    write_pty(channel,strlen(tmp),tmp);
    sd[channel].login_stat=1;
    return(1);
  }

  /* login_stat = 0 -> first contact
   * login_stat = 1 -> password required after login
   * login_stat = 2 -> login successful, command_mode
   * login_stat > 2 -> data transfer to connected station
   */

  if(sd[channel].login_stat==1) {
    if(!strncmp(sd[channel].user_pw,buffer,strlen(sd[channel].user_pw))
       && sd[channel].level!=0) {
      sprintf(tmp,"\nlogin acknowledged.\n\nCmd> ");
      write_pty(channel,strlen(tmp),tmp);
      sprintf(tmp,"I %s",sd[channel].user_id); /* set users call to tnc */
      cmd_input(channel,M_CMDSCRIPT,tmp,strlen(tmp),0);
      sd[channel].login_stat=2;
      return(1);
    }
    else {
      /* an intruder, close socket descriptor, thereby terminating telnet */
      sprintf(tmp,"\nlogin failed, permission denied.\n\n");
      write_pty(channel,strlen(tmp),tmp);
      sd[channel].login_stat=0;
      sh_stat[channel].connected=0;
      close(sh_stat[channel].pty);
      return(1);
    }
  }
  
  if(sd[channel].login_stat==2) {
    /* set usercall prior to each command (reseted after disconnect !) */
    sprintf(tmp,"I %s",sd[channel].user_id);
    cmd_input(channel,M_CMDSCRIPT,tmp,strlen(tmp),0);

    /* connect command */
    if(!strncmp(buffer,"c",1)) {
      for(i=0;i<strlen(buffer);i++)
	if(buffer[i]==' ') break;
      if(i<strlen(buffer) || (strcmp(sd[channel].connect,"none"))) {
	sprintf(tmp,"* Ok, back to cmd-mode with ':cmd' *\n");
	write_pty(channel,strlen(tmp),tmp);
	sprintf(tmp,"I %s",sd[channel].user_id); /* just to be sure */
	cmd_input(channel,M_CMDSCRIPT,tmp,strlen(tmp),0);
	if(strcmp(sd[channel].connect,"none")) /* autoconnect */
	  sprintf(tmp,"C %s",sd[channel].connect);
	else
	  sprintf(tmp,"C %s",&buffer[i+1]);
	i=strlen(tmp)-1;
	while(buffer[i]=='\n' || buffer[i]=='\r') {
	  buffer[i]=0; i--;
	}
	cmd_input(channel,M_SOCKET,tmp,strlen(tmp),0);
	sd[channel].login_stat=3;
	return(1);
      }
      else {
	sprintf(tmp,"* no callsign ? *\n\nCmd> ");
	write_pty(channel,strlen(tmp),tmp);
	return(1);
      }
    }

    /* disconnect command */
    if(!strncmp(buffer,"d",1)) {
      cmd_input(channel,M_SOCKET,"D",2,0);
      sd[channel].login_stat=2;
      return(1);
    }

    /* exit command (quit) */
    if(!strncmp(buffer,"e",1) || !strncmp(buffer,"q",1)) {
      cmd_input(channel,M_CMDSCRIPT,"D",2,0);
      sprintf(tmp,"*Ok*\n\n");
      write_pty(channel,strlen(tmp),tmp); 
      sd[channel].login_stat=0;
      sh_stat[channel].connected=0;
      close(sh_stat[channel].pty);
      return(1); /* was a command */
    }

    /* extended commands, just for users level=9 ! */
    if(buffer[0]==':') {
      if(sd[channel].level>=9) {
	strncpy(tmp,&buffer[1],sizeof(tmp));
	i=strlen(tmp)-1;
	while(tmp[i]=='\n' || tmp[i]=='\r') {
	 tmp[i]=0; i--;
	}
	cmd_input(channel,M_SOCKET,tmp,strlen(tmp),0);
	return(1);
      }
      else {
	sprintf(tmp,"\n* permission denied *\n\nCmd> ");
	write_pty(channel,strlen(tmp),tmp);
	return(1);
      }
    }

    /* do a little missionary work, help the user, hi */
    if(buffer[0]!='\r' || ch_stat[channel].state<4) {
      sprintf(tmp,"\nInvalid command.\n\n");
      strcat(tmp,"- (c)onnect <callsign>\n");
      strcat(tmp,"- (d)isconnect\n");
      strcat(tmp,"- (e)xit\n");
      strcat(tmp,"\n  To change from converse-mode to command-mode\n");
      strcat(tmp,"  enter ':cmd' while connected to a station.\n\nCmd> ");
      write_pty(channel,strlen(tmp),tmp);
      return(1);
    }
    else { /* switch back to conversation if connected */
      sprintf(tmp,"* Back to converse mode *\n");
      write_pty(channel,strlen(tmp),tmp);
      sd[channel].login_stat=3;
      return(1);
    }

  }

  /* check for escape-string from converse mode to command mode */
  if(sd[channel].login_stat>2) {
    if(!strncmp(buffer,":cmd",4)) {
      sd[channel].login_stat=2;
      sprintf(tmp,"*Ok*\n\nCmd> ");
      write_pty(channel,strlen(tmp),tmp);
      return(1);
    }

    /*
     * here special level for mail delivery, level=99 allows any
     * extended command via '::command' (to open logfiles or uploads)
     */
     
    if(sd[channel].level==99) {
      if(buffer[0]==':' && buffer[1]==':') {
        strncpy(tmp,&buffer[2],sizeof(tmp));
        i=strlen(tmp)-1;
        /* strip off LF or CR at end of input */
        while(tmp[i]=='\n' || tmp[i]=='\r') {
          tmp[i]=0; i--;
        }
        cmd_input(channel,M_CMDSCRIPT,tmp,strlen(tmp),0);
        sprintf(tmp,"\n*Ok*> ");
        write_pty(channel,strlen(tmp),tmp);
        return(1);
      }
    }
    return(0);
  }
  return(0); /* was no command */
}
#endif

int output_screen(channel)
int channel;
{
  if (ptyecho_flag) return(1);
  if (sh_stat[channel].active) return(0);
  return(1);
}

#ifdef USE_SOCKET
int output_screen_sock(channel)
int channel;
{
  if (ptyecho_flag) return(1);
  if (sh_stat[channel].active == A_SOCKET) return(0);
  return(1);
}
#endif

#ifdef USE_SOCKET
void out_socket(channel,str)
int channel;
char *str;
{
  char tmpstr[200];
  
  if (sh_stat[channel].active != A_SOCKET) return;
  strcpy(tmpstr,"\n");
  strcat(tmpstr,rem_tnt_str);
  strcat(tmpstr,str);
  write_pty(channel,strlen(tmpstr),tmpstr);
}
#endif

int shell_active(channel)
int channel;
{
  if (sh_stat[channel].active) return(1);
  return(0);
}

