/* tnt: Hostmode Terminal for TNC
   Copyright (C) 1993,1994 by Mark Wahl
   For license details see documentation
   Procedures for window handling (window.c)
   created: Mark Wahl DL4YBG 93/08/01
   updated: Mark Wahl DL4YBG 94/12/20
*/

#undef TNT
#include "tnt.h"

extern char attc_normal;
extern char attc_statline;
extern char attc_monitor;
extern char attc_cstatline;
extern char attc_controlchar;
extern char attc_remote;
extern char attm_normal;
extern char attm_statline;
extern char attm_monitor;
extern char attm_cstatline;
extern char attm_controlchar;
extern char attm_remote;
extern int color;
extern int termcap;

char att_normal;
char att_statline;
char att_monitor;
char att_cstatline;
char att_controlchar;
char att_remote;
int LINES,COLS;
char *term;
int auto_newline;
int supp_hicntl;
int pagesize;

/* Terminal capabilities */
static char *AL, *DL, *IC, *DC, *WR;
static char *CL, *CM, *CS, *IS, *BC;
static char *ND, *CE, *NL, *SO, *SE;
static char *US, *UE, *VE, *VI, *RS;
static char *MB, *MR, *MD, *MH, *ME;
static char CRNL[10];
static char RUBOUT[10];
static int scroll_reg;
static int ins_char;
static int del_char;
static int ins_line;
static int del_line;
static int cur_always_vis;
static int no_attr;
char *_tptr = NULL;

#ifndef HAS_MEMMOVE
void *memmove(dst,src,len)
void *dst;
const void *src;
unsigned int len;
{
  char buffer[256];
  
  if (len > 256) return;
  memcpy(buffer,src,len);
  memcpy(dst,buffer,len);
}
#endif

static struct win_info real_info[5]; /* real scr contains up to 5 windows */

static void getrowcols(rows, cols)
int *rows;
int *cols;
{
#ifdef TIOCGWINSZ
	struct winsize ws;

	if (ioctl(0, TIOCGWINSZ, &ws) < 0) {
		*rows = 0;
		*cols = 0;
	} else {
		*rows = ws.ws_row;
		*cols = ws.ws_col;
	}	
#else
#  ifdef TIOCGSIZE
	struct ttysize ws;

	if (ioctl(0, TIOCGSIZE, &ws) < 0) {
		*rows = 0;
		*cols = 0;
	} else {
		*rows = ws.ts_lines;
		*cols = ws.ts_cols;
	}
#  else 
	char *p, *getenv();

	if (p = getenv("LINES"))
		*rows = atoi(p);
	else
		*rows = 0;
	if (p = getenv("COLUMNS"))
		*cols = atoi(p);
	else
		*cols = 0;
#  endif
#endif	
}

static void phy_strout(str)
char *str;
{
  write(1,str,strlen(str));
}

int term_init()
{
  static char *tbuf;
  static char *cbuf;
  char *p;

 if ((term = getenv("TERM")) == NULL) {
   printf("Environment variable TERM not set\n");
   return(1);
 }
 if (strcmp(term,"console") != 0) {
   termcap = 1;
   color = 0;
 }
 if (color) {
   att_normal = attc_normal;
   att_statline = attc_statline;
   att_monitor = attc_monitor;
   att_cstatline = attc_cstatline;
   att_controlchar = attc_controlchar;
   att_remote = attc_remote;
 }
 else {
   att_normal = attm_normal;
   att_statline = attm_statline;
   att_monitor = attm_monitor;
   att_cstatline = attm_cstatline;
   att_controlchar = attm_controlchar;
   att_remote = attm_remote;
 }
 if (termcap) {
  if ((tbuf = (char *)malloc(1024)) == NULL ||
  	(cbuf = (char *)malloc(1024)) == NULL) {
  	printf("Out of memory.\n");
  	return(1);
  }
  switch(tgetent(cbuf, term)) {
  	case 0:
  		printf("No termcap entry for %s\n", term);
  		return(1);
  	case -1:
  		printf("No /etc/termcap present!\n");
  		return(1);
  	default:
  		break;
  }
  _tptr = tbuf;

  if ((CM = tgetstr("cm", &_tptr)) == NULL) {
  	printf("No cursor motion capability (cm)\n");
  	return(1);
  }
  getrowcols(&LINES, &COLS);
  if (LINES == 0)
    if ((LINES = tgetnum("li")) <= 0) {
      printf("Number of terminal lines unknown\n");
      return(1);
  }
  if (LINES < 16) {
    printf("Illegal number of lines on terminal: %d (x >= 16)\n",LINES);
    return(1);
  }
  if (COLS == 0)
    if ((COLS = tgetnum("co")) <= 0) {
      printf("Number of terminal columns unknown\n");
      return(1);
  }
  if ((COLS < MINCOLS) || (COLS > MAXCOLS)) {
    printf("Illegal number of columns on terminal: %d (80 <= x <= 160)\n",COLS);
    return(1);
  }

  /* Terminal Capabilities */
  if ((AL = tgetstr("al", &_tptr)) == NULL) {
    ins_line = 0;
  }
  else {
    ins_line = 1;
  }
  if ((DL = tgetstr("dl", &_tptr)) == NULL) {
    del_line = 0;
  }
  else {
    del_line = 1;
  }
  if ((IC = tgetstr("ic", &_tptr)) == NULL) {
    ins_char = 0;
  }
  else {
    ins_char = 1;
  }
  if ((DC = tgetstr("dc", &_tptr)) == NULL) {
    del_char = 0;
  }
  else {
    del_char = 1;
  }
  if ((CL = tgetstr("cl", &_tptr)) == NULL) {
  	printf("No clear screen capability (cl)\n");
  	return(1);
  }
  if ((IS = tgetstr("is", &_tptr)) == NULL)
    IS = "\033[r";
  RS = tgetstr("rs", &_tptr);
  if ((ND = tgetstr("nd", &_tptr)) == NULL) {
  	printf("No cursor right capability (nd)\n");
  	return(1);
  }
  if ((CE = tgetstr("ce", &_tptr)) == NULL) {
  	printf("No clear to end of line capability (ce)\n");
  	return(1);
  }
  if ((WR = tgetstr("cr", &_tptr)) == NULL)
    WR = "\015";
  if ((NL = tgetstr("nl", &_tptr)) == NULL)
    NL = "\012";
  if (tgetflag("bs") || ((BC = tgetstr("bc", &_tptr)) == NULL))
    BC = "\010";
  if ((CS = tgetstr("cs", &_tptr)) == NULL)
    scroll_reg = 0;
  else
    scroll_reg = 1;
  if ((!scroll_reg) && !(ins_line && del_line)) {
    printf
     ("Neither scroll region(cs), nor insert/delete line (al,dl) capability\n");
    return(1);
  }
  VE = tgetstr("ve", &_tptr);
  VI = tgetstr("vi", &_tptr);
  if ((VE == NULL) || (VI == NULL )) cur_always_vis = 1;

  SO = tgetstr("so", &_tptr);
  SE = tgetstr("se", &_tptr);
  US = tgetstr("us", &_tptr);
  UE = tgetstr("ue", &_tptr);
  MB = tgetstr("mb", &_tptr);
  MR = tgetstr("mr", &_tptr);
  MD = tgetstr("md", &_tptr);
  MH = tgetstr("mh", &_tptr);
  ME = tgetstr("me", &_tptr);
  if ((ME == NULL) && (SE == NULL) && (SO == NULL)) no_attr = 1; 
  
  strcpy(CRNL,WR);
  strcat(CRNL,NL);
  strcpy(RUBOUT,BC);
  strcat(RUBOUT," ");
  strcat(RUBOUT,BC);
  return(0);
 }
 else {
  getrowcols(&LINES, &COLS);
  if (LINES == 0)
    if ((LINES = tgetnum("li")) <= 0) {
  	printf("Number of terminal lines unknown\n");
  	return(1);
  }
  if (COLS == 0)
    if ((COLS = tgetnum("co")) <= 0) {
  	printf("Number of terminal columns unknown\n");
  	return(1);
  }
  AL = "\033[L";
  ins_line = 1;
  DL = "\033[M";
  del_line = 1;
  IC = "\033[@";
  ins_char = 1;
  DC = "\033[P";
  del_char = 1;
  CL = "\033[H\033[J";
  IS = "\033[r";
  RS = NULL;
  ND = "\033[a";
  CE = "\033[0K";
  WR = "\015";
  NL = "\012";
  BC = "\010";
  VE = "\033[?25h";
  VI = "\033[?25l";
  cur_always_vis = 0;
  scroll_reg = 1;
  SO = "\033[1m";
  SE = "\033[21m";
  US = "\033[4m";
  UE = "\033[24m";
  MB = "\033[5m";
  MR = "\033[7m";
  MD = "\033[1m";
  MH = "\033[2m";
  ME = "\033[m";
  no_attr = 0; 
  strcpy(CRNL,WR);
  strcat(CRNL,NL);
  strcpy(RUBOUT,BC);
  strcat(RUBOUT," ");
  strcat(RUBOUT,BC);
  return(0);
 }
}

static void phy_treset() 
{
  write(1,"\033c",2);
}

static void phy_newline() 
{
  phy_strout(CRNL);
}

static void phy_insline()
{
  phy_strout(AL);
}

static void phy_delline()
{
  phy_strout(DL);
}

static void phy_charins()
{
  phy_strout(IC);
}

static void phy_chardel()
{
  phy_strout(DC);
}

static void phy_cr()
{
  phy_strout(WR);
}

static void phy_curvis()
{
  if (!cur_always_vis)
    phy_strout(VE);
}

static void phy_curinvis()
{
  if (!cur_always_vis)
    phy_strout(VI);
}

static void phy_clrscr()
{
  phy_strout(CL);
}

static void phy_gotoxy(col,lin)
int col;
int lin;
{
  char tmpstr[20];

  if (!termcap) {
    sprintf(tmpstr,"\033[%u;%uH",lin+1,col+1);
    phy_strout(tmpstr);
  }
  else phy_strout(tgoto(CM,col,lin));
}

static void phy_par_scrdwn(lin1,lin2)
int lin1;
int lin2;
{
  /* cursor to lin2 */
  phy_gotoxy(0,lin2);
  /* delete line */
  phy_delline();
  /* cursor to lin1 */
  phy_gotoxy(0,lin1);
  /* insert line */
  phy_insline();
}

static void phy_par_scrup(lin1,lin2)
int lin1;
int lin2;
{
  /* cursor to lin1 */
  phy_gotoxy(0,lin1);
  /* delete lin */
  phy_delline();
  /* cursor to lin2 */
  phy_gotoxy(0,lin2);
  /* insert lin */
  phy_insline();
}

static void phy_scrreg(lin1,lin2)
int lin1;
int lin2;
{
  char tmpstr[20];
 
  if (scroll_reg) {
    if (termcap) {
      phy_strout(tgoto(CS,lin2,lin1));
    }
    else {
      sprintf(tmpstr,"\033[%u;%ur",lin1+1,lin2+1);
      phy_strout(tmpstr);
    }
  }
}

static void phy_scrnor()
{
  if (scroll_reg) {
    if (termcap) {
      if (RS != NULL)
        phy_strout(RS);
      else
        phy_strout(IS);
    }
    else {
      write(1,"\033[r",3);
    }
  }
}

static void phy_rubout()
{
  phy_strout(RUBOUT);
}

static void phy_curleft()
{
  phy_strout(BC);
}

static void phy_curright()
{
  phy_strout(ND);
}

static void phy_clreol()
{
  phy_strout(CE);
}

static void phy_charout(ch)
char ch;
{
  write(1,&ch,1);
}

static void str_attrout(str,ch)
char *str;
char ch;
{
  char tch;

  if (color) {
    strcat (str,"\033[");
    if (ch & 0x80) strcat(str,"5;");
    else strcat(str,"25;");
    if (ch & 0x08) strcat(str,"1;");
    else strcat(str,"21;");
    strcat(str,"3");
    tch = (ch & 07)+'0';
    strncat(str,&tch,1);
    strcat(str,";4");
    tch = ((ch >> 4) & 07)+'0';
    strncat(str,&tch,1);
    strcat(str,"m");
  }
  else if (!no_attr) {
    /* reset attributes */
    if (ME != NULL) strcat(str,ME);
    else if (SE != NULL) strcat(str,SE);
    else if (UE != NULL) strcat(str,UE);
    if ((ch & 0x01) && (SO != NULL) && !((SE == NULL) && (ME == NULL)))
      strcat(str,SO);
    if ((ch & 0x02) && (US != NULL) && !((UE == NULL) && (ME == NULL)))
      strcat(str,US);
    if ((ch & 0x04) && (MB != NULL) && (ME != NULL)) strcat(str,MB);
    if ((ch & 0x08) && (MR != NULL) && (ME != NULL)) strcat(str,MR);
    if ((ch & 0x10) && (MD != NULL) && (ME != NULL)) strcat(str,MD);
    if ((ch & 0x20) && (MH != NULL) && (ME != NULL)) strcat(str,MH);
  }
}

static void phy_attrout(ch)
char ch;
{
  char str[20];
  int i;

  if (!no_attr) {
    strcpy (str,"");
    str_attrout(str,ch);
    if (i = strlen(str))
      write(1,str,i);
  }
}

static void phy_no_transl()
{
  write(1,"\033(U",3);
}

static void phy_normal_transl()
{
  write(1,"\033(B",3);
}

void beep()
{
  write(1,"\007",1);
}

int window_init()
{
  int i;

  for (i = 0; i < 5; i++) {
    real_info[i].win = NULL;
  }
  phy_clrscr();
}

int window_exit()
{
  phy_clrscr();
  phy_scrnor();
  phy_treset();
}

/* test, if insert mode active */
int insert_active()
{
  return (real_info[0].win->insertflag);
}

/* open new window, returns 0 on success, 1 if no memory available */
int open_window(lines,win,bot)
int lines;
struct window *win;
int bot;
{
  int buffer_len;
  char *add;
  
  win->num_lines = lines;
  if (bot)
    win->line = lines - 1;
  else
    win->line = 0;
  win->column = 0;
  win->line_offset = 0;
  win->phys_line = win->line; /* equal because line_offset == 0 */
  win->real = -1;
  win->attribut = att_normal;
  win->insertflag = 0;
  buffer_len = lines * 2 * COLS;
  win->window_buffer = (char *)malloc(buffer_len);
  if (win->window_buffer == NULL) return(1);
  add = win->window_buffer;
  while (buffer_len > 0) {
    *add++ = att_normal;
    *add++ = SPACE;
    buffer_len -= 2 ;
  }
  return(0);
}

/* clears window (clear screen) */
int clear_window(win)
struct window *win;
{
  int buffer_len;
  char *add;
  int lines;
  
  lines = win->num_lines;
  win->line = lines - 1;
  win->column = 0;
  win->line_offset = 0;
  win->phys_line = win->line; /* equal because line_offset == 0 */
  win->real = -1;
  win->attribut = att_normal;
  win->insertflag = 0;
  buffer_len = lines * 2 * COLS;
  add = win->window_buffer;
  while (buffer_len > 0) {
    *add++ = att_normal;
    *add++ = SPACE;
    buffer_len -= 2 ;
  }
}

/* closes window and frees memory, returns 1 if window already closed */
int close_window(win)
struct window *win;
{
  if (win->window_buffer == NULL) return(1);
  if (win->real != -1) {
    real_info[win->real].win = NULL;
  }
  free(win->window_buffer);
  win->window_buffer = NULL;
  return(0);
}

/* increment window-line-offset with boundary-check */
static void inc_line_offset(win)
struct window *win;
{
  if (++win->line_offset == win->num_lines) win->line_offset = 0;
}

/* increment window-physical-line with boundary-check */
static void inc_phys_line(win)
struct window *win;
{
  if (++win->phys_line == win->num_lines) win->phys_line = 0;
}

/* decrement window-physical-line with boundary-check */
static void dec_phys_line(win)
struct window *win;
{
  if (win->phys_line == 0) win->phys_line = win->num_lines - 1;
  else win->phys_line--;
}

/* clear to end of line from col(included) in physical line  */
static void cl_eoline(win)
struct window *win;
{
  char *add;
  int i;
  
  add = win->window_buffer + 2 * (COLS * win->phys_line + win->column);
  i = COLS - win->column;
  while (i > 0) {
    *add++ = att_normal;
    *add++ = SPACE;
    i--;    
  }
}

/* function returns 1 if line is the last line of the physical screen */
int last_phy_line(win,line)
struct window *win;
int line;
{
  int last_line;
  
  if (real_info[win->real].last_real_line == (LINES - 1)) {
    last_line = real_info[win->real].last_log_line;
    if ((last_line += win->line_offset) >= win->num_lines)
      last_line -= win->num_lines;
    if (last_line == line) return(1);
  }
  return(0);
}

/* copy contents of one windowline to real screen
   assumes real cursor is already positioned at proper position */
static void update_line(win,phy_line)
struct window *win;
int phy_line;
{
  char att;
  char *addr;
  int i;
  int max_col;
  char *tmpstr;

  tmpstr = (char *)malloc(15*COLS+1);
  max_col = COLS;
  if (auto_newline) {
    if (last_phy_line(win,phy_line)) max_col = COLS - 1;
  }
  addr = win->window_buffer + phy_line * 2 * COLS;
  phy_attrout(att = *addr);
  strcpy(tmpstr,"");
  for (i = 0; i < max_col; i++) {
    /* if attribute not current, change it */
    if (*addr != att) {
      att = *addr;
      str_attrout(tmpstr,att);
    }
    addr++;
    strncat(tmpstr,addr,1);
    addr++;
  }
  if (!termcap) phy_no_transl();
  phy_strout(tmpstr);
  free(tmpstr);
  if (!termcap) phy_normal_transl();
}

/* physical window will be scrolled up and last line will be updated */
static void win_scrollup(part)
int part;
{
  struct window *win;
  int line;
  int phy_line;

  win = real_info[part].win;
  line = real_info[part].last_real_line;
  if (!scroll_reg)
    phy_par_scrup(real_info[part].first_real_line,line);
  else {  
    phy_gotoxy(0,line);
    phy_newline(part);
  }
  line = real_info[part].last_log_line;
  if ((phy_line = line + win->line_offset) >= win->num_lines)
    phy_line -= win->num_lines; 
  update_line(real_info[part].win,phy_line);
  if (auto_newline) {
    phy_gotoxy(0,real_info[part].last_real_line);
  }
  else {
    phy_cr();
  }
}

/* put cursor in window to new line and scroll if needed */
static void newline(win)
struct window *win;
{
  /* new column is 0 */
  win->column = 0;
  /* next line */
  if (win->line == (win->num_lines - 1)) {
    /* in last line, scroll buffer */
    inc_line_offset(win);
    inc_phys_line(win);
    cl_eoline(win);
    /* check if window in real screen */
    if (win->real != -1) {
      if (!scroll_reg) {
        /* scroll physical window 1 line up and update line */
        win_scrollup(win->real);
      }
      else {
        if (real_info[win->real].last_log_line == (win->num_lines - 1)) {
          /* newline does the scroll */
          phy_newline(win->real);
        }
        else {
          /* scroll physical window 1 line up and update line */
          win_scrollup(win->real);
        }
      }
    }
  }
  else {
    /* not in last line, no scroll needed */
    win->line++;
    inc_phys_line(win);
    /* check if window in real screen */
    if (win->real != -1) {
      /* if new line in physical window (except first line) do
         physical new line */
      if ((win->line > (real_info[win->real].first_log_line)) &&
          (win->line <= (real_info[win->real].last_log_line))) {
        phy_newline(win->real);
      }
      /* if newline first in physical window, cursor must be positioned */
      else if (win->line == (real_info[win->real].first_log_line)) {
        phy_gotoxy(0,real_info[win->real].first_real_line);
      }
      else if (win->line > (real_info[win->real].last_log_line)) {
        if (win->real == 0) {
          real_window_down(win->real,0);
        }
      }
    }
  }
}

/* put cursor to start of line (for statuslines) */
static void crstatline(win)
struct window *win;
{
  win->column = 0;
  if (win->real != -1) {
    phy_cr();
  }
}

/* copy contents of window to real screen */
static void update_window(part)
int part;
{
  struct window *win;
  int line;
  int phy_line;

  phy_gotoxy(0,real_info[part].first_real_line);  
  win = real_info[part].win;
  line = real_info[part].first_log_line;
  if ((phy_line = line + win->line_offset) >= win->num_lines)
    phy_line -= win->num_lines;
  while (line <= real_info[part].last_log_line) {
    update_line(win,phy_line);
    /* newline not after last line or if auto_newline */
    if (!auto_newline) {
      if (line != real_info[part].last_log_line) phy_newline(part);
    }
    line++;
    if ((phy_line = line + win->line_offset) >= win->num_lines)
      phy_line -= win->num_lines;
  }
}

static void set_cursor(part)
int part;
{
  struct window *win;
  int line;
  
  win = real_info[part].win;
  if (scroll_reg)
    phy_scrreg(real_info[part].first_real_line,real_info[part].last_real_line);
  if ((win->line >= real_info[part].first_log_line) &&
      (win->line <= real_info[part].last_log_line)) {
    line = win->line - real_info[part].first_log_line
                     + real_info[part].first_real_line;
    phy_gotoxy(win->column,line);
    phy_attrout(win->attribut);
    if (part != 0) phy_curinvis(); 
    else phy_curvis();
  }
  else {
    phy_curinvis();
  }
}

int real_screen(layout)
struct real_layout layout[];
{
  int i;
  
  for (i = 0; ((real_info[i].win != NULL) && (i < 5)); i++) {
    real_info[i].win->real = -1;
  }
  phy_clrscr();
  phy_curinvis();
  phy_scrnor();
  for (i = 0; ((layout[i].win != NULL) && (i < 5)); i++) {
    real_info[i].win = layout[i].win;
    real_info[i].win->real = i;
    real_info[i].first_real_line = layout[i].first_real_line;
    real_info[i].pagesize = layout[i].pagesize;
    
    if ((real_info[i].win_num_lines = layout[i].win_num_lines) >
         real_info[i].win->num_lines)
      real_info[i].win_num_lines = real_info[i].win->num_lines;
    real_info[i].last_real_line =
      layout[i].first_real_line + real_info[i].win_num_lines - 1;
    real_info[i].last_log_line = real_info[i].win->num_lines - 1;
    real_info[i].first_log_line =
      real_info[i].last_log_line - (real_info[i].win_num_lines - 1);
    if (real_info[i].win->line < real_info[i].first_log_line) {
      real_info[i].first_log_line = real_info[i].win->line;
      real_info[i].last_log_line = real_info[i].first_log_line
                                 + real_info[i].win_num_lines - 1;
    }
    else if (real_info[i].win->line > real_info[i].last_log_line) {
      real_info[i].last_log_line = real_info[i].win->line;
      real_info[i].first_log_line = real_info[i].last_log_line
                                 - (real_info[i].win_num_lines - 1);
    }
    update_window(i);
  }
  set_cursor(0);
}

/* scroll real windowpart line or page up */
int real_window_up(part,page)
int part;
int page;
{
  int lines;
  int line;
  int phy_line;
  struct window *win;
  
  if (real_info[part].first_log_line == 0) return;
  if (page)
    lines = real_info[part].pagesize;
  else
    lines = 1;
  if ((lines == 1) && ins_line) {
    real_info[part].first_log_line--;
    real_info[part].last_log_line--;
    set_cursor(part);
    phy_curinvis();
    line = real_info[part].first_real_line;
    if (!scroll_reg) {
      phy_par_scrdwn(real_info[part].first_real_line,
                     real_info[part].last_real_line);
    }
    else {
      phy_gotoxy(0,line);
      phy_insline();
    }
    line = real_info[part].first_log_line;
    win = real_info[part].win;
    if ((phy_line = line + win->line_offset) >= win->num_lines)
      phy_line -= win->num_lines;
    update_line(win,phy_line);
    set_cursor(0);
  }
  else {
    if (real_info[part].first_log_line < lines) {
      lines = real_info[part].first_log_line;
    }  
    real_info[part].first_log_line -= lines;
    real_info[part].last_log_line -= lines;
    set_cursor(part);
    phy_curinvis();
    update_window(part);
    set_cursor(0);
  }
}

/* scroll real windowpart to top */
int real_window_top(part)
int part;
{
  if (real_info[part].first_log_line == 0)
    return;
  real_info[part].first_log_line = 0;
  real_info[part].last_log_line = real_info[part].win_num_lines - 1;
  set_cursor(part);
  phy_curinvis();
  update_window(part);
  set_cursor(0);
}


/* scroll real windowpart line or page down */
int real_window_down(part,page)
int part;
int page;
{
  int lines;
  int line;
  int phy_line;
  struct window *win;
  
  if (real_info[part].last_log_line == (real_info[part].win->num_lines - 1))
    return;
  if (page)
    lines = real_info[part].pagesize;
  else
    lines = 1;
  if ((lines == 1) && del_line) {
    real_info[part].first_log_line++;
    real_info[part].last_log_line++;
    set_cursor(part);
    phy_curinvis();
    line = real_info[part].first_real_line;
    if (!scroll_reg) {
      phy_par_scrup(real_info[part].first_real_line,
                    real_info[part].last_real_line);
    }
    else {
      phy_gotoxy(0,line);
      phy_delline();
    }
    line = real_info[part].last_real_line;
    phy_gotoxy(0,line);
    line = real_info[part].last_log_line;
    win = real_info[part].win;
    if ((phy_line = line + win->line_offset) >= win->num_lines)
      phy_line -= win->num_lines;
    update_line(win,phy_line);
    set_cursor(0);
  }
  else {
    if ((real_info[part].last_log_line + lines) >
        (real_info[part].win->num_lines - 1)) {
      lines = real_info[part].win->num_lines - 1 - 
              real_info[part].last_log_line;
    }     
    real_info[part].first_log_line += lines;
    real_info[part].last_log_line += lines;
    set_cursor(part);
    phy_curinvis();
    update_window(part);
    set_cursor(0);
  }
}

/* scroll real windowpart to bottom */
int real_window_bot(part)
int part;
{
  if (real_info[part].last_log_line == (real_info[part].win->num_lines - 1))
    return;
  real_info[part].first_log_line = 
    real_info[part].win->num_lines - 1 - real_info[part].win_num_lines + 1;
  real_info[part].last_log_line = real_info[part].win->num_lines - 1;
  set_cursor(part);
  phy_curinvis();
  update_window(part);
  set_cursor(0);
}

/* execute cursor backspace */
static void backspace(win)
struct window *win;
{
  char *pos;
  int line;
  
  if (win->column == 0) return;
  win->column--;
  /* calculate character position and store in buffer */
  pos = win->window_buffer + win->phys_line*2*COLS + 2*win->column;
  if (win->insertflag) {
    memmove(pos,pos+2,2*(COLS - 1 - win->column));
    pos = win->window_buffer + (win->phys_line+1)*2*COLS - 2;
  } 
  *pos++ = win->attribut;
  *pos = SPACE;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      if (win->insertflag) {
        phy_curleft();
        if (del_char)
          phy_chardel();
        else {
          phy_cr();
          update_line(win,win->phys_line);
          line = win->line - real_info[win->real].first_log_line
                           + real_info[win->real].first_real_line;
          phy_gotoxy(win->column,line);
        }       
      }
      else {
        phy_rubout();
      }
    }
  }
}

/* execute delete character */
static void cur_delchar(win)
struct window *win;
{
  char *pos;
  int line;
  
  if (win->column >= COLS - 1) return;
  /* calculate character position and store in buffer */
  pos = win->window_buffer + win->phys_line*2*COLS + 2*win->column;
  memmove(pos,pos+2,2*(COLS - 1 - win->column));
  pos = win->window_buffer + (win->phys_line+1)*2*COLS - 2;
  *pos++ = win->attribut;
  *pos = SPACE;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      if (del_char)
        phy_chardel();
      else {
        phy_cr();
        update_line(win,win->phys_line);
        line = win->line - real_info[win->real].first_log_line
                         + real_info[win->real].first_real_line;
        phy_gotoxy(win->column,line);
      }       
    }
  }
}

/* execute cursor left */
static void cur_left(win)
struct window *win;
{
  if (win->column == 0) return;
  win->column--;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      phy_curleft();
    }
  }
}

/* execute cursor right */
static void cur_right(win)
struct window *win;
{
  if (win->column == (COLS - 1)) return;
  win->column++;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      phy_curright();
    }
  }
}

/* execute cursor start of line */
static void cur_stline(win)
struct window *win;
{
  if (win->column == 0) return;
  win->column = 0;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      phy_cr();
    }
  }
}

/* execute cursor end of line */
static void cur_eoline(win)
struct window *win;
{
  char *pos;
  int column;
  int line;
  
  column = COLS-1;
  pos = win->window_buffer + win->phys_line*2*COLS + 2*column + 1;
  /* find first position not SPACE */
  while ((*pos == SPACE) && (column > 0)) {
    pos -= 2;
    column--;
  }
  if (column < COLS-1) column++; /* position at space after char */
  win->column = column;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      line = win->line - real_info[win->real].first_log_line
                       + real_info[win->real].first_real_line;
      phy_gotoxy(win->column,line);
    }
  }
}

/* execute cursor erase to end of line */
static void cur_eroline(win)
struct window *win;
{
  cl_eoline(win);
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      phy_clreol();
    }
  }
}

/* execute cursor up */
static void cur_up(win)
struct window *win;
{
  int line;
  
  if (win->line == 0) return;
  win->line--;
  dec_phys_line(win);
  if (win->line < real_info[win->real].first_log_line) {
    real_window_up(0,0);
  }
  line = win->line - real_info[win->real].first_log_line
                   + real_info[win->real].first_real_line;
  phy_gotoxy(win->column,line);
}

/* execute cursor page up */
static void cur_up_page(win)
struct window *win;
{
  if (win->line == 0) return;
  if ((win->line -= real_info[0].pagesize) < 0) win->line = 0;
  if ((win->phys_line = win->line + win->line_offset) >= win->num_lines)
    win->phys_line -= win->num_lines; 
  real_window_up(0,1);
  set_cursor(0);
}

/* execute cursor top of window */
static void cur_top(win)
struct window *win;
{
  if (win->line == 0) return;
  win->line = 0;
  if ((win->phys_line = win->line + win->line_offset) >= win->num_lines)
    win->phys_line -= win->num_lines; 
  real_window_top(0);
  set_cursor(0);
}

/* execute cursor down */
static void cur_down(win)
struct window *win;
{
  int line;
  
  if (win->line == (win->num_lines - 1)) return;
  win->line++;
  inc_phys_line(win);
  if (win->line > real_info[win->real].last_log_line) {
    real_window_down(0,0);
  }
  line = win->line - real_info[win->real].first_log_line
                   + real_info[win->real].first_real_line;
  phy_gotoxy(win->column,line);
}

/* execute cursor page down */
static void cur_down_page(win)
struct window *win;
{
  if (win->line == (win->num_lines - 1)) return;
  if ((win->line += real_info[0].pagesize) > (win->num_lines - 1))
    win->line = win->num_lines - 1;
  if ((win->phys_line = win->line + win->line_offset) >= win->num_lines)
    win->phys_line -= win->num_lines; 
  real_window_down(0,1);
  set_cursor(0);
}

/* execute cursor bottom of window */
static void cur_bot(win)
struct window *win;
{
  if (win->line == (win->num_lines - 1)) return;
  win->line = win->num_lines - 1;
  if ((win->phys_line = win->line + win->line_offset) >= win->num_lines)
    win->phys_line -= win->num_lines; 
  real_window_bot(0);
  set_cursor(0);
}

/* output of character, no control-chars allowed */
static void win_charout_noctl(ch,win,ins)
char ch;
struct window *win;
int ins;
{
  char *pos;
  int line;

  if (auto_newline) {
    /* don't print last character of screen */
    if ((win->column == (COLS - 1)) &&
        last_phy_line(win,win->phys_line)) {
      return;
    }
  }
  if (supp_hicntl) {
    if ((ch >= 128) && (ch <= 160)) ch = '.'; 
  }
  /* no scroll for statuslines, ignore characters at end of line */
  if ((win->column == COLS) && (win->num_lines == 1)) return;
  /* calculate character position and store in buffer */
  pos = win->window_buffer + win->phys_line*2*COLS + 2*win->column;
  if (ins && (win->insertflag) && (win->column < COLS - 2)) {
    memmove(pos+2,pos,2*(COLS - 1 - win->column));
  }
  *pos++ = win->attribut;
  *pos = ch;
  /* output on real screen ? */
  if (win->real != -1) {
    if ((win->line >= real_info[win->real].first_log_line) &&
        (win->line <= real_info[win->real].last_log_line)) {
      if (ins && (win->insertflag) && (win->column < COLS - 2)) {
        if (ins_char) {
          phy_charins();
          phy_charout(ch);
        }
        else {
          phy_cr();
          update_line(win,win->phys_line);
          line = win->line - real_info[win->real].first_log_line
                           + real_info[win->real].first_real_line;
          phy_gotoxy(win->column,line);
          phy_curright();
        }       
      }
      else phy_charout(ch);
    }
  }
  /* advance cursor in buffer */
  if (++win->column == COLS) { /* was in last column ? */
    if (win->num_lines != 1) newline(win);
  }
}

/* output of character */
void win_charout_2(ch,win,conv)
char ch;
struct window *win; 
int conv;
{
  char newch1,newch2;
  int cnt;
  
  if ((ch < SPACE) || (ch == DEL) || (ch == ESC2)) {
    switch (ch) {
    case CR:
      /* no newline on statuslines ! */
      if (win->num_lines != 1) newline(win);
      else crstatline(win);
      break;
    case DEL:
      if (termcap) { 
        win_charout_noctl('^',win,0);
        win_charout_noctl('0',win,0);
      }
      else win_charout_noctl(DEL,win,0);
      break;
    case ESC2:
      if (termcap) { 
        win_charout_noctl('^',win,0);
        win_charout_noctl('1',win,0);
      }
      else win_charout_noctl(ESC2,win,0);
      break;
    default:
      win_charout_noctl('^',win,0);
      win_charout_noctl(ch + '@',win,0);
      break;
    }
  }
  else {
    if (termcap && conv && (ch >= 0x80)) {
      cnt = conv_rx_to_local(ch,&newch1,&newch2);
      if (cnt) { 
        win_charout_noctl(newch1,win,0);
        if (cnt == 2)
          win_charout_noctl(newch2,win,0);
      }
      else win_charout_noctl(ch,win,0);
    }
    else win_charout_noctl(ch,win,0);
  }
}

/* output of character */
void win_charout(ch,win,conv)
char ch;
struct window *win;
int conv; 
{
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(win->real);
  }
  win_charout_2(ch,win,conv);
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(0);
  }
}

/* output of 0-terminated string */
void win_stringout(str,win,conv)
char *str;
struct window *win;
int conv;
{
  char ch;
  
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(win->real);
  }
  if (!termcap && (win->real != -1)) {
    phy_no_transl();
  }
  while ((ch = *str++) != 0) win_charout_2(ch,win,conv);
  if (!termcap && (win->real != -1)) {
    phy_normal_transl();
  }
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(0);
  }
}

/* output of string, length in 1.byte (0 = 256 byte) */
void win_textout(str,win,conv)
char *str;
struct window *win;
int conv;
{
  int len;
  
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(win->real);
  }
  if (!termcap && (win->real != -1)) {
    phy_no_transl();
  }
  len = (*str++) + 1;
  while (len--) win_charout_2(*str++,win,conv);
  if (!termcap && (win->real != -1)) {
    phy_normal_transl();
  }
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(0);
  }
}

/* output of string, length in len */
void win_textout_len(str,len,win,conv)
char *str;
int len;
struct window *win;
int conv;
{
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(win->real);
  }
  if (!termcap && (win->real != -1)) {
    phy_no_transl();
  }
  while (len--) win_charout_2(*str++,win,conv);
  if (!termcap && (win->real != -1)) {
    phy_normal_transl();
  }
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(0);
  }
}

/* set new attribute */
void win_attrib(ch,win)
char ch;
struct window *win;
{
  win->attribut = ch;
  if (win->real == 0)
    phy_attrout(ch);
}

/* get line window */
int get_line(str,win)
char *str;
struct window *win;
{
  int i;
  char *ptr;
  char attr;
  char newchr;
  
  ptr = win->window_buffer + win->phys_line * 2 * COLS;
  for (i = 0; i < win->column; i++) {
    attr = *ptr++;
    str[i] = *ptr++;
    if (attr != att_normal) {
      str[i] -= '@';
    }
    if (str[i] >= 0x80) {
      if (conv_umlaut_to_tx(str[i],&newchr))
        str[i] = newchr;
    }
  }
  str[i++] = CR;
  return(i);
}

/* get line window without conversion */
int get_line_noconv(str,win)
char *str;
struct window *win;
{
  int i;
  char *ptr;
  char attr;
  char newchr;
  
  ptr = win->window_buffer + win->phys_line * 2 * COLS;
  for (i = 0; i < win->column; i++) {
    ptr++;
    str[i] = *ptr++;
  }
  str[i++] = CR;
  return(i);
}

/* output of character and controlcodes */
void win_charout_cntl(ch,win)
char ch;
struct window *win; 
{
  char newch1,newch2;
  int cnt;

  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(win->real);
  }
  if ((ch < SPACE) || (ch == DEL) || (ch == ESC2)) {
    switch (ch) {
    case CR:
      /* no newline on statuslines ! */
      if (win->num_lines != 1) newline(win);
      else crstatline(win);
      break;
    case DEL:
    case BS:
      backspace(win);
      break;
    case C_CULEFT:
      cur_left(win);
      break;
    case C_CURIGHT:
      cur_right(win);
      break;
    case C_CUUP:
      cur_up(win);
      break;
    case C_CUDWN:
      cur_down(win);
      break;
    case C_WINUP:
      cur_up_page(win);
      break;
    case C_WINDWN:
      cur_down_page(win);
      break;
    case C_CUTOP:
      cur_top(win);
      break;
    case C_CUBOT:
      cur_bot(win);
      break;
    case C_STLINE:
      cur_stline(win);
      break;
    case C_EOLINE:
      cur_eoline(win);
      break;
    case C_INSERT:
      if (win->insertflag)
        win->insertflag = 0;
      else
        win->insertflag = 1;
      statlin_update();
      break;
    case C_DELLINE:
      cur_eroline(win);
      break;
    case C_DELCHAR:
      cur_delchar(win);
      break;
    case ESC2:
      win_charout_noctl('^',win,1);
      win_charout_noctl('1',win,1);
      break;
    default:
      win_charout_noctl('^',win,1);
      win_charout_noctl(ch + '@',win,1);
      break;
    }
  }
  else {
    if (!termcap && (win->real != -1)) {
      phy_no_transl();
    }
    if (ch >= 0x80) {
      cnt = conv_local_to_umlaut(ch,&newch1,&newch2);
      if (cnt) {
        win_charout_noctl(newch1,win,1);
        if (cnt == 2)
          win_charout_noctl(newch2,win,1);
      }
      else win_charout_noctl(ch,win,1);
    }
    else win_charout_noctl(ch,win,1);
    if (!termcap && (win->real != -1)) {
      phy_normal_transl();
    }
  }
  if ((win->real != -1) && (win->real != 0)) {
    set_cursor(0);
  }
}

